type t = Kernel.actype

(************ POLY functions ***************)

type 'a alphabet =
    AlphaArray of 'a array
  | AlphaFun of (int * (int -> 'a))

let ios ?(printer=(function _ -> "?")) alphabet = match alphabet with 
    AlphaArray alphabet -> (
      let n = Array.length alphabet - 1 in
      let rec doit x i = 
	if i < 0 then raise (Kernel.BadACML ("State ("^(printer x)^") outside given array alphabet"))
	else if alphabet.(i) = x then i 
	else doit x (i-1) in
      let ht = Hashtbl.create n in
	(fun x ->
	   try Hashtbl.find ht x
	   with Not_found -> let sudo_find_it = doit x n in
	     Hashtbl.add ht x sudo_find_it;
	     sudo_find_it
	)
    )
  | AlphaFun (n, soi) ->
      (fun x -> 
	 let rec doit i =
	   if i < 0 then raise (Kernel.BadACML ("State ("^(printer x)^") outside image of given functional alphabet"))
	   else if soi i = x then i else doit (i-1) in
	   doit (n-1)
      )

let soi = function 
    AlphaArray alphabet -> (fun i -> alphabet.(i))
  | AlphaFun (n,soi) -> soi
      
let cardinal = function
    AlphaArray a -> Array.length a
  | AlphaFun (n,_) -> n

(************************************************)

let radius  = Kernel.getradius 
let dimension  = Kernel.getdimension 
let arity  = Kernel.arity 
let cardinality = fun a -> a.Kernel.cardinality

     
     
let param = Kernel.param
let bool_param = Kernel.param_bool
let float_param = Kernel.param_float
let choice_param = Kernel.param_choice
let choice_array_param a ?choice n arr =
  let l = Array.to_list arr in
  let c = match choice with
      None -> Kernel.param_choice a n l
    | Some i -> if 0<= i && i < Array.length arr then Kernel.param_choice a ~choice:arr.(i) n l else Kernel.param_choice a n l
  in
  let rec idx i =
    if i=0 then 0 
    else if arr.(i) = c then i
    else idx (i-1)
  in idx (Array.length arr - 1)
let file_param = Kernel.param_string true
let text_param ac name x = (
  let a = Kernel.param_string false ac name x in
    Kernel.param_set_text ac name;
    a
)
let string_param = Kernel.param_string false
let set_param = Kernel.set_param
let set_bool_param = Kernel.set_bool_param
let set_float_param = Kernel.set_float_param
let set_string_param = Kernel.set_string_param
let set_choice_param = Kernel.set_choice_param
let get_param = Kernel.get_param
let get_bool_param = Kernel.get_bool_param
let get_float_param = Kernel.get_float_param
let get_string_param = Kernel.get_string_param
let get_choice_param = Kernel.get_string_param

let add_mem_info = Kernel.add_meminfo

let clean_decorations a = a.Kernel.parameters <- Kernel.defaultparamschem ()
let remove_param = Kernel.clear_param
let filter_params = Kernel.filter_params

let (declare_object, get_objects) = Kernel.declare_object, Kernel.get_objects

let declare_file_type ?(transform=(fun x -> x)) name extensions =
  if !Interactive.interactive then Interactive.add_file_filter transform name (List.map (fun x -> "*."^x) extensions)

let apply = Kernel.apply_builtin
let get_builtin_names = Kernel.get_builtin_nicks
let get_external_names = Kernel.get_ext_nicks
let initialize = Kernel.makeinitial

let doc_param a n s = 
  let n = Kernel.real_param_name n in
  a.Kernel.parameters.Kernel.metaparams <- (n, Kernel.Doc s)::(List.remove_assoc n a.Kernel.parameters.Kernel.metaparams)

let doc_action a n s = 
  a.Kernel.parameters.Kernel.metaparams <- (n, Kernel.Doc s)::(List.remove_assoc n a.Kernel.parameters.Kernel.metaparams)

let group_param a n l = a.Kernel.parameters.Kernel.metaparams <- 
  (n, Kernel.Grouparam l)::(List.remove_assoc n a.Kernel.parameters.Kernel.metaparams)

let set_dimension a d =
  if Kernel.neednewac a then 
    a.Kernel.dimension <- d

let set_cardinality a n =
  if Kernel.neednewac a then (
    if n>=0 then
      a.Kernel.cardinality <- n
    else raise (Kernel.BadACML "Non-positive number of states")
  )

let set_moore_neighbourhood a r =
  if Kernel.neednewac a then (
    if r>=0 then
      a.Kernel.neighbourhood <- Kernel.Moore(r)
    else raise (Kernel.BadACML "Non-positive radius")
  )
    
let set_cross_neighbourhood a r =
  if Kernel.neednewac a then (
    if r>=0 then
      a.Kernel.neighbourhood <- Kernel.Cross(r)
    else raise (Kernel.BadACML "Non-positive radius")
  )
    
let set_vonneumann_neighbourhood a r =
  if Kernel.neednewac a then (
    if r>=0 then
      a.Kernel.neighbourhood <- Kernel.VonNeumann(r)
    else raise (Kernel.BadACML "Non-positive radius")
  )	
  
let set_moore_neighborhood = set_moore_neighbourhood
let set_cross_neighborhood = set_cross_neighbourhood
let set_vonneumann_neighborhood = set_vonneumann_neighbourhood

let set_state_name = Kernel.set_statename
let get_state_name = Kernel.get_statename

let poly_set_state_name ac ~alphabet st name =
  let n = cardinal alphabet in
  set_cardinality ac n;
  set_state_name ac (ios alphabet st) name

let poly_get_state_name ac ~alphabet st =
  get_state_name ac (ios alphabet st)

let std_params ?dimension ?(dimension_fixed=false) ?states ?(states_fixed=false) ?radius ?(radius_fixed=false) ?neighborhood ?(neighborhood_fixed=false) a =
  let getopion n = function
      Some x -> x
    | _ -> n
  in
  let pred = (getopion 2 dimension) in
  let pres = (getopion 2 states) in
  let prer = (getopion 1 radius) in
  let do_param fixed a st info x y v = if not fixed then (
    doc_param a st info;
    param a st x y v 
  ) else v in
  let d = do_param dimension_fixed a "dimension" "Lattice dimension" 1 2 pred in
  let s = do_param states_fixed a "states" "Number of states" 2 (1 lsl 29) pres in
  let r = do_param radius_fixed a "radius" "Radius of the neighborhood" 1 (1 lsl 20) prer in
  let obey_neighb x = match x with
      "vonNeumann" -> (set_vonneumann_neighbourhood a r; x)
    | "Cross" -> (set_cross_neighbourhood a r; x)
    | _ -> (set_moore_neighbourhood a r; x)
  in
  let n = if neighborhood_fixed then (match neighborhood with
					| Some "vonNeumann" -> (set_vonneumann_neighbourhood a r; "vonNeumann")
					| Some "Cross" -> (set_cross_neighbourhood a r; "Cross")
					| _ -> (set_moore_neighbourhood a r; "Moore"))
  else (
    doc_param a "neighborhood" "Shape of the neighborhood";
    match neighborhood with
      | Some "vonNeuman" -> obey_neighb (choice_param a "neighborhood" ["vonNeumann"; "Moore"; "Cross"])
      | Some "Cross" -> obey_neighb (choice_param a "neighborhood" [ "Cross"; "vonNeumann"; "Moore"])
      | _ ->  obey_neighb (choice_param a "neighborhood" ["Moore"; "vonNeumann"; "Cross"]) 
  ) in
    (
      set_dimension a d;
      set_cardinality a s;
      (d,s,r,n)
    )

let infomsg s =
  Debug.debugline Debug.Dbginteract s;
  if !Interactive.interactive then Interactive.setstatus s

let myargnb  = Kernel.myargnb 
let mygetarg  = Kernel.mygetarg 

let centercell  = Kernel.centercell 
let relativecell  = Kernel.relativecell 
let neighbors = Kernel.neighbors


let initial  = Kernel.initial 

let ca_random x = 
  if x > 0 then Kernel.myrandom x else 0
let conf_random x  = 
  if x > 0 then Kernel.myrandomnocount x else 0
let ca_random_float = Kernel.myrandom_float
let conf_random_float = Kernel.myrandomnocount_float

let needversion  = Kernel.needversion 
let giveaname  = Kernel.giveaname 
let set_name  = Kernel.giveaname 
let get_name = Kernel.whatsyourname
let set_object_name  = Kernel.giveanobjname
let get_object_name = Kernel.whatsyourobjname

let neednewac  = Kernel.neednewac 
let neednewconf  = Kernel.neednewconf 

let valid  f = 
  let g a = (f a; a) in
  Kernel.valid g

let set_rule a ?cache ?no_table f = 
  a.Kernel.transitionrule <- Kernel.defaulttablerule ();
  let mod_cache = match cache with
      None -> None
    | Some c -> Some (fun f -> let f' = f 0 in 
		      let c' = c f' in (fun _ x -> c' x))
  in
    if Kernel.allow_table a then (match no_table with
				      (Some true) -> a.Kernel.determinization <- Kernel.DeterministicButSlow
				    | _ -> ());
    Kernel.table_from_fun_array mod_cache a (fun c -> f)

let set_rule_fun a f = 
  a.Kernel.transitionrule <- Kernel.defaulttablerule ();
  Kernel.table_from_fun_fun a f

let set_multiset_rule a ?cache ?no_table f = 
  a.Kernel.transitionrule <- Kernel.defaultmultisettablerule ();
  if Kernel.allow_table a then (match no_table with
				    (Some true) -> a.Kernel.determinization <- Kernel.DeterministicButSlow
				  | _ -> ());
  Kernel.table_from_fun_array cache a f

let set_multiset_rule_fun a f = 
  a.Kernel.transitionrule <- Kernel.defaultmultisettablerule ();
  Kernel.table_from_multiset_fun_fun a f


let declare_stochastic a = a.Kernel.determinization <- Kernel.Stochastic
let declare_deterministic a = a.Kernel.determinization <- Kernel.Deterministic
let ensure_determinism a = a.Kernel.determinization <- Kernel.WantToBeDeterministic


type representation =
    Unicolor of int * int *int
  | Transparent
  | NorthArrow
  | SouthArrow
  | EastArrow
  | WestArrow
  | Bordercolor of int * int * int
  | Multiple of representation list

let color_gradient ac t i =
  let n = Array.length t in
    if n = 0 || cardinality ac = 1 then Unicolor (0,0,0) else
      let where = ((n-1)*i)/(cardinality ac) in
      let i = ((n-1)*i - (where*(cardinality ac))) in
      let ((ar,av,ab),(br,bv,bb)) = (t.(where),t.((where+1) mod n)) in
      let vr = (i*(br-ar))/(cardinality ac-1) in 
      let vv = (i*(bv-av))/(cardinality ac-1) in 
      let vb = (i*(bb-ab))/(cardinality ac-1) in Unicolor (ar+vr,av+vv,ab+vb)

let rec translate = function 
  | Unicolor (a,b,c) -> Kernel.Uni (a,b,c)
  | Transparent -> Kernel.Nocolor
  | NorthArrow -> Kernel.Up
  | SouthArrow -> Kernel.Down
  | EastArrow -> Kernel.Right
  | WestArrow -> Kernel.Left
  | Bordercolor (a,b,c) -> Kernel.Arround (a,b,c)
  | Multiple x -> Kernel.Overdraw (List.map translate x)

let rec untranslate = function 
  | Kernel.Uni (a,b,c) -> Unicolor (a,b,c)
  | Kernel.Nocolor -> Transparent 
  | Kernel.Up -> NorthArrow 
  | Kernel.Down -> SouthArrow 
  | Kernel.Right -> EastArrow
  | Kernel.Left -> WestArrow
  | Kernel.Arround (a,b,c) -> Bordercolor (a,b,c)
  | Kernel.Overdraw x -> Multiple (List.map untranslate x)

let get_representation ac reader posx posy =
  let state = reader posx posy in
  untranslate (Kernel.colorofstate ac state reader posx posy)

let poly_get_representation ac ~alphabet reader posx posy =
  let reader a b = ios alphabet (reader a b) in
  get_representation ac reader posx posy

let add_colormap ac ?set name cm =
  let cmaps = ac.Kernel.parameters.Kernel.colormaps in
    (
      ac.Kernel.parameters.Kernel.colormaps <- 
	(name, (Kernel.General (fun i -> translate (cm i))))::
	(List.remove_assoc name cmaps);
      (match set with 
	   Some true -> ac.Kernel.colormap <- name
	 | _ -> ()
      )
    )

let poly_add_colormap ac ?set name ~alphabet cm =
  let state_of_int = soi alphabet in
    add_colormap ac ?set name (fun i -> cm (state_of_int i))

let add_view ac ?set name f =
  let dn = Kernel.current_domain_name () in
  let cmaps = ac.Kernel.parameters.Kernel.colormaps in
    (
      ac.Kernel.parameters.Kernel.colormaps <- 
	(name, (Kernel.Viewer (fun i j read -> Kernel.enter_domain_name dn; 
				 let res = translate (f i j read) in Kernel.exit_domain_name (); res
			      )))::(List.remove_assoc name cmaps);
      (match set with 
	   Some true -> ac.Kernel.colormap <- name
	 | _ -> ()
      )
    )

let poly_add_view ac ?set name ~alphabet f =
  let state_of_int = soi alphabet in
    add_view ac ?set name (fun i j read -> f i j (fun x y -> state_of_int (read x y)))


type 'a modifier =
    Forced of 'a
  | Map of ('a -> 'a)
  | GeneralZone of ('a -> (int * int * int * int) -> int -> int -> 'a)
  | SimpleZone of ('a -> int -> int -> 'a)
  | FreeZone of ((int * int * int * int) -> (int -> int -> 'a) -> (int -> int -> 'a -> unit)-> unit)

let do_the_add_stuff ac set name obj =
  let mods = List.remove_assoc name ac.Kernel.parameters.Kernel.actions in
    ac.Kernel.parameters.Kernel.actions <- if set then obj::mods else mods@[obj]
  

let add_force_modif ac name state set = do_the_add_stuff ac set name (name, Kernel.Force state)
      
let add_general_modif ac name funk set = do_the_add_stuff ac set name (name, Kernel.Funkmod funk)
      
let add_free_modif ac name funk set = do_the_add_stuff ac set name (name, Kernel.Freemod funk)
      
let add_modif ac ?key ?set ?left ?oriented name what = 
  let doset = (match set with 
       Some true -> true
    | _ -> false) in
  let oriented = (match oriented with
      Some true -> true
    | _ -> false) in
  let do_orient = if not oriented then (fun (ux,uy,vx,vy) -> 
    (min ux vx), (min uy vy), (max ux vx), (max uy vy) 
  ) else (fun x -> x) in
  (match what with
    Forced i -> add_force_modif ac name i doset
  | Map f -> let dn = Kernel.current_domain_name () in
	     add_general_modif ac name (fun c z x y -> Kernel.enter_domain_name dn; let res = f c in
										    Kernel.exit_domain_name (); res) doset
  | GeneralZone f -> let dn = Kernel.current_domain_name () in
		     add_general_modif ac name (fun c z x y -> Kernel.enter_domain_name dn; let res = f c (do_orient z) x y in
											    Kernel.exit_domain_name (); res) doset
  | SimpleZone f -> let dn = Kernel.current_domain_name () in
		    add_general_modif ac name (fun c z x y -> Kernel.enter_domain_name dn; let res = f c x y in
											   Kernel.exit_domain_name (); res) doset
  | FreeZone f -> let dn = Kernel.current_domain_name () in
		  add_free_modif ac name (fun z re wr -> Kernel.enter_domain_name dn; f (do_orient z) re wr;
		    Kernel.exit_domain_name ()) doset
  );
  (match key with
    Some k -> Kernel.setaction ac (Kernel.SrcKeyClick k) name
  | None -> ()
  );
  if doset then 
    (match left with
	Some true -> Kernel.setaction ac Kernel.SrcLeftclick name
      | Some false -> Kernel.setaction ac Kernel.SrcRightclick name
      | _ -> (
	Kernel.setaction ac Kernel.SrcLeftclick name;
	Kernel.setaction ac Kernel.SrcRightclick name
      )
    )

let poly_add_modif ac ?key ?set ?left ?oriented name ~alphabet what = 
  let ios = ios alphabet and soi = soi alphabet in
  let translate_to_int = function 
      Forced x -> Forced (ios x)
    | Map f -> Map (fun i -> ios (f (soi i)))
    | GeneralZone f -> GeneralZone (fun s z x y -> ios (f (soi s) z x y))
    | SimpleZone f -> SimpleZone (fun s x y -> ios (f (soi s) x y))
    | FreeZone f -> FreeZone (fun z read write -> f z
				(fun a b -> soi (read a b))
				(fun a b s -> write a b (ios s))
			     )
  in
    add_modif ac ?key ?set ?left ?oriented name (translate_to_int what)


let set_universe_size a x y = a.Kernel.universe <- (x,y)

let get_universe_size a = a.Kernel.universe

type 'a initconf =
  General of (int -> int -> 'a)
| GenArray of (unit -> 'a array array) 
| Bernoulli of ('a -> float)
| RandomBernoulli of 'a array
| RandomUniform
| Uniform of 'a
| Mask of (int -> int -> bool) * ('a initconf)
| FixedSize of int * int * ('a initconf)
| MultiLayer of ('a initconf) list
    
let add_init_conf ac ?set name ic = 
  let rec myconvert = function
      RandomBernoulli x -> Kernel.Random (Kernel.Bernouilli x)
    | Uniform x -> Kernel.Monochrome x
    | RandomUniform -> Kernel.Random (Kernel.Uniform)
    | General f -> Kernel.Custom f
    | Bernoulli p ->
       let n = cardinality ac in
       let rec lebon alpha acc i =
	 if i = n-1 then n-1
	 else if alpha <= acc then i
	 else lebon alpha (acc +. (p (i+1))) (i+1)
       in
       Kernel.Custom (fun x y ->
	 let alpha = conf_random_float 1. in
	 lebon alpha (p 0) 0
       )
    | GenArray f -> Kernel.CustomArray f
    | Mask (mfun, conf) -> Kernel.Masked (mfun, myconvert conf)
    | FixedSize (w,h, conf) -> Kernel.Sized (w,h, myconvert conf)
    | MultiLayer lst -> Kernel.Layers (List.map myconvert lst)
  in
    ac.Kernel.initconfs <- (name, (myconvert ic))::(List.remove_assoc name ac.Kernel.initconfs);
    if initial ac || (Kernel.plugin_mode ()) then (match set with
	 Some true -> ac.Kernel.theinitconf <- name
       | _ -> ()
    )
      
  
let active_only ac activefun =
  let rec extract_rule = function
      Kernel.Sparse (s,r) -> r
    | Kernel.MultisetFunArray _ | Kernel.MultisetTable _ -> raise (Kernel.Unimplemented "Sparse multiset rule")
    | r -> r
  in
  let rec acl acc n =
    if n = -1 then acc
    else if activefun n then acl (Kernel.ActivSet.add n acc) (n-1) else acl acc (n-1)
  in 
    ac.Kernel.transitionrule <- Kernel.Sparse ((acl Kernel.ActivSet.empty (cardinality ac-1)),
					extract_rule ac.Kernel.transitionrule
				       )

let poly_set_rule ac ?active ?(printer = (function _ -> "(from rule)")) ~alphabet local =
  let n = cardinal alphabet in
  let state_of_int = soi alphabet in
  let int_of_state = ios ~printer alphabet in
  let new_local x = int_of_state (local (Array.map state_of_int x)) in
    (
      set_cardinality ac n;
      set_rule ac new_local;
      match active with
	  Some actfun -> active_only ac (fun x -> actfun (state_of_int x))
	| _ -> ()
    )

let poly_set_rule_fun ac ?active ?(printer = (function _ -> "(from rule)")) ~alphabet local =
  let n = cardinal alphabet in
  let state_of_int = soi alphabet in
  let int_of_state = ios ~printer alphabet in
  let new_local reader = int_of_state (local (fun x y -> state_of_int (reader x y))) in
    (
      set_cardinality ac n;
      set_rule_fun ac new_local;
      match active with
	  Some actfun -> active_only ac (fun x -> actfun (state_of_int x))
	| _ -> ()
    )


let poly_set_multiset_rule ac ?active ?(printer = (function _ -> "(from rule)")) ~alphabet local =
  let n = cardinal alphabet in
  let state_of_int = soi alphabet in
  let int_of_state = ios ~printer alphabet in
  let new_local c x = int_of_state (local (state_of_int c) (fun sta -> x.(int_of_state sta))) in
    (
      set_cardinality ac n;
      set_multiset_rule ac new_local;
      match active with
	  Some actfun -> active_only ac (fun x -> actfun (state_of_int x))
	| _ -> ()
    )
	
let poly_add_init_conf ac ?set name ~alphabet ic = 
  let ios = ios alphabet in
  let rec myconvert = function
      RandomBernoulli x -> Kernel.Random (Kernel.Bernouilli (Array.map ios x))
    | Uniform x -> Kernel.Monochrome (ios x)
    | RandomUniform -> Kernel.Random (Kernel.Uniform)
    | General f -> Kernel.Custom (fun x y -> ios (f x y))
    | Bernoulli p ->
       let n = cardinal alphabet in
       let rec lebon alpha acc i =
	 if i = n-1 then n-1
	 else if alpha < acc then i
	 else lebon alpha (acc +. (p (soi alphabet i))) (i+1)
       in
       Kernel.Custom (fun x y ->
	 let alpha = conf_random_float 1. in
	 lebon alpha (p (soi alphabet 0)) 0
       )
    | GenArray f -> 
      let newgen () = let a = f () in Array.map (Array.map ios) a in
      Kernel.CustomArray newgen
    | FixedSize (w,h, conf) -> Kernel.Sized (w,h, myconvert conf)
    | Mask (mfun, conf) -> Kernel.Masked (mfun, myconvert conf)
    | MultiLayer lst -> Kernel.Layers (List.map myconvert lst)
  in
    ac.Kernel.initconfs <- (name, (myconvert ic))::(List.remove_assoc name ac.Kernel.initconfs);
    if initial ac then (match set with
	 Some true -> ac.Kernel.theinitconf <- name
       | _ -> ()
    )

let add_shortcut = Kernel.add_shortcut

type gui_t = Interactive.api_gui_element
type draw_context = Interactive.api_draw_context


let gui_window ?(height=200) ?(width=200) title = Interactive.api_window title width height
let gui_vbox = Interactive.api_vbox
let gui_hbox = Interactive.api_hbox
let gui_textbox s = fst (Interactive.api_textbuffer s)
let gui_textbuffer = Interactive.api_textbuffer
let gui_button = Interactive.api_button
let gui_label = Interactive.api_label
let gui_full_label = Interactive.api_full_label
let gui_drawing_area ?(height=200) ?(width=200) ?(background=(0,0,0)) () = Interactive.api_draw_area height width background
let size dc = dc.Interactive.dc_size ()
let point dc ?(color=(250,250,250)) x y = dc.Interactive.dc_point color x y
let clear dc = dc.Interactive.dc_clear ()
let rectangle dc ?(color=(250,250,250)) x = dc.Interactive.dc_rectangle color x

let add_snapshot ac ?set name snap_method = 
  let dn = Kernel.current_domain_name () in
  let newsnap x y fonc = Kernel.enter_domain_name dn; snap_method x y fonc;
    Kernel.exit_domain_name ()
  in
    Kernel.addaction ac name (Kernel.Snapshot newsnap);
    (match set with 
	 Some true -> Kernel.setaction ac Kernel.SrcSnapshot name
       | _ -> ()
    )

let poly_add_snapshot ac ?set name ~alphabet snap_method = 
  let soi = soi alphabet in
    add_snapshot ac ?set name (fun x y read -> snap_method x y (fun a b -> soi (read a b)))

let add_measurement ac ?(trigger=false) ?(set=true) name measure_method = 
  let dn = Kernel.current_domain_name () in
  let newmeasure t x y rfonc wfonc = Kernel.enter_domain_name dn; measure_method t x y rfonc wfonc;
    Kernel.exit_domain_name ()
  in
    Kernel.addaction ac name (Kernel.Measurement 
				((if trigger then Kernel.MeTrigger else Kernel.MeOnOff (ref set)), newmeasure));
    Kernel.setanotheraction ac Kernel.SrcMeasurement name

let poly_add_measurement ac ?(trigger=false) ?(set=true) name ~alphabet measure_method = 
  let soi = soi alphabet and ios = ios alphabet in
    add_measurement ac name ~trigger ~set (fun t x y read write -> measure_method t x y 
				    (fun a b -> soi (read a b))
				    (fun a b s -> write a b (ios s))
				 )


type execution = 
    EachStep
  | EachNewObject_after
  | EachNewObject_before
  | ProgramStartup
  | UserDemand

let exec_to_kernel_type = function
    EachNewObject_after -> Kernel.Apreschaqueobjet
  | EachNewObject_before -> Kernel.Avantchaqueobjet
  | EachStep -> Kernel.Apreschaqueetape
  | UserDemand -> Kernel.Alademande
  | _ -> Kernel.Apresinitialisation

let valid_plugin ?name ?description ?activate ?executed_at ?start ?finish maker_fun = match executed_at with
    None -> Kernel.valid_plugin ?activate ?name ?description ?start ?finish ~repetition:(exec_to_kernel_type ProgramStartup) maker_fun
  | Some x -> Kernel.valid_plugin ?activate ?name ?description ?start ?finish ~repetition:(exec_to_kernel_type x) maker_fun

type regenerate =
    Automaton
  | Configuration
  | Both
  | Nothing
  | Skip

let valid_hook_plugin ?activate ?name ~pattern ?description ?start ?finish  hook = 
  let reg_typ = function
    | Automaton -> (true,true,false)
    | Configuration -> (true,false,true)
    | Both -> (true,true,true)
    | Nothing -> (true,false,false)
    | Skip -> (false,false,false)
  in 
  let new_hook s ac (w,h) reader writer = 
    let reg,s_list = hook s ac (w,h) reader writer in
      ((reg_typ reg), s_list)
  in
    Kernel.valid_hook_plugin ?name ?activate ?description ?start ?finish ~pattern new_hook

let kernel_cast a = a

let get_stop_time () = !Kernel.stoptime

let set_stop_time t = Kernel.stoptime := t

let add_description a title text = 
  let list_rep = ["&","&amp;" ; ">","&gt;" ; "<","&lt;"] in
  let clean_txt x = List.fold_left (fun acc x -> Str.global_replace (Str.regexp (fst x)) (snd x) acc) x list_rep in
    a.Kernel.description <- ((clean_txt title),(clean_txt text))::a.Kernel.description

let get_views a = List.map fst a.Kernel.parameters.Kernel.colormaps
let set_view a quoi = a.Kernel.colormap <- quoi
let get_view a = a.Kernel.colormap

type action = Save | Shoot | New | Stop

let convert_action = function 
  | Save -> Keymap.Saverule
  | Shoot -> Keymap.Screenshot
  | New -> Keymap.Newboth
  | Stop -> Keymap.Stop

let clear_scenario = Keymap.clear_scenario
let addto_scenario x = Keymap.addto_scenario (convert_action x)

