(* compilation_options: -package cairo,sdl,sdl.sdlimage *)

open Acml


let color_of_state n state =
  if n < 256 then 
    let gray_val = state * (255/n) in
      (Int32.add (Int32.shift_left (Int32.of_int gray_val) 16) (Int32.of_int  (gray_val lsl 8 + gray_val)))
  else Int32.of_int state

let state_of_rgb n (r,g,b) =
  let tmp = 
    if n < 256 then r / (255/n)
    else (r lsl 16) + (g lsl 8) + b
  in tmp mod n

let write_png_grayscale filename reader n w h =
  let ba = Bigarray.Array2.create Bigarray.int32 Bigarray.c_layout h w in
    for x=0 to w-1 do
      for y=0 to h-1 do
	ba.{h-1-y,x} <- color_of_state n (reader x y)
      done;
    done;
    Cairo_png.surface_write_to_file (Cairo_bigarray.of_bigarr_32 ~alpha:false ba) filename

let write_ocaml_array filename reader w h =
  let file = open_out filename in
  Printf.fprintf file "let configuration = [|\n";
  for y=0 to h-1 do
    Printf.fprintf file "[| ";
    for x=0 to w-1 do
      Printf.fprintf file "%d; " (reader x y)
    done;
    Printf.fprintf file "|];\n";
  done;
  Printf.fprintf file "|]\n";
  close_out file

let fresh_file_name base ext =
  let ithname i = base^"_"^(string_of_int i)^ext in
  let i = ref 0 in 
    while Sys.file_exists (ithname !i) do
      incr i
    done;
    ithname !i

let maker a=
  let n = cardinality a in
  let d = dimension a in
    add_modif a "save as PNG" (FreeZone
      (fun (ax,ay,bx,by) read write ->
	 let (w,h) = (bx-ax+1),(by-ay+1) in
	 let reader a b = read (ax+a) (ay+b) in
	 let name = fresh_file_name "region" ".png" in
	   write_png_grayscale name reader n w h;
	   infomsg ("File '"^name^"' successfully written.")
      ));
    add_modif a "save as OCaml array" (FreeZone
      (fun (ax,ay,bx,by) read write ->
	 let (w,h) = (bx-ax+1),(by-ay+1) in
	 let reader a b = read (ax+a) (ay+b) in
	 let name = fresh_file_name "region" ".ml" in
	   write_ocaml_array name reader w h;
	   infomsg ("File '"^name^"' successfully written.")
      ));
    if d=1 then (
      add_snapshot a "1D-configuration as PNG"
	(fun w h reader ->
	  let name = fresh_file_name "configuration" ".png" in
	  write_png_grayscale name reader n w 1;
	  infomsg ("File '"^name^"' successfully written.")
	);
      add_snapshot a "1D-configuration as OCaml array"
	(fun w h reader ->
	  let name = fresh_file_name "configuration" ".ml" in
	  write_ocaml_array name reader w 1;
	  infomsg ("File '"^name^"' successfully written.")
	);
      add_snapshot a "2D-space-time as PNG"
	(fun w h reader ->
	  let name = fresh_file_name "configuration" ".png" in
	  write_png_grayscale name reader n w h;
	  infomsg ("File '"^name^"' successfully written.")
	);
      add_snapshot a "2D-space-time as OCaml array"
	(fun w h reader ->
	  let name = fresh_file_name "configuration" ".ml" in
	  write_ocaml_array name reader w h;
	  infomsg ("File '"^name^"' successfully written.")
	)
    ) else (
      add_snapshot a "Configuration as PNG"
	(fun w h reader ->
	  let name = fresh_file_name "configuration" ".png" in
	  write_png_grayscale name reader n w h;
	  infomsg ("File '"^name^"' successfully written.")
	);
      add_snapshot a "Configuration as OCaml array"
	(fun w h reader ->
	  let name = fresh_file_name "configuration" ".ml" in
	  write_ocaml_array name reader w h;
	  infomsg ("File '"^name^"' successfully written.")
	)
    )


let _ = valid_plugin ~executed_at:EachNewObject_after ~name:"PNG configuration/region saver" ~activate:true ~description:"It adds actions to save a configuration or a region as a PNG image or OCaml array source code."  maker

let _ = declare_file_type "PNG/JPEG files" ["png"; "PNG"; "jpg"; "jpeg"; "JPG"; "JPEG"]
    
let _ = valid_hook_plugin ~activate:true ~pattern:"\\(.*\\.[pP][nN][gG]$\\)\\|\\(.*\\.[jJ][pP][gG]$\\)\\|\\(.*\\.[jJ][pP][eE][gG]$\\)" ~name:"PNG/JPEG conf. loader"
  (fun str ac (w,h) read write ->
     try
       let surf = Sdlloader.load_image str in
       let (w,h,_) = Sdlvideo.surface_dims surf in
       let n = cardinality ac in
       let (_,oldh) = get_universe_size ac in
       let nh = if dimension ac = 1 && (not (initial ac)) then oldh else h in
       add_init_conf ac ~set:true (Filename.basename str)
	 (FixedSize (w,nh,(General (fun a b -> state_of_rgb n (Sdlvideo.get_pixel_color surf a (nh-b-1))))));
       (Configuration,[])
     with _ -> (infomsg ("Problem in loading file '"^str^"'. Aborted."); (Nothing,[]))
  )

exception Pb

let matrix_from_rle_file str =
  try
    let file = 
      try 
	open_in str 
      with Sys_error err -> (infomsg ("File '"^str^"' can not be opened: "^err); raise Pb)
    in
      (* try to be compatible with DOS end-of-line convention  *)
    let header = Str.regexp "[ \t]*x[ \t]*=[ \t]*\\([0-9]+\\)[ \t]*,[ \t]*y[ \t]*=[ \t]*\\([0-9]+\\)[ \t]*\\(,[ \t]*rule[ \t]*=[ \t]*\\([^\r]*\\)\\)?\r?$" in
    let comment = Str.regexp "#[cCdDoO]? \\(.*\\)\r?$" in
    let skip = Str.regexp "\\(#.*$\\)\\|\\([ \r\t]*$\\)" in
    let data = Str.regexp "[ \t]*\\([0-9]*\\)\\([bo.A]\\|\\([p-y]?[A-X]\\)\\|\\$\\)[ \t]*" in
    let theend = Str.regexp "!.*" in
(*     let strange_newline = Str.regexp "[\r]" in *)
(*     let line_acc = ref [] in *)
(*     let input_line file = match !line_acc with *)
(*       | [] -> let line = input_line file in *)
(* 	      let lst = Str.split strange_newline *)
    let rec scan_line s n acc =
      if n>= String.length s then acc else
	try
	  if Str.string_match theend s n then acc else
	    let _ = Str.search_forward data s n in
	    let howmany = Str.matched_group 1 s in
	    let ofwhat = Str.matched_group 2 s in
	      scan_line s (Str.match_end ()) ((howmany,ofwhat)::acc)
	with Not_found -> acc
    in
    let rec go_scan acc_comments acc = try
      let s = input_line file in
	if Str.string_match comment s 0 then go_scan ((Str.matched_group 1 s)::acc_comments) acc
	else if Str.string_match skip s 0 then go_scan acc_comments acc
	else go_scan acc_comments (scan_line s 0 acc)
    with End_of_file -> ((List.rev acc), (List.rev acc_comments))
    in 
    let rec scan_header acc_comments = 
      let s = input_line file in
	if Str.string_match comment s 0 then scan_header ((Str.matched_group 1 s)::acc_comments)
	else if Str.string_match skip s 0 then scan_header acc_comments
	else if Str.string_match header s 0 then ((int_of_string (Str.matched_group 1 s)),
						  (int_of_string (Str.matched_group 2 s)),
						  (try Some (Str.matched_group 4 s) with Not_found -> None),
						  acc_comments
						 )
	else (infomsg ("Header not found in file '"^str^"'!"); raise Pb)
    in
    let (w,h,rule_spec,pre_comments) = scan_header [] in
    let data,comments = go_scan pre_comments [] in
    let real_w = ref 0 in
    let cur_w = ref 0 in
    let real_h = ref 1 in
    let char_index refchar candidate = 1 + (Char.code candidate) - (Char.code refchar)  in
    let state_of_string = function 
      | "." | "b" -> 0
      | "o" | "A" -> 1
      | x -> if String.length x = 1 then char_index 'A' x.[0] else 24*(char_index 'p' x.[0])+(char_index 'A' x.[1])
    in
    let clean_data = List.map (fun (n,o) -> ((if n="" then 1 else int_of_string n), o)) data
    in (
	List.iter (fun (n,o) -> if o="$" then (real_h:= !real_h+n; cur_w:= 0) else (cur_w:= !cur_w+n; real_w := max !real_w !cur_w)) clean_data;
	let inv_h = !real_h in
	let m = Array.make_matrix !real_w !real_h 0 in
	let rec spit howmany what = if howmany > 0 then (m.(!cur_w).(inv_h-1 - !real_h) <- what; incr cur_w; spit (howmany-1) what)
	in
	  cur_w:= 0; real_h:= 0;
	  List.iter (fun (n,o) -> 
		       if o="$" then (real_h:= !real_h+n; cur_w:= 0) else spit n (state_of_string o)) clean_data;	
	  incr real_h;
	  (m, !real_w, !real_h, w,h,rule_spec,comments)
      )
  with x -> raise x 

   
let _ = declare_file_type ~transform:(fun x -> "region:"^x) "RLE as region" ["rle";"RLE"]
let _ = declare_file_type "RLE as configuration/rule" ["rle";"RLE"]
    

let _ = valid_hook_plugin ~activate:true ~pattern:"\\(region:\\)?.*[rR][lL][eE]$" ~name:"RLE loader"
  ~description:"Loads an RLE file as a configuration and eventually changes the rule as specified in the file.\n If the prefix \"region:\" is added to the filename, neither the configuration nor the rule is changed, but an action is added that can put the configuration content of the file wherever the user asks using the mouse."
  (fun str ac (w,h) read write ->
     if (String.length str >= 7) && (String.sub str 0 7 = "region:") then (
       let filename = String.sub str 7 (String.length str - 7) in
       let (m, real_w, real_h, w,h,rule_spec,comments) = matrix_from_rle_file filename in
	 if List.length comments > 0 then add_description ac (Filename.basename str) (String.concat "\n" comments);
	 add_modif ac ~set:true str 
	   (FreeZone (fun (xa,ya,xb,yb) read write -> 
			for i=0 to xb-xa do
			  for j=0 to yb-ya do
			    write (xa+i) (ya+j) (if i < real_w && j < real_h then m.(i).(j) else 0)
			  done
			done
		     ));
	 (Nothing,[])
     ) else (
       let (m, real_w, real_h, w,h,rule_spec,comments) = matrix_from_rle_file str in
       let (oldw,oldh) = get_universe_size ac in
	 if List.length comments > 0 then add_description ac (Filename.basename str) (String.concat "\n" comments);
	 add_init_conf ac ~set:true (Filename.basename str)
	   (if oldw > w && oldh > h then 
	      let sx,sy = ((oldw-w)/2),((oldh-h)/2) in 
		(General (fun a b -> if a < real_w+sx && b < real_h+sy && a>=sx && b>= sy then m.(a-sx).(b-sy) else 0))
	    else FixedSize (w, h, (General (fun a b -> if a < real_w && b < real_h then m.(a).(b) else 0)))
	   );
	 (Configuration,(match rule_spec with 
			     (* Heuristic to mimic Golly on .table extension *)
			   | Some r -> (
			       let nks = Kernel.get_all_nicks () in
				 if List.mem r nks then [r] 
				 else (
				   try Kernel.try_shortcuts r; 
				     (* Golly uses a modified rule name to speak about the universe...  *)
				     [(Filename.concat (Filename.dirname str)
					 (if Str.string_match (Str.regexp "^\\([^:]*\\):.*$") r 0 
				       then Str.matched_group 1 r 
				       else r))^".table"] 
				   with Kernel.Shortcut _ -> [r]
				 )
			     )
			   | None -> []))
     )
  )

