let family_name = "Random cyclic automata"
let description = "States are arranged in a cycle. A cell can either update its state to the next in the cycle or stay unchanged."

open Acml;;
let objects = [
"cyclic1D", ["dimension", "1"; "states", "4"; "radius", "4"; "threshold", "1"; "multiset", "false"; "instability", "0.45";];
"cyclicmultiset1D", ["dimension", "1"; "states", "5"; "radius", "1"; "threshold", "0"; "multiset", "true"; "instability", "0.5";];
"cyclic2D", ["dimension", "2"; "states", "5"; "radius", "1"; "threshold", "2"; "multiset", "true"; "instability", "0.7"; "neighborhood", "Moore"; "forward jump", "2"];
]

let declare = 
  function name, params -> (
    add_shortcut ("^"^name^"$") (fun _ -> params);
    declare_object name
  )


let _ = (
  add_shortcut "^[Rr]andom[- ]?[Cc]yclic$" (fun _ -> []);
  List.iter declare objects
)

let maker a = 
  (
    let (d,n,r,neig) = std_params a ~states:3 in
    let instability = float_param a "instability" 0. 1. 0.6 in
    let jump = param a "forward jump" 1 n 1 in
    let minnext = param a "threshold" 0 (arity a) 2 in
    ensure_determinism a;
    if bool_param a "multiset" true then 
      set_multiset_rule a (fun cell howmany ->
	let next = (cell + jump) mod n in
	if howmany.(next) < minnext then cell
	else if ca_random_float 1. <= instability then next else cell
      ) 
    else set_rule a (fun x -> 
      let next = (x.(centercell a) + jump) mod n in
      let rec find howmany p =
	if howmany = 0 then true 
	else if p < 0 then false
	else if x.(p) = next && howmany = 1 then true 
	else if x.(p) = next then find (howmany-1) (p-1)
	else find howmany (p-1)
      in 
      if not (find minnext (Array.length x - 1)) then x.(centercell a)
      else if ca_random_float 1. <= instability then next else x.(centercell a)
    );
    set_name a family_name;
    add_description a "Rule family" description;
    doc_param a "instability" "The probability that a transition which statisfies constraints given by other parameters do actually update the cell to the next state";
    doc_param a "threshold" "How many occurrences of next state in the neighbourhood are required to allow update";
    doc_param a "forward jump" "k such that a cell in state i tries to update to (i+k) mod N";
    doc_param a "multiset" "Whether the rule is invariant by any permutation of the neighborhood";
  )

let _ =  valid maker
