let family_name = "Table format"

open Acml

exception Pb

type table_item =
    Varuse of string
  | Value of int

type table_exec =
    Vardef of string * (table_item list)
  | Transition of table_item list

let scan_table input_line =
    (* regexp made compatible with DOS newline convention *)
  let header_states = Str.regexp "^[ \t]*n_states:[ \t]*\\([0-9]+\\)" in
  let header_neighb = Str.regexp "^[ \t]*neighborhood:[ \t]*\\([^# \r]+\\)\r?" in
  let header_sym = Str.regexp "^[ \t]*symmetries:[ \t]*\\([^# \r]+\\)\r?" in
  let variable = Str.regexp "^[ \t]*var \\([a-zA-Z][a-zA-Z0-9]*\\)[ \t]*=[ \t]*{\\([^}]+\\)}" in
  let variable_single = Str.regexp "^[ \t]*var \\([a-zA-Z][a-zA-Z0-9]*\\)[ \t]*=[ \t]*\\([a-zA-Z0-9]+\\)" in
  let transition_compact = Str.regexp "^\\([0-9]+\\)[ \t]*\\(#.*\\)?\r?$" in
  let transition = Str.regexp "^\\(\\(\\(\\([a-zA-Z][a-zA-Z0-9]*\\)\\|\\([0-9]+\\)\\),\\)*\\(\\([a-zA-Z][a-zA-Z0-9]*\\)\\|\\([0-9]+\\)\\)\\)[ \t]*\\(#.*\\)?\r?$" in
  let comment = Str.regexp "^\\([ \t]*#.*\\)" in
  let int_or_string i = try Value (int_of_string i) with _ -> Varuse i in
  let il_of_string s = List.map int_or_string (Str.split (Str.regexp " *, *") s) in
  let decompact s = Transition (List.map (fun x -> Value (int_of_string x)) (Str.split (Str.regexp "") s)) in
  let linepos = ref 0 in
  let rec scan_file states neighb sym acc  = 
    try 
      let s = input_line () in
	incr linepos;
	if Str.string_match comment s 0 or String.length s <= 1 then scan_file states neighb sym acc
	else if Str.string_match header_states s 0 then scan_file (int_of_string (Str.matched_group 1 s)) neighb sym acc
	else if Str.string_match header_neighb s 0 then scan_file states (Str.matched_group 1 s) sym acc
	else if Str.string_match header_sym s 0 then scan_file states neighb (Str.matched_group 1 s) acc
	else if Str.string_match variable s 0 then 
	  let v = Str.matched_group 1 s and l = Str.matched_group 2 s in
	    scan_file states neighb sym ((!linepos, (Vardef (v, (il_of_string l))))::acc)
	else if Str.string_match variable_single s 0 then 
	  let v = Str.matched_group 1 s and l = Str.matched_group 2 s in
	    scan_file states neighb sym ((!linepos, (Vardef (v, [int_or_string l])))::acc)
	else if Str.string_match transition_compact s 0 then scan_file states neighb sym 
	  ((!linepos, (decompact (Str.matched_group 1 s)))::acc)
	else if Str.string_match transition s 0 then scan_file states neighb sym 
	  ((!linepos, (Transition (il_of_string (Str.matched_group 1 s))))::acc)
	else (infomsg (Printf.sprintf "Syntax error in line %d" !linepos); raise Pb)
    with End_of_file -> (states, neighb, sym, List.rev acc)
  in
    (scan_file 1 "moore" "none" [])
    
exception Analysis_error of string
exception No_more_choice

module IS = Set.Make (struct 
  type t=int    
  let compare = compare
end)

module SM = Map.Make (struct 
  type t=string
  let compare = compare
end)

let set_from_list =
  let rec doit s = function
    | [] -> s
    | t::q -> doit (IS.add t s) q
  in doit IS.empty 


let next_choice k test_val =
  let rec doit i =
    if i=k then raise No_more_choice
    else let (vref,vec) = test_val.(i) in (
      if !vref = Array.length vec - 1 then (vref:=0; doit (i+1))
      else (incr vref)
    ) in
      doit 0 

let fun_from_table_def p_trans k env fixed_pos sym_readers trans_patterns centerof =
  let compatible env a x = IS.mem x (SM.find a env) in
  let easy_trans_match prep_trans  x =
    let everything = snd prep_trans in
    let pre,finalize = fst everything in
    let pattern = snd everything in
    let rec doit sym_read i s_list = 
      if i=k then Some 0(* (if fst finalize then snd finalize else x.(sym_read (snd finalize))) *)
      else if not (IS.mem x.(sym_read i) pattern.(i)) then iter_syms s_list else doit sym_read (i+1) s_list
    and iter_syms = function
      | [] -> None
      | t::q -> if pre (fun i-> x.(t i)) then doit t 0 q else iter_syms q
    in
      iter_syms sym_readers
  in
  let read = List.hd sym_readers in
  let rec trans_fixed_match env pattern i x =
    if i=k then Some env 
    else if fixed_pos i then (match pattern.(i)  with 
				  Value v -> if v = x.(read i) then trans_fixed_match env pattern (i+1) x else None
				| Varuse a -> if compatible env a x.(read i) then trans_fixed_match (SM.add a (IS.add x.(read i) IS.empty) env) pattern (i+1) x else None)
    else trans_fixed_match env pattern (i+1) x
  in
  let trans_sym_match prep_trans env pattern line x =
    let test_val = fst prep_trans in
    let everything = snd prep_trans in
    let pre,finalize = fst everything in
    let rec is_good sym i =
      if i=k then true
      else if not (fixed_pos i) then (
	let p,possible = test_val.(i) in
	  if x.(sym i) = possible.(!p) then is_good sym (i+1)
	  else false
      ) else is_good sym (i+1) in
    let rec try_each = function
      | [] -> None
      | sym::q -> if is_good sym 0 then Some (if fst finalize then snd finalize else x.(sym (snd finalize)))
	else try_each q
    in try_each sym_readers  in
  let rec try_choices n env q u line x = 
    let test_val = fst p_trans.(n) in
      (match trans_sym_match p_trans.(n) env u line x with
	 | None -> (
	     if (try
		   next_choice k test_val;
		   true
		 with No_more_choice -> false)
	     then try_choices n env q u line x else compute_transition (n+1) q x)
	 | Some v -> v)
  and compute_transition n l x = match l with
    | [] -> centerof x
    | (line,u)::q -> 
	(match u.(k) with
	     (* if no variable in output then use easy matching *)
	   | Value v -> (match easy_trans_match p_trans.(n) x with
			   | Some _ -> v
			   | None -> compute_transition (n+1) q x
			)
	   | _ -> (* variable in output so use slow matching  *)
	       (* first 'pattern match' the fixed positions *) 
	       (match trans_fixed_match env u 0 x with
		  | None -> compute_transition (n+1) q x
		  | Some env -> (* fixed positions ok, variables bound if any *)
		      (* reset choice array *)
		      Array.iter (fun (r,_) -> r:=0) (fst p_trans.(n));
		      (* loop on symmetries for each choice of variables values *) 
		      try_choices n env q u line x
	       )
	)
  in compute_transition 0 trans_patterns

(* DEBUG stats for cache *)
let cnt = ref 0
let deb_cache () = (
  incr cnt;
  infomsg ("[cache] "^(string_of_int !cnt))
)

(* heuristic cache mechanism *)
let cache f =
  let size = 1 lsl 16 in
  let nb = 2 in
  let harray = Array.init nb (fun _ -> Hashtbl.create size) in
  let usage = ref 0 in
  let pos = ref 0 in
  let _ = add_mem_info (nb*size*Kernel.bytes_per_word) in
  let rec find i x =
    try 
      Hashtbl.find harray.(i) x
    with Not_found -> (
      if i = ((nb + !pos - 1) mod nb) then (
	let r = f x in
	incr usage;
(* 	deb_cache (); *)
	Hashtbl.add harray.(!pos) (Array.copy x) r;
	if !usage >= size then (
	  pos := (!pos + 1) mod nb;
	  Hashtbl.clear harray.(!pos);
	  usage:= 0;
	);
	r
      ) else find ((i+1) mod nb) x
    )
  in (fun x -> find !pos x)

      
let rec env_trans env trans = function
    [] -> (List.rev env, List.rev trans)
  | (line, (Vardef (x,l)))::q -> env_trans ((x,l)::env) trans q
  | (line, (Transition l))::q -> env_trans env ((line,(Array.of_list l))::trans) q

let rec check_length length_check = function
  | [] -> ()
  | (line,a)::q -> 
    if Array.length a <> length_check then raise (Analysis_error (Printf.sprintf "bad transition length for declared neighborhood on line %d (got %d / expected %d)" 
								   line (Array.length a) length_check ))
    else check_length length_check q

let _ = declare_file_type "Table rule format" ["table"]


type multiset_variable =
    { mutable multiplicity : int;
      possibles : int array;
      mutable current : int;
    }

let mk_mv m pl =
  {multiplicity = m; possibles = Array.of_list pl; current =0}

(* translate a transition in a compact and more efficient form for multiset rules *)

let prepare_multiset_trans n env envi (line,trans) =
  let k = Array.length trans in
    (* initial will contain the multiset of states formed by constant values in the transition, but excluding the central cell *)
  let initial = Array.make n 0 in
    (* result is sorted in increasing order *)
  let find_all a = List.assoc a envi in
  let rec doit central acc_list var_ref i =
    if i = 0 then   (* we exclude central cell from the multiset choices since it is fixed during the choice loop process *)
      match trans.(i) with
	| Value v -> doit (IS.add v IS.empty) acc_list var_ref (i+1)
	| Varuse a -> let central = 
	    try 
	      List.assoc a env
	    with Not_found -> raise (Analysis_error ("undefined variable '"^a^"' on line "^(string_of_int line))) in
	    doit central acc_list var_ref (i+1)
    else if i = k-1 then match trans.(i) with
      | Value v -> (initial, central, None, v, Array.of_list (List.rev acc_list))
      | Varuse a -> let p = try
	  List.assoc a var_ref
	with Not_found -> raise (Analysis_error ("variable '"^a^"' is not bound but used as transition result on line "^(string_of_int line))) in
	  (initial, central,(Some p), 0, Array.of_list (List.rev acc_list))
    else match trans.(i) with
      | Value v -> (initial.(v) <- initial.(v)+1; doit central acc_list var_ref (i+1))
      | Varuse a -> 
	  if List.exists (fun y -> fst y = a) var_ref then let p = List.assoc a var_ref in (
	    p.multiplicity <- p.multiplicity + 1;
	    doit central acc_list var_ref (i+1)
	  ) else let vr = mk_mv 1 (find_all a) in
	    doit central (vr::acc_list) ((a,vr)::var_ref) (i+1)
  in doit IS.empty [] [] 0


(* translate a transition in a compact and more efficient form for general rules *)
let prepare_trans fixed_pos env envi (line,trans) =
  let k = Array.length trans in
  let find_all a = List.assoc a envi in
  let prepared_choices =
    let test_val = Array.make k ((ref 0), (Array.make 1 0)) in
    let rec choices_vec var_pos i =
      if i < k-1 then (
	if (fixed_pos i) then choices_vec var_pos (i+1) 
	else (* if not (fixed_pos i) then  *)
	  (match trans.(i) with
	       Value v -> (test_val.(i) <- ((ref 0), (Array.make 1 v)); choices_vec var_pos (i+1))
	     | Varuse a -> (test_val.(i) <- ((ref 0), (Array.of_list (find_all a))); choices_vec ((a,i)::var_pos) (i+1))
	  )
      ) in
      (choices_vec [] 0; test_val) in
  let rec identities i var_pos acc_fun =
    if i= k-1 then match trans.(i) with
      | Value v -> acc_fun,(true,v)
      | Varuse a -> 
	  let p = try
	    List.assoc a var_pos
	  with Not_found -> raise (Analysis_error ("variable '"^a^"' is not bound but used as transition result on line "^(string_of_int line)))
	  in acc_fun,(false,p)
    else match trans.(i) with
      | Varuse a -> 
	  if List.exists (fun y -> fst y = a) var_pos then let p = List.assoc a var_pos in 
	    identities (i+1) var_pos (fun reader -> reader i = reader p && acc_fun reader)
	  else identities (i+1) ((a,i)::var_pos) acc_fun
      | _ -> identities (i+1) var_pos acc_fun
  in
    (prepared_choices),((identities 0 [] (fun _ -> true)),
			(Array.map (function Value v -> IS.add v IS.empty | Varuse a -> 
				      try 
					List.assoc a env
				      with Not_found -> 
					raise (Analysis_error ("undefined variable '"^a^"' on line "^(string_of_int line)))
				   ) trans))

let multiset_fun_from_table_def m_trans n =
  let multiset = Array.make n 0 in
  let next variables candidate =
    let up i oldv newv =
      let who = variables.(i).possibles.(oldv) in
	multiset.(who) <- multiset.(who) - variables.(i).multiplicity;
      let who = variables.(i).possibles.(newv) in
	multiset.(who) <- multiset.(who) + variables.(i).multiplicity;
    in
    let k = Array.length variables in
    let rec doit i =
      if i=k then raise No_more_choice
    else if variables.(i).current = Array.length variables.(i).possibles - 1 then 
      (up i variables.(i).current 0;  variables.(i).current <- 0; doit (i+1))
    else let v = variables.(i).current in (
      up i v (v+1);
      variables.(i).current <- variables.(i).current + 1;
      if multiset.(v) <> candidate.(v) or multiset.(v+1)<>candidate.(v+1) then doit 0
    ) in  doit 0 
  in
  let init_multiset c variables candidate initial =
    for i=0 to n-1 do
      multiset.(i)<-initial.(i);
    done;
    multiset.(c) <- multiset.(c) + 1;
    let rec choose who p m =
      if p=m then ()
      else 
	let what = who.possibles.(p) in
	  if candidate.(what) >= multiset.(what) + who.multiplicity then (
	    who.current <- p;
	    multiset.(what) <- multiset.(what) + who.multiplicity
	  ) else choose who (p+1) m
    in
    let k = Array.length variables in
    for i=0 to k-1 do
      let who = variables.(i) in
      let m = Array.length who.possibles in
	choose who 0 m;
    done
  in
  let pattern_equal_multiset c variables candidate =
    let rec test i =
      if i=n then true
      else if multiset.(i) = candidate.(i) then test (i+1) 
      else false
    in test 0
  in
  let stop = Array.length m_trans in
  let rec try_transition p c x =
    let (_,_,fin_var, fin_val, variables) = m_trans.(p) in
    if pattern_equal_multiset c variables x then (
      match fin_var with
	| None -> fin_val
	| Some vr -> vr.possibles.(vr.current)
    )
    else  if (try next variables x; true  with No_more_choice -> false) then
      try_transition p c x
    else foreach_transition (p+1) c x
  and foreach_transition p c x = 
      if p = stop then c
	(* first test central cell *)
      else let (initial,central,_, _, variables) = m_trans.(p) in
	if IS.mem c central then (
	  if (try init_multiset c variables x initial; true  with No_more_choice -> false) then
	    try_transition p c x
	  else foreach_transition (p+1) c x	  
	)
	else foreach_transition (p+1) c x
  in
    foreach_transition 0

let gameoflife =
"
# Game of Life
n_states:2
neighborhood:Moore
symmetries:permute

var a={0,1}
var b={0,1}
var c={0,1}
var d={0,1}

# birth
0,0,0,0,0,0,1,1,1,1
# death
1,a,b,c,d,1,1,1,1,0
1,0,0,0,0,0,0,0,a,0
# in any other case, survive or don't birth
"

let transfo ac =
  let defby = choice_param ac "Define by" ["source text"; "source file"] in
  let str = file_param ac "Source file" "test.table" in
  let text = text_param ac "Source text" gameoflife in
    if neednewac ac then 
      try
	set_name ac family_name;
	set_object_name ac "see source text";
	let get_input =
	  match defby with
	    | "source text" -> let l = ref (Str.split (Str.regexp "\n") text) in
		(fun () -> match !l with
		   | [] -> raise End_of_file
		   | t::q -> l:=q; t
		)
	    | _ ->
 		let file = 
		  try 
		    set_object_name ac (Filename.basename str);
		    open_in str 
		  with Sys_error err -> (infomsg ("File '"^str^"' can not be opened: "^err); raise Pb)
		in (fun () -> input_line file)
	in
	let (n,neighb,sym,list) = scan_table get_input in
	let rec expand_all accu = function
	  | [] -> accu
	  | (v,l)::q -> 
	      let rec do_expand acc = function
		| [] -> acc
		| (Value i)::q -> do_expand (i::acc) q
		| (Varuse z)::q -> let l1 = List.assoc z accu in 
		    do_expand (List.rev_append (List.rev l1) acc) q
	      in expand_all ((v,(do_expand [] l))::accu) q
	in
	let env,trans = env_trans [] [] list in
	let envi = expand_all [] env in
	let c = ref 1 in
	let expanded_size = List.fold_left (fun acc (li,tra) ->
	  c:= 1;
	  for i = 0 to Array.length tra - 2 do
	    match tra.(i) with
		Varuse v -> let howmany = List.length (List.assoc v envi) in 
			    c:= !c * (if howmany = n then 1 else howmany)
	      | _ -> ()
	  done;
	  infomsg (Printf.sprintf "Line %d: %d trans\n" li !c);
	  !c + acc) 0 trans in
	let _ = infomsg ("#expanded="^(string_of_int expanded_size)) in
	let simplify_singleton_variables (l,transition) =
	  match transition.(Array.length transition - 1) with
	      Varuse v -> (
		try let e = List.assoc v envi in
		    if List.length e = 1 then (
		      transition.(Array.length transition - 1) <- Value (List.hd e)
		    )
		with _ -> ()
	      )
	    | _ -> ()
	in
	let _ = List.iter simplify_singleton_variables trans in
	  (* warning: in exp_env we possibly change original order *)
	let exp_env = List.map (fun (a,l) -> (a,(set_from_list (List.assoc a envi)))) envi in
	let sm_env = List.fold_left (fun acc x -> SM.add (fst x) (snd x) acc) (SM.empty) exp_env in
	  (match String.lowercase neighb with
	     | "vonneumann" -> (
		 set_dimension ac 2;
		 set_cardinality ac n;
		 set_vonneumann_neighborhood ac 1;
		 let _ = check_length (arity ac +1) trans in
		 let centerof x = x.(centercell ac) in
		 let neighb = relativecell ac in
		 let base_perm = [|(centercell ac); (neighb 0 1); (neighb 1 0); (neighb 0 (-1)); (neighb (-1) 0) |] in
		 let rotate i p = if p=0 then base_perm.(0) else base_perm.(1+((p-1+i) mod 4)) in
		 let reflect i = if i=0 then 0 else 1+((i+1) mod 4) in
		 let comp f g x = f (g x) in
		 let fixed_pos,sym_readers = match sym with
		   | "rotate4" -> (fun x-> x=0), [rotate 0; rotate 1; rotate 2; rotate 3]
		   | "rotate4reflect" -> (fun x-> x=0), [rotate 0; rotate 1; rotate 2; rotate 3; 
							 comp (rotate 0) reflect; comp (rotate 1) reflect; comp (rotate 2) reflect; comp (rotate 3) reflect]
		   | _ -> (fun x -> true),[Array.get base_perm] 
		 in
		   if sym = "permute" then
		     let m_trans = Array.of_list (List.map (prepare_multiset_trans n exp_env envi) trans) in
		       set_multiset_rule ac (multiset_fun_from_table_def m_trans (cardinality ac))
		   else let p_trans = Array.of_list (List.map (prepare_trans fixed_pos exp_env envi) trans) in
		     set_rule ac ~cache ~no_table:(n>=15) (fun_from_table_def p_trans (arity ac) sm_env fixed_pos sym_readers trans centerof)
	       )
	     | "moore" -> (
		 set_dimension ac 2;
		 set_cardinality ac n;
		 set_moore_neighborhood ac 1;
		 let _ = check_length (arity ac +1) trans in
		 let centerof x = x.(centercell ac) in
		 let neighb = relativecell ac in
		 let base_perm = [|centercell ac; neighb 0 1; neighb 1 1; neighb 1 0; neighb 1 (-1); neighb 0 (-1);
				   neighb (-1) (-1); neighb (-1) 0; neighb (-1) 1; |] in
		 let rotate8 i p = if p=0 then base_perm.(0) else base_perm.(1+((p-1+i) mod 8)) in
		 let rotate4 i p = if p=0 then base_perm.(0) else base_perm.(1+((p-1+2*i) mod 8)) in
		 let fixed_pos,sym_readers = match sym with
		   | "rotate4" -> (fun x-> x=0), [rotate4 0; rotate4 1; rotate4 2; rotate4 3]
		   | "rotate8" -> (fun x-> x=0), [rotate8 0; rotate8 1; rotate8 2; rotate8 3; rotate8 4; rotate8 5; rotate8 6; rotate8 7]
		   | _ -> (fun x -> true),[Array.get base_perm]
		 in 
		   if sym = "permute" then (
 		     let m_trans = Array.of_list (List.map (prepare_multiset_trans n exp_env envi) trans) in
		       set_multiset_rule ac (multiset_fun_from_table_def m_trans (cardinality ac))
		   ) else 
		     let p_trans = Array.of_list (List.map (prepare_trans fixed_pos exp_env envi) trans) in
		     set_rule ac ~cache (fun_from_table_def p_trans (arity ac) sm_env fixed_pos sym_readers trans centerof)
	       )
	     | _ -> raise (Analysis_error ("Unknown neighbourhood type '"^neighb^"'"))
	  )
      with e -> (
	infomsg ("Problem in loading file '"^str^"': "^(Printexc.to_string e)^". Aborted."); raise Pb)
    
let _ = add_shortcut ".*\\.table" (function x -> ["Source text", "";  "Source file",x; "Define by", "source file";])

let _ = valid transfo

