let family_name = "Table format"

open Acml

exception Pb

type table_item =
    Varuse of string
  | Value of int

type table_exec =
    Vardef of string * (table_item list)
  | Transition of table_item list

let scan_table input_line =
    (* regexp made compatible with DOS newline convention *)
  let header_states = Str.regexp "^[ \t]*n_states:[ \t]*\\([0-9]+\\)" in
  let header_neighb = Str.regexp "^[ \t]*neighborhood:[ \t]*\\([^# \r]+\\)\r?" in
  let header_sym = Str.regexp "^[ \t]*symmetries:[ \t]*\\([^# \r]+\\)\r?" in
  let variable = Str.regexp "^[ \t]*var \\([0-9]*[a-zA-Z][a-zA-Z0-9]*\\)[ \t]*=[ \t]*{\\([^}]+\\)}" in
  let variable_single = Str.regexp "^[ \t]*var \\([0-9]*[a-zA-Z][a-zA-Z0-9]*\\)[ \t]*=[ \t]*\\([a-zA-Z0-9]+\\)" in
  let transition_compact = Str.regexp "^\\([0-9]+\\)[ \t]*\\(#.*\\)?\r?$" in
  let transition = Str.regexp "^\\(\\(\\(\\([0-9]*[a-zA-Z][a-zA-Z0-9]*\\)\\|\\([0-9]+\\)\\),\\)*\\(\\([0-9]*[a-zA-Z][a-zA-Z0-9]*\\)\\|\\([0-9]+\\)\\)\\)[ \t]*\\(#.*\\)?\r?$" in
  let comment = Str.regexp "^\\([ \t]*#.*\\)" in
  let int_or_string i = try Value (int_of_string i) with _ -> Varuse i in
  let il_of_string s = List.map int_or_string (Str.split (Str.regexp " *, *") s) in
  let decompact s = Transition (List.map (fun x -> Value (int_of_string x)) (Str.split (Str.regexp "") s)) in
  let linepos = ref 0 in
  let rec scan_file states neighb sym acc  = 
    try 
      let s = input_line () in
	incr linepos;
	if Str.string_match comment s 0 || String.length s <= 1 then scan_file states neighb sym acc
	else if Str.string_match header_states s 0 then scan_file (int_of_string (Str.matched_group 1 s)) neighb sym acc
	else if Str.string_match header_neighb s 0 then scan_file states (Str.matched_group 1 s) sym acc
	else if Str.string_match header_sym s 0 then scan_file states neighb (Str.matched_group 1 s) acc
	else if Str.string_match variable s 0 then 
	  let v = Str.matched_group 1 s and l = Str.matched_group 2 s in
	    scan_file states neighb sym ((!linepos, (Vardef (v, (il_of_string l))))::acc)
	else if Str.string_match variable_single s 0 then 
	  let v = Str.matched_group 1 s and l = Str.matched_group 2 s in
	    scan_file states neighb sym ((!linepos, (Vardef (v, [int_or_string l])))::acc)
	else if Str.string_match transition_compact s 0 then scan_file states neighb sym 
	  ((!linepos, (decompact (Str.matched_group 1 s)))::acc)
	else if Str.string_match transition s 0 then scan_file states neighb sym 
	  ((!linepos, (Transition (il_of_string (Str.matched_group 1 s))))::acc)
	else (infomsg (Printf.sprintf "Syntax error in line %d" !linepos); raise Pb)
    with End_of_file -> (states, neighb, sym, List.rev acc)
  in
    (scan_file 1 "moore" "none" [])
    
exception Analysis_error of string
exception No_more_choice

module IS = Set.Make (struct 
  type t=int    
  let compare = compare
end)

let set_from_list =
  let rec doit s = function
    | [] -> s
    | t::q -> doit (IS.add t s) q
  in doit IS.empty 


(* DEBUG stats for cache *)
let cnt = ref 0
let deb_cache () = (
  incr cnt;
  infomsg ("[cache] "^(string_of_int !cnt))
)

(* heuristic cache mechanism *)
let cache f =
  let size = 1 lsl 16 in
  let nb = 2 in
  let harray = Array.init nb (fun _ -> Hashtbl.create size) in
  let usage = ref 0 in
  let pos = ref 0 in
  let _ = add_mem_info (nb*size*Kernel.bytes_per_word) in
  let rec find i x =
    try 
      Hashtbl.find harray.(i) x
    with Not_found -> (
      if i = ((nb + !pos - 1) mod nb) then (
	let r = f x in
	incr usage;
(* 	deb_cache (); *)
	Hashtbl.add harray.(!pos) (Array.copy x) r;
	if !usage >= size then (
	  pos := (!pos + 1) mod nb;
	  Hashtbl.clear harray.(!pos);
	  usage:= 0;
	);
	r
      ) else find ((i+1) mod nb) x
    )
  in (fun x -> find !pos x)

      
let rec env_trans env trans = function
    [] -> (List.rev env, List.rev trans)
  | (line, (Vardef (x,l)))::q -> env_trans ((x,l)::env) trans q
  | (line, (Transition l))::q -> env_trans env ((line,(Array.of_list l))::trans) q

let rec check_length length_check = function
  | [] -> ()
  | (line,a)::q -> 
    if Array.length a <> length_check then raise (Analysis_error (Printf.sprintf "bad transition length for declared neighborhood on line %d (got %d / expected %d)" 
								   line (Array.length a) length_check ))
    else check_length length_check q

let _ = declare_file_type "Table rule format" ["table"]


type multiset_variable =
    { mutable multiplicity : int;
      possibles : int array;
      mutable current : int;
    }

let mk_mv m pl =
  {multiplicity = m; possibles = Array.of_list pl; current =0}

(* translate a transition in a compact and more efficient form for multiset rules *)

let prepare_multiset_trans n env envi (line,trans) =
  let k = Array.length trans in
    (* initial will contain the multiset of states formed by constant values in the transition, but excluding the central cell *)
  let initial = Array.make n 0 in
    (* result is sorted in increasing order *)
  let find_all a = List.assoc a envi in
  let rec doit central acc_list var_ref i =
    if i = 0 then   (* we exclude central cell from the multiset choices since it is fixed during the choice loop process *)
      match trans.(i) with
	| Value v -> doit (IS.add v IS.empty) acc_list var_ref (i+1)
	| Varuse a -> let central = 
	    try 
	      List.assoc a env
	    with Not_found -> raise (Analysis_error ("undefined variable '"^a^"' on line "^(string_of_int line))) in
	    doit central acc_list var_ref (i+1)
    else if i = k-1 then match trans.(i) with
      | Value v -> (initial, central, None, v, Array.of_list (List.rev acc_list))
      | Varuse a -> let p = try
	  List.assoc a var_ref
	with Not_found -> raise (Analysis_error ("variable '"^a^"' is not bound but used as transition result on line "^(string_of_int line))) in
	  (initial, central,(Some p), 0, Array.of_list (List.rev acc_list))
    else match trans.(i) with
      | Value v -> (initial.(v) <- initial.(v)+1; doit central acc_list var_ref (i+1))
      | Varuse a -> 
	  if List.exists (fun y -> fst y = a) var_ref then let p = List.assoc a var_ref in (
	    p.multiplicity <- p.multiplicity + 1;
	    doit central acc_list var_ref (i+1)
	  ) else let vr = mk_mv 1 (find_all a) in
	    doit central (vr::acc_list) ((a,vr)::var_ref) (i+1)
  in doit IS.empty [] [] 0



let multiset_fun_from_table_def m_trans n =
  let multiset = Array.make n 0 in
  let next variables candidate =
    let up i oldv newv =
      let who = variables.(i).possibles.(oldv) in
	multiset.(who) <- multiset.(who) - variables.(i).multiplicity;
      let who = variables.(i).possibles.(newv) in
	multiset.(who) <- multiset.(who) + variables.(i).multiplicity;
    in
    let k = Array.length variables in
    let rec doit i =
      if i=k then raise No_more_choice
    else if variables.(i).current = Array.length variables.(i).possibles - 1 then 
      (up i variables.(i).current 0;  variables.(i).current <- 0; doit (i+1))
    else let v = variables.(i).current in (
      up i v (v+1);
      variables.(i).current <- variables.(i).current + 1;
      if multiset.(v) <> candidate.(v) || multiset.(v+1)<>candidate.(v+1) then doit 0
    ) in  doit 0 
  in
  let init_multiset c variables candidate initial =
    for i=0 to n-1 do
      multiset.(i)<-initial.(i);
    done;
    multiset.(c) <- multiset.(c) + 1;
    let rec choose who p m =
      if p=m then ()
      else 
	let what = who.possibles.(p) in
	  if candidate.(what) >= multiset.(what) + who.multiplicity then (
	    who.current <- p;
	    multiset.(what) <- multiset.(what) + who.multiplicity
	  ) else choose who (p+1) m
    in
    let k = Array.length variables in
    for i=0 to k-1 do
      let who = variables.(i) in
      let m = Array.length who.possibles in
	choose who 0 m;
    done
  in
  let pattern_equal_multiset c variables candidate =
    let rec test i =
      if i=n then true
      else if multiset.(i) = candidate.(i) then test (i+1) 
      else false
    in test 0
  in
  let stop = Array.length m_trans in
  let rec try_transition p c x =
    let (_,_,fin_var, fin_val, variables) = m_trans.(p) in
    if pattern_equal_multiset c variables x then (
      match fin_var with
	| None -> fin_val
	| Some vr -> vr.possibles.(vr.current)
    )
    else  if (try next variables x; true  with No_more_choice -> false) then
      try_transition p c x
    else foreach_transition (p+1) c x
  and foreach_transition p c x = 
      if p = stop then c
	(* first test central cell *)
      else let (initial,central,_, _, variables) = m_trans.(p) in
	if IS.mem c central then (
	  if (try init_multiset c variables x initial; true  with No_more_choice -> false) then
	    try_transition p c x
	  else foreach_transition (p+1) c x	  
	)
	else foreach_transition (p+1) c x
  in
    foreach_transition 0

let gameoflife =
"
# Game of Life
n_states:2
neighborhood:Moore
symmetries:permute

var a={0,1}
var b={0,1}
var c={0,1}
var d={0,1}

# birth
0,0,0,0,0,0,1,1,1,1
# death
1,a,b,c,d,1,1,1,1,0
1,0,0,0,0,0,0,0,a,0
# in any other case, survive or don't birth
"

let prepare_exp_transitions transi exp_env =
  let m = List.length transi in
  let res = Array.make m 0 in
  let foo = Array.make 1 IS.empty in
  let cond = Array.make m foo in
  let get_value = function
    | Value v -> v
    | _ -> (infomsg "Internal error"; raise Pb)
    in
  let rec transfer p = function
    | [] -> res,cond
    | (line,t)::q -> (
      res.(p) <- get_value t.(Array.length t - 1);
      cond.(p) <- (Array.map (function Value v -> IS.add v IS.empty | Varuse a -> 
	try 
	  List.assoc a exp_env
	with Not_found -> 
	  raise (Analysis_error ("undefined variable '"^a^"' on line "^(string_of_int line)))
      ) t);
      transfer (p+1) q
    )
  in
  transfer 0 transi

type trans_tree =
  | Test of (int * trans_tree array)
  | Final of (int array) * (int list)

let fun_from_expanded_transitions tree transi resu centerof =
  let rec find_first lespos x = function
    | [] ->  centerof x 
    | t::q -> let rec test p i =
	   if i=p then resu.(t)
	   else if IS.mem x.(lespos.(i)) transi.(t).(lespos.(i)) then test p (i+1)
	   else find_first lespos x q
	 in
	 test (Array.length lespos) 0
  in
  let rec find_tree x = function
    | Final (lespos, lestrans) -> find_first lespos x lestrans
    | Test (pos, choices) -> find_tree x choices.(x.(pos))
  in
  (fun x -> find_tree x tree)

exception Done

let tree_size = ref 0

let rec subdivide_tree n p done_pos trans_tabref trans_list =
  let themin = ref (-1) in
  let theminlist = ref [] in
  let thepos = ref 0 in
  let update_min pos nv nl =
    if !themin < 0 || nv < !themin then (
      themin := nv;
      thepos := pos;
      theminlist := nl
    ) in
  let rec find_new_pos i =
    if i>=p then raise Done
    else if List.mem i done_pos then find_new_pos (i+1)
    else i
  in
  let rec split pos j acc = if j=n then List.rev acc
    else split pos (j+1) ((List.filter (fun l -> IS.mem j trans_tabref.(l).(pos)) trans_list)::acc)
  in
  let rec dothejob after =
    let pos = find_new_pos after in
    let nl = split pos 0 [] in
    let nv = List.fold_left (fun acc l -> List.length l + acc) 0 nl in
    update_min pos nv nl;
    dothejob (after+1)
  in
  if List.length done_pos < p && List.length trans_list > 10 && !tree_size < 1 lsl 18 then
    try dothejob 0 with Done -> (
      tree_size := !tree_size + n;
      let sub = List.map (subdivide_tree n p (!thepos::done_pos) trans_tabref) !theminlist in
      Test (!thepos, (Array.of_list sub))
    )
  else (
    let rec pending_pos acc t =
      if t = p then  Array.of_list acc
      else if List.mem t done_pos then pending_pos acc (t+1)
      else pending_pos (t::acc) (t+1)
    in Final ((pending_pos [] 0), trans_list)
  )

let transfo ac =
  let defby = choice_param ac "Define by" ["source text"; "source file"] in
  let str = file_param ac "Source file" "test.table" in
  let text = text_param ac "Source text" gameoflife in
    if neednewac ac then 
      try
	set_name ac family_name;
	set_object_name ac "see source text";
	let get_input =
	  match defby with
	    | "source text" -> let l = ref (Str.split (Str.regexp "\n") text) in
		(fun () -> match !l with
		   | [] -> raise End_of_file
		   | t::q -> l:=q; t
		)
	    | _ ->
 		let file = 
		  try 
		    set_object_name ac (Filename.basename str);
		    open_in str 
		  with Sys_error err -> (infomsg ("File '"^str^"' can not be opened: "^err); raise Pb)
		in (fun () -> input_line file)
	in
	let (n,neighb,sym,list) = scan_table get_input in
	let rec expand_all accu = function
	  | [] -> accu
	  | (v,l)::q -> 
	      let rec do_expand acc = function
		| [] -> acc
		| (Value i)::q -> do_expand (i::acc) q
		| (Varuse z)::q -> let l1 = List.assoc z accu in 
		    do_expand (List.rev_append (List.rev l1) acc) q
	      in expand_all ((v,(do_expand [] l))::accu) q
	in
	let env,trans = env_trans [] [] list in
	let envi = expand_all [] env in
	let expand_variables_in_transitions (l,transition) =
	  match transition.(Array.length transition - 1) with
	      Varuse v -> (
		let pos = ref [] in
		for i = 0 to Array.length transition - 1 do
		  match transition.(i) with
		      Varuse u when u=v -> pos:= i::(!pos)
		    | _ -> ()
		done;
		let vals = 
		  try 
		    List.assoc v envi
		  with Not_found -> raise (Analysis_error ("undefined variable '"^v^"' on line "^(string_of_int l)))
		in
		let spit v = let r = Array.make (Array.length transition) (Value 0) in
			     for i = 0 to Array.length transition - 1 do
			       if List.mem i !pos then r.(i) <- Value v else r.(i) <- transition.(i)
			     done; (l,r)
		in List.map spit vals
	      )
	    | _ -> [l,transition]
	in
	let trans = List.flatten (List.map expand_variables_in_transitions trans) in
	let expand_symetries syms transitions =
	  let do_expand l t s =
	    let r = Array.make (Array.length t) (Value 0) in
	    for i=0 to Array.length t - 2 do
	      r.(s i) <- t.(i)
	    done;
	    r.(Array.length t - 1) <- t.(Array.length t - 1);
	    (l,r)
	  in List.flatten (List.map (fun (l,t) -> List.map (do_expand l t) syms) transitions)
	in 
	  (* warning: in exp_env we possibly change original order *)
	let exp_env = List.map (fun (a,l) -> (a,(set_from_list (List.assoc a envi)))) envi in
	  (match String.lowercase_ascii neighb with
	     | "vonneumann" -> (
		 set_dimension ac 2;
		 set_cardinality ac n;
		 set_vonneumann_neighborhood ac 1;
		 let _ = check_length (arity ac +1) trans in
		 let centerof x = x.(centercell ac) in
		 let neighb = relativecell ac in
		 let base_perm = [|(centercell ac); (neighb 0 1); (neighb 1 0); (neighb 0 (-1)); (neighb (-1) 0) |] in
		 let rotate i p = if p=0 then base_perm.(0) else base_perm.(1+((p-1+i) mod 4)) in
		 let reflect i = if i=0 then 0 else 1+((i+1) mod 4) in
		 let comp f g x = f (g x) in
		 let fixed_pos,sym_readers = match sym with
		   | "rotate4" -> (fun x-> x=0), [rotate 0; rotate 1; rotate 2; rotate 3]
		   | "rotate4reflect" -> (fun x-> x=0), [rotate 0; rotate 1; rotate 2; rotate 3; 
							 comp (rotate 0) reflect; comp (rotate 1) reflect; comp (rotate 2) reflect; comp (rotate 3) reflect]
		   | _ -> (fun x -> true),[Array.get base_perm] 
		 in
		   if sym = "permute" then
		     let m_trans = Array.of_list (List.map (prepare_multiset_trans n exp_env envi) trans) in
		       set_multiset_rule ac (multiset_fun_from_table_def m_trans (cardinality ac))
		   else 
		     let transi = expand_symetries sym_readers trans in
		     let re,tr = prepare_exp_transitions transi exp_env in
		     let thetree = subdivide_tree n (arity ac) [] tr (Array.to_list (Array.init (Array.length tr) (fun x -> x))) in
		     set_rule ac ~cache ~no_table:(n>=15) (fun_from_expanded_transitions thetree tr re centerof)
	       )
	     | "moore" -> (
		 set_dimension ac 2;
		 set_cardinality ac n;
		 set_moore_neighborhood ac 1;
		 let _ = check_length (arity ac +1) trans in
		 let centerof x = x.(centercell ac) in
		 let neighb = relativecell ac in
		 let base_perm = [|centercell ac; neighb 0 1; neighb 1 1; neighb 1 0; neighb 1 (-1); neighb 0 (-1);
				   neighb (-1) (-1); neighb (-1) 0; neighb (-1) 1; |] in
		 let rotate8 i p = if p=0 then base_perm.(0) else base_perm.(1+((p-1+i) mod 8)) in
		 let rotate4 i p = if p=0 then base_perm.(0) else base_perm.(1+((p-1+2*i) mod 8)) in
		 let reflect = function
		   | 0 -> 0
		   | 1 -> 1
		   | 5 -> 5
		   | i -> 2 + (8-i) in
		 let comp f g x = f (g x) in
		 let fixed_pos,sym_readers = match sym with
		   | "rotate4" -> (fun x-> x=0), [rotate4 0; rotate4 1; rotate4 2; rotate4 3]
		   | "rotate4reflect" -> (fun x-> x=0), [rotate4 0; rotate4 1; rotate4 2; rotate4 3; comp (rotate4 0) reflect; comp (rotate4 1) reflect; comp (rotate4 2) reflect; comp (rotate4 3) reflect]
		   | "rotate8" -> (fun x-> x=0), [rotate8 0; rotate8 1; rotate8 2; rotate8 3; rotate8 4; rotate8 5; rotate8 6; rotate8 7]
		   | _ -> (fun x -> true),[Array.get base_perm]
		 in 
		   if sym = "permute" then (
 		     let m_trans = Array.of_list (List.map (prepare_multiset_trans n exp_env envi) trans) in
		       set_multiset_rule ac (multiset_fun_from_table_def m_trans (cardinality ac))
		   ) else 
		     let transi = expand_symetries sym_readers trans in
		     let re,tr = prepare_exp_transitions transi exp_env in
		     let thetree = subdivide_tree n (arity ac) [] tr (Array.to_list (Array.init (Array.length tr) (fun x -> x))) in
		     set_rule ac ~cache (fun_from_expanded_transitions thetree tr re centerof)
	       )
	     | _ -> raise (Analysis_error ("Unknown neighbourhood type '"^neighb^"'"))
	  )
      with e -> (
	infomsg ("Problem in loading file '"^str^"': "^(Printexc.to_string e)^". Aborted."); raise Pb)
    
let _ = add_shortcut ".*\\.table" (function x -> ["Source text", "";  "Source file",x; "Define by", "source file";])

let _ = valid transfo

