open Debug
open Num

exception Unimplemented of string
exception Unknownparam of string
exception Badparam of string
exception BadACML of string
exception Debug of string
exception RuleTooLarge of Big_int.big_int
exception AbortRule of string
exception ComputationInterrupted

(* Shorcutrecursive = string list *)
exception Shortcut of (bool * string * ((string * string) list))

exception NoNeedCache

(************* TYPE AND TYPE INTERFACE *****************)

(* from time to time I increase this value which is used to handle backward compatibility with old saved files *)
let theversion = 17
let stoptime = ref (-1)

let rec count_bits value bits =
  if value > 0 then count_bits (value lsr 1) (bits+1)
  else bits

(*************** Eval proc cache *******************)

exception Res of int 

let get_cache () =
  try 
    let f = open_in "/proc/cpuinfo" in
    let re = Str.regexp "cache size.* \\([0-9]*\\) .*" in
      try
	while true do
	  let s = (input_line f) in 
	    if Str.string_match re s 0 then raise (Res (int_of_string (Str.matched_group 1 s)))
	done; 0
      with e -> close_in f; raise e 
  with 
      Res x -> (x lsl 10)
    | _ -> (1 lsl 19)

let bytes_in_cache = get_cache ()

(*****************************************************)

let max_int_bit = count_bits max_int 0
let max_array_bit = count_bits Sys.max_array_length 0
let bytes_per_word = Sys.word_size / 8
let ints_in_cache = bytes_in_cache / bytes_per_word

type fast_array =
    Native of int array
  | Medium of int array
  | Big of (int, Bigarray.int_elt, Bigarray.c_layout) Bigarray.Array1.t

type compact_info = 
    {
      log_blocs_per_int : int;
      bloc_size : int;
    }

type fast_matrix =
    MatNative of int array array
  | MatMixed of compact_info*(int array array)

type matrix =
    Unidim of int array array
    | Bidim of fast_matrix
    | Zerodim of int ref

type position = 
    Line of int
    | Plane of int * int
    | Center
(*     | Shift of position*position *)

module SpotSet = Set.Make
  (
    struct
      type t = position
      let compare = compare
    end
  )
  
module ActivSet = Set.Make
  (
    struct
      type t = int
      let compare = compare
    end
  )
  
type relposition =
    Birelpos of int * int
  | Unirelpos of int

type neighbourhoodtype =
    Moore of int
    | VonNeumann of int
    | Cross of int

type determintype =
    WantToBeDeterministic
  | Deterministic
  | DeterministicButSlow
  | Stochastic

type coretype =
{
  transirule : ruletype;
  cardinal : int;
  neighb : neighbourhoodtype;  
}

and ruletype =
    Table of fast_array
    | XTable of int array (* (int, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t *)
    | MultisetTable of (int, Bigarray.int_elt, Bigarray.c_layout) Bigarray.Array1.t
    | MultisetFunArray of (int -> int array -> int)
    | FunArray of (int array -> int)
    | FunFun of ((int -> int -> int) -> int)
    | MultisetFunFun of (int -> (int -> int) -> int)
    | Sparse of ActivSet.t * ruletype
    | Transducer

type measuretype =
    Uniform
  | Bernouilli of int array

type confzone =
    Rectangular of position * position
  | Randomrectangular of position * int * int

type moduloarray =
    Udim of int array
  | Bdim of int array array

type conftype =
    Random of measuretype
  | Finite of (position * int) list
  | Randomtranslate of conftype
  | Subzone of confzone * conftype
  | Monochrome of int
  | Sized of int * int * conftype
  | Layers of conftype list
  | Repeat of int * conftype
  | Modulo of moduloarray
  | Custom of (int -> int -> int)
  | CustomArray of (unit -> int array array)
  | Masked of (int -> int -> bool) * conftype

type drawing =
    Up
  | Down
  | Left
  | Right
  | Nocolor
  | Uni of (int*int*int)
  | Arround of (int*int*int)
  | Overdraw of drawing list
      
type colormaptype =
    Auto
  | General of (int -> drawing)
  | Genscale of ((int*int*int)*(int*int*int))
  | Tabscale of (int*int*int) array
  | Graycycle
  | Simple of (int*int*int) array
  | Viewer of (int -> int -> (int -> int -> int) -> drawing)

let defaultcolormapname = "Gray scale"

let thecolormaps = [
  defaultcolormapname, Tabscale [|(0,0,0); (255, 255, 255)|];
]

let colormapnames = List.map fst thecolormaps


type requesttype =
    Newall
  | Newac
  | Newconf
  | NeitherNor
  | Blocked of requesttype

type idrawing =     
    GenGrid of int * int * int * int * int (** xstep ystep and color components *)
  | Markcell of ((int * int) * (int * int * int)) (** coord/color  *)

type infodrawings =
    { mutable under : idrawing list;
      mutable over : idrawing list;
    }

type intparam =
    {
      mutable pvalue : int;
      mutable pdefault : int;
      mutable pmin : int;
      mutable pmax : int
    }

type floatparam =
    {
      mutable pfvalue : float;
      mutable pfdefault : float;
      mutable pfmin : float;
      mutable pfmax : float
    }

type boolparam =
    {
      mutable pbvalue : bool;
      mutable pbdefault : bool
    }

type choiceparam =
    {
      mutable pcvalue : string;
      mutable pchoices : string list;
    }

type stringparam =
    {
      mutable psvalue : string;
      mutable psdefault : string;
    }

type paramcontent =
    ParInt of intparam
  | ParFloat of floatparam
  | ParBool of boolparam
  | ParChoice of choiceparam
  | ParString of stringparam

type metaparam =
    Filechooser
  | Grouparam of string list (* the name associated to this is the
				name of the group and the list is the
				list of params in the group *)
  | Doc of string (* Some explanation about a parameter or an action *)
  | Multiline

(* (name,content)  lists *)
type 'a itemlist = (string * 'a) list

type rst = Random2.State.t

type seedtype = 
    {
      (* rule random state *)
      mutable rs_rule : rst;
      (* other random state *)
      mutable rs_conf : rst;
      mutable ordinal : int;
    }


(* Type for information saving *)
(* Union type to ensure backward compatibility of saving format *) 
(* idea: use list of "saving" items *)

type saving =
    SavSeed of seedtype
  | SavObjname of string
  | SavParams of (paramcontent itemlist)
  | SavUniverse of (int*int)
  | SavKernelSig of string*float
  | SavAcmlSig of string*float
  | SavRandom2Sig of string*float
      
type randominfotype = Big_int.big_int option


(* Whether we use rounding to power of two *)
type powerTwoInfo =
    QueDalle
  | Vcard of int
  | Varit of int
      (* first cardinal, then arity *)
  | Vboth of int * int


type actionsource =
    SrcLeftclick
  | SrcRightclick
  | SrcMiddleclick
  | SrcSave
  | SrcSnapshot
  | SrcMeasurement
  | SrcKeyClick of string


type measurementtype = 
    MeTrigger
  | MeOnOff of (bool ref)

(* An action can be of 4 kinds:
   - something called on a selection of cells (typically a modification of cells value)
   - something called on a file name or an output stream and an acml object and a configuration
   - something called on a configuration
   - anything else
*)
type actiontype =
    (*MODIF*) Cycle
  | (*MODIF*) Force of int
  | (*MODIF*) Funkmod of (int -> (int*int*int*int) -> int -> int -> int)
  | (*MODIF*) Freemod of ((int * int * int * int) -> (int -> int -> int) -> (int -> int -> int -> unit)-> unit)
  | (*SAV*) Source
  | (*SAV*) GlobalSav of (string * (out_channel -> actype -> (int array array) -> unit)) (* File extension, output stream, etc *)
  | (*SNAPSHOT*) Snapshot of (int -> int -> (int -> int -> int) -> unit)
  | (*SNAPSHOT*) Wysiwyg
  | (*MEASUREMENT*) Measurement of (measurementtype * (int -> int -> int -> (int -> int -> int) -> (int -> int -> int -> unit) -> unit))
  | NoAction
and paramschema = 
    { 
      mutable params : paramcontent itemlist;
      mutable metaparams : metaparam itemlist;
      mutable colormaps : colormaptype itemlist;
      mutable actions : actiontype itemlist;
      mutable statesnames : ((int,string) Hashtbl.t) option;
    }
and actype =
    { 
      mutable transitionrule : ruletype;
      mutable seed : seedtype;
      mutable undo_seed : seedtype;
      mutable cardinality : int;
      mutable dimension : int;
      mutable neighbourhood : neighbourhoodtype;  
      mutable radius : int option;

      mutable universe : int*int;
      mutable theinitconf : string;
      mutable initconfs : conftype itemlist;
      mutable spots : SpotSet.t;
      mutable modified : SpotSet.t;
      mutable complexconf : bool;

      mutable action : (actionsource * string) list;
      mutable colormap : string;
      mutable parameters : paramschema;

      mutable name : string option;
      (* nick_name should be a unique identifier referring to the generator used to produce the object *)
      mutable nick_name : string option;
      mutable obj_name : string option;
      mutable randominfo : randominfotype;
      mutable infotodraw : infodrawings;
      mutable colorcache : int32 array;
      mutable wanted : requesttype;
      mutable srcfile : string option;
      mutable dbg : Debug.debugcats list;
      mutable determinization : determintype;
      mutable binary_power : powerTwoInfo;
      mutable meminfo : int;
      mutable refmeminfo : int ref;
      mutable description : (string * string) list;
      version : int;
    }


type acprototype = 
    Acerror of string
  | Acompletion of int
  | Acfinished of actype
(*   | Acpidinfo of int *)


(* x = number of different values *)
let round_power_two x =
  let rec doit i =
    if i = max_int_bit then internalbug "round_power_two";
    if x <= 1 lsl i then 1 lsl i  else doit (i+1)
  in doit 0


(* x = max digit we want to represent *)
let exp_round_power_two x =
  let rec doit i =
    if i = max_int_bit then internalbug "round_power_two";
    (* if x <= (1 lsl i) - 1 then i  else doit (i+1) *)
    if x <= (1 lsl i) - 1 then i  else doit (i+1)
  in doit 1

(************************************************************)
(***  Fast arrays ***)
(************************************************************)

let max_native = (min (Sys.max_array_length) ints_in_cache) / 3

let make_fast_array size max_value =
  if size < max_native then Native (Array.make size 0)
(* BUG in medium arrays *)
(*   else if max_value < (1 lsl 4) then Medium (Array.make ((size lsr 3)+1) 0) *)
  (* else if max_value < (1 lsl 8) then Medium (Bigarray.Array1.create Bigarray.int8_unsigned Bigarray.c_layout size) *)
  else Big (Bigarray.Array1.create Bigarray.int Bigarray.c_layout size)

let fast_array_type = function
    Native _ -> "table (full-sized)"
  | Medium _ -> "table (compacted)"
  | Big _ -> "table (big)"

let fast_array_size = function
    Native t -> Array.length t
  | Medium t -> (Array.length t) lsl 3
  | Big t -> Bigarray.Array1.dim t

let fast_array_memsize = function
    Native t -> bytes_per_word*(Array.length t)
  | Medium t -> bytes_per_word*(Array.length t)
  | Big t -> bytes_per_word*(Bigarray.Array1.dim t)

let getter_from_fast_array = function
    Native t -> (fun i -> Array.unsafe_get t i)
  | Medium t -> (fun i -> ((Array.unsafe_get t (i lsr 3)) lsr (3*(i land ((1 lsl 3) - 1)))) land ((1 lsl 3) - 1))
  | Big t -> (fun i -> Bigarray.Array1.unsafe_get t i)

let setter_from_fast_array = function
    Native t -> (fun i c -> Array.unsafe_set t i c)
  | Medium t -> (fun i c -> 
		   let where = (i lsr 3) in
		   let shift = (i land 7) * 3 in
		     Array.unsafe_set t where (((Array.unsafe_get t where) land (lnot (7 lsl shift))) lor 
						   (c lsl shift))
		)
  | Big t -> (fun i c -> Bigarray.Array1.unsafe_set t i c)

(************************************************************)
(***  Fast matrix ***)
(************************************************************)

let max_mat_native = ints_in_cache

let rec loghowmanyblocs now size =
  if (1 lsl (now+1))*size <= max_int_bit then loghowmanyblocs (now+1) size else now

let make_fast_matrix sizex sizey max_value =
  if (sizex lsr 2)*sizey < max_mat_native then MatNative (Array.make_matrix sizex sizey 0)
  else
    let bloc_s = exp_round_power_two max_value in
    let bloc_logn = loghowmanyblocs 1 bloc_s in
      (* Oh ! joie du +1 qui m'a coûté plusieurs heures *) 
    let new_y = (sizey lsr bloc_logn) + 1 in
      debugline Dbgdev (Printf.sprintf "New matrix x=%d y=%d mask=%d newy=%d" sizex sizey ((1 lsl (bloc_logn)) - 1) new_y);
      MatMixed ({bloc_size = bloc_s; log_blocs_per_int = bloc_logn},(Array.make_matrix sizex new_y 0))
	
let getter_from_fast_matrix = function
    MatNative t -> (fun i j -> Array.unsafe_get (Array.unsafe_get t i) j)
  | MatMixed (nfo,t) -> 
	let y_mask = (1 lsl (nfo.log_blocs_per_int)) - 1 in
	let bloc_mask = (1 lsl (nfo.bloc_size)) - 1 in
      (fun i j -> 
	 ((Array.unsafe_get (Array.unsafe_get t i) (j lsr nfo.log_blocs_per_int)) lsr 
	    ((j land y_mask) * nfo.bloc_size)) land bloc_mask
      )

let setter_from_fast_matrix = function
    MatNative t -> (fun i j c -> Array.unsafe_set (Array.unsafe_get t i) j c)
  | MatMixed (nfo,t) -> 
	let y_mask = (1 lsl (nfo.log_blocs_per_int)) - 1 in
	let bloc_mask = (1 lsl (nfo.bloc_size)) - 1 in
	  (fun i j c ->
	     let where = (j lsr nfo.log_blocs_per_int) in
	     let shift = (j land y_mask) * nfo.bloc_size in
	     let ith = Array.unsafe_get t i in
	       Array.unsafe_set ith where (((Array.unsafe_get ith where) land (lnot (bloc_mask lsl shift))) lor 
					     (c lsl shift))
	  )

let fast_matrix_type = function
    MatNative _ -> "full-size"
  | MatMixed (nfo,_) -> Printf.sprintf "compacted %dx blocs_size=%d" (1 lsl nfo.log_blocs_per_int) nfo.bloc_size

let fast_matrix_size = function
    MatNative t -> (Array.length t, Array.length t.(0))
  | MatMixed (nfo,t) -> (Array.length t, (1 lsl nfo.log_blocs_per_int)*(Array.length t.(0)))

let copy_from_to x y m nm =
  for i=0 to x-1 do
    for j=0 to y-1 do
      Array.unsafe_set (Array.unsafe_get nm i) j (Array.unsafe_get (Array.unsafe_get m i) j)
    done
  done

let clone_matrix m =
  let x,y = Array.length m, Array.length m.(0) in
  let nm = Array.make_matrix x y 0 in
    copy_from_to x y m nm; nm

let clone_fast_matrix = function
    MatNative t -> MatNative (clone_matrix t)
  | MatMixed (nfo,m) -> MatMixed (nfo, clone_matrix m)

let copy_fast_matrix dep arr = 
  if fast_matrix_size dep = fast_matrix_size arr then 
    (
      match (dep,arr) with
	  (MatNative t, MatNative s) -> copy_from_to (Array.length t) (Array.length t.(0)) t s
	| (MatMixed (nfo,m), MatMixed (nfp,n)) -> if nfo=nfp then copy_from_to (Array.length m) (Array.length m.(0)) m n
	| _ -> internalbug "copy fast matrix type mismatch"
    )

(************************************************************)

let mixwanted ac com =
  match ac.wanted with
      Blocked x -> ac.wanted <- Blocked com
    | _ -> ac.wanted <- com

let block a =
  match a.wanted with 
      Blocked _ -> ()
    | x -> a.wanted <- Blocked x

let unblock a =
  match a.wanted with 
      Blocked x -> a.wanted <- x
    | x -> ()

let needresetconf a =
  match a.wanted with
      Newall | Newac -> true
    | _ -> false

let needresetactions a =
  match a.wanted with
      Newall | Newac | NeitherNor -> true
    | _ -> false


let needresetallparameters a =
  match a.wanted with
      Blocked _ -> false
    | _ -> true

let needresetdescription = needresetactions
      
let neednewac a =
  let rec doit = function 
  Newall | Newac -> true
    | Blocked x -> doit x
    | _ -> false
  in doit a.wanted

let neednewac_unblocked a =
  match a.wanted with
      Newall | Newac -> true
    | _ -> false
	
let neednewconf a =
  let rec doit = function 
  Newall | Newconf -> true
    | Blocked x -> doit x
    | _ -> false
  in doit a.wanted
  
let get_initconf a =
  try
    List.assoc a.theinitconf a.initconfs
  with Not_found -> Random(Uniform)

let get_colormap a =
  try 
    List.assoc a.colormap a.parameters.colormaps
  with Not_found -> Auto
      
let modif_name = "configuration modifier"
let save_name = "saving method"
let snap_name = "snapshot method"
let measure_name = "measurement method"

let kind_of_action = function
    Cycle | Force _  | Funkmod _ | Freemod _ -> modif_name
  | Source | GlobalSav _ -> save_name
  | Snapshot _ | Wysiwyg -> snap_name
  | Measurement _ -> measure_name
  | _ -> "generic action"

let get_modifs a =
  let filter x = kind_of_action (snd x) = modif_name
  in List.filter filter a.parameters.actions
      
let get_savers a =
  let filter x = kind_of_action (snd x) = save_name
  in List.filter filter a.parameters.actions
      
let get_snaps a =
  let filter x = kind_of_action (snd x) = snap_name
  in List.filter filter a.parameters.actions

let get_measures a =
  let filter x = kind_of_action (snd x) = measure_name
  in List.filter filter a.parameters.actions
      
let set_statename a i s =
  let ht =
    (match a.parameters.statesnames with
	None -> let h = Hashtbl.create a.cardinality in (a.parameters.statesnames <- Some h; h)
      | Some h -> h
    ) in
  Hashtbl.replace ht i s

let get_statename a i =
  match a.parameters.statesnames with
      None -> raise Not_found
    | Some h -> Hashtbl.find h i
      

let setaction ac src act =
  ac.action <- (src,act)::(List.remove_assoc src ac.action)

let setanotheraction ac src act =
  if List.for_all (fun x -> src <> fst x || act <> snd x) ac.action then ac.action <- (src,act)::(ac.action)

let addaction ac name act =
  ac.parameters.actions <- (name,act)::(List.remove_assoc name ac.parameters.actions)

let getactionname ac src =
  try
    List.assoc src ac.action
  with Not_found -> "nothing"

let getactionnamefiltered ac filter out =
  let rec collect acc = function
    | [] -> acc
    | (src,name)::q -> if filter src then collect ((out (src,name))::acc) q else collect acc q
  in collect [] ac.action

let getaction ac src =
  try 
    List.assoc (List.assoc src ac.action) ac.parameters.actions
  with Not_found ->  NoAction

let getallactions ac src =
  let rec find acc = function
      [] -> List.rev acc
    | t::q -> 	if fst t = src then 
	try find ((snd t, (List.assoc (snd t) ac.parameters.actions))::acc) q 
	with Not_found -> find acc q
      else find acc q
  in find [] ac.action
       
let do_all_measurements ac curt x y read write =
  let single_do (name,content) = match content with
      Measurement (MeOnOff boolref, howto) -> 
	if !boolref then (
	  debugline Dbgdev ("Applying measurement '"^name^"'");
	  howto curt x y read write
	)
    | _ -> ()
  in List.iter single_do (getallactions ac SrcMeasurement)

let defaultaction ac src act =
  try 
    ignore (List.assoc src ac.action)
  with Not_found ->  ac.action <- (src,act)::(List.remove_assoc src ac.action)

(* Make sure that all choices made via GUI refer to existing actions *)
let purge_actions ac =
  ac.action <- List.filter (fun (_,name) -> List.exists (fun x -> fst x=name) ac.parameters.actions) ac.action;
  let doit src getter extract =
    if not (List.exists (fun x -> fst x = src) ac.action) then 
      let acts = getter ac in
	if acts<>[] then defaultaction ac src (fst (extract acts))
  in
    doit SrcLeftclick get_modifs List.hd;
    doit SrcRightclick get_modifs (function
				       t::(u::q) -> u
				     | t::q -> t
				     | _ -> assert false
				  );
    doit SrcSnapshot get_snaps List.hd;
    doit SrcSave get_savers List.hd;
    doit SrcMeasurement get_measures List.hd;
    if not (List.exists (fun x -> fst x = ac.theinitconf) ac.initconfs) then ac.theinitconf <- fst (List.hd (ac.initconfs))

(* Tells whether colors are local and storable in an array cache *)
let complex_colors a =
  if a.cardinality > min Sys.max_array_length (1 lsl 29) then true else
    match get_colormap a with
	Genscale _ | Tabscale _ | Graycycle | Simple _ -> false
      | _ -> true

let uses_view a =
  match get_colormap a with
      Genscale _ | Tabscale _ | Graycycle | Simple _ | General _ -> false
    | _ -> true


let cloneparam (s,par) = match par with
    ParInt p -> (s,ParInt {pdefault=p.pdefault; pmin=p.pmin; pmax=p.pmax; pvalue=p.pvalue})
  | ParFloat p -> (s,ParFloat {pfdefault=p.pfdefault; pfmin=p.pfmin; pfmax=p.pfmax; pfvalue=p.pfvalue})
  | ParBool p -> (s,ParBool {pbdefault=p.pbdefault; pbvalue=p.pbvalue})
  | ParChoice p -> (s,ParChoice {pchoices=p.pchoices; pcvalue=p.pcvalue})
  | ParString p -> (s,ParString {psdefault=p.psdefault; psvalue=p.psvalue})

let cloneseed s =
  {
    rs_rule = Random2.State.copy s.rs_rule;
    rs_conf = Random2.State.copy s.rs_conf;
    ordinal = s.ordinal;
  }

let copycolormap = function
  | General x -> General x
  | Tabscale x -> Tabscale (Array.copy x)
  | Simple x -> Simple (Array.copy x)
  | e -> e

let copyparameters p =
  {
    params = List.map cloneparam p.params;
    actions = p.actions;
    metaparams = p.metaparams;
    colormaps = List.map (fun (n,c) -> (n,copycolormap c)) p.colormaps;
    statesnames = (match p.statesnames with
	None -> None
      | Some h -> Some (Hashtbl.copy h)
    )
  }

let copyrule x = x
	
    

let cloneac x =
  {
    parameters = copyparameters x.parameters;
    transitionrule=copyrule x.transitionrule;
    seed = cloneseed x.seed; undo_seed = cloneseed x.undo_seed; theinitconf = x.theinitconf; initconfs = x.initconfs; randominfo = x.randominfo;
    cardinality = x.cardinality; neighbourhood = x.neighbourhood; dimension = x.dimension; radius = x.radius;
    name = x.name; obj_name = x.obj_name; action = x.action; infotodraw = x.infotodraw; 
    colorcache = Array.copy x.colorcache;
    wanted = x.wanted; srcfile = x.srcfile; version = x.version; dbg = x.dbg; spots = x.spots; modified = x.modified;
    universe = x.universe; complexconf = x.complexconf; colormap = x.colormap; determinization=x.determinization;
    binary_power = x.binary_power; meminfo = x.meminfo; refmeminfo = x.refmeminfo; nick_name = x.nick_name; description = x.description
      
  }
    
(*********************************************)
(* PRNGs for any usage (not saved)           *)
(*********************************************)

let (neutralrandom, neutralfloatrandom, initneutralrandom) = 
  let st = ref None in
  let doit f = match (!st) with
      None -> (let s = Random2.State.copy (Random2.get_state ()) in
		 st := Some s; f s)
    | Some s -> f s
  in
    ((fun i -> doit (fun s -> Random2.State.int s i)),
     (fun i -> doit (fun s -> Random2.State.float s i)),
     (fun x -> doit (fun s -> Random2.State.full_init s x))
    )


(***************************************************)
(* PRNGs for rules and configurations generation   *)
(***************************************************)

let nbofbits i =
  let rec doit acc j =
    if j / 2 > 0 then doit (1+acc) (j/2)
    else acc
  in doit 1 i

let (myrandom, myrandomnocount, myrandom_float, myrandomnocount_float, resetrandom, setrandomcount, getcount, getnocount, setcount, setnocount, hasard, neutralize_myrandom, std_myrandom) =
  let countstate = Random2.State.copy (Random2.get_state ()) in
  let nocountstate = Random2.State.copy (Random2.get_state ()) in
  let std_my_count = Random2.State.int countstate in
  let std_my_nocount = Random2.State.int nocountstate in
  let std_my_count_float = Random2.State.float countstate in
  let std_my_nocount_float = Random2.State.float nocountstate in
  let s_m_n = ref std_my_nocount in
  let s_m_c = ref std_my_count in
  let s_m_n_float = ref std_my_nocount_float in
  let s_m_c_float = ref std_my_count_float in
  let rnd = ref Big_int.unit_big_int in
    ((fun i -> rnd := Big_int.add_int_big_int (nbofbits i) !rnd;
	!s_m_c i),
     (fun i ->  !s_m_n i),
     (fun f -> rnd := Big_int.add_int_big_int 90 !rnd;
       !s_m_c_float f),
     (fun f ->  !s_m_n_float f),
     (fun () -> rnd:= Big_int.unit_big_int),
     (fun a -> a.randominfo <- (if Big_int.eq_big_int !rnd Big_int.unit_big_int 
				then None
				else Some !rnd)),
     (fun () -> Random2.State.copy countstate),
     (fun () -> Random2.State.copy nocountstate),
     (fun s -> Random2.State.assign countstate s),
     (fun s -> Random2.State.assign nocountstate s),
     (fun () ->
	Random2.State.assign countstate (Random2.get_state ());
	Random2.State.assign nocountstate (Random2.get_state ())
     ),
     (fun () -> s_m_n := neutralrandom; s_m_c := neutralrandom; s_m_n_float := neutralfloatrandom; s_m_c_float := neutralfloatrandom),
     (fun () -> s_m_n := std_my_nocount; s_m_c := std_my_count; s_m_n_float := std_my_nocount_float; s_m_c_float := std_my_count_float)
    )


let is_sparse ac = match ac.transitionrule with
    Sparse _ -> true
  | _ -> false

(* let colormap_from_string s ac = *)
(*   try  *)
(*     List.assoc s ac.parameters.colormaps *)
(*   with Not_found -> Auto *)


let rec action_to_string = function
    Newall -> "Newall"
  | Newac -> "Neawac"
  | Newconf -> "Newconf"
  | NeitherNor -> "Newother"
  | Blocked x -> "Blocked "^(action_to_string x)

let seed_ordinal_zero_to_string s =
  let (evaltab,evali) = Random2.explicit_get_state s.rs_rule in
  let (runtab,runi) = Random2.explicit_get_state s.rs_conf in
    Printf.sprintf "{\n\tKernel.rs_rule = Random2.make_state [|%s|] %d; \n\tKernel.rs_conf = Random2.make_state [|%s|] %d; \n\tKernel.ordinal = 0\n}"
      (Array.fold_left (fun acc i -> acc^(Printf.sprintf "%d;" i)) "" evaltab)  evali
      (Array.fold_left (fun acc i -> acc^(Printf.sprintf "%d;" i)) "" runtab)  runi

(********************************************************************************)
(** PARAMETERS **)
(********************************************************************************)



let showbool b = if b then "true" else "false"
let showstrlist l = "["^(List.fold_left (fun acc e -> acc^"\""^e^"\";") "" l)^"]"

let float_to_caml f =
  let (m,e) = frexp f in
  let sm = Printf.sprintf "%.64f" m in
  Printf.sprintf "(ldexp %s (%d))" sm e

let param_to_string (s,par) =
  match par with
      ParInt p ->  Printf.sprintf "(\"%s\", Kernel.ParInt {Kernel.pdefault= %d; Kernel.pmin= %d; Kernel.pmax= %d; Kernel.pvalue= %d})" s p.pdefault p.pmin p.pmax p.pvalue
    | ParFloat p ->  Printf.sprintf "(\"%s\", Kernel.ParFloat {Kernel.pfdefault= %s; Kernel.pfmin= %s; Kernel.pfmax= %s; Kernel.pfvalue= %s})" 
      s (float_to_caml p.pfdefault) (float_to_caml p.pfmin) (float_to_caml p.pfmax) (float_to_caml p.pfvalue)
    | ParBool p ->  Printf.sprintf "(\"%s\", Kernel.ParBool {Kernel.pbdefault= %s; Kernel.pbvalue= %s})" s (showbool p.pbdefault) (showbool p.pbvalue)
    | ParString p ->  Printf.sprintf "(\"%s\", Kernel.ParString {Kernel.psdefault= \"%s\"; Kernel.psvalue= \"%s\"})" s p.psdefault p.psvalue
    | ParChoice p ->  Printf.sprintf "(\"%s\", Kernel.ParChoice {Kernel.pchoices= %s; Kernel.pcvalue= \"%s\"})" s (showstrlist p.pchoices) p.pcvalue

let universe_to_string (x,y) =
  Printf.sprintf "(%d,%d)" x y

let show_param x  = 
  let (name,e) = x in
    (match e with
	ParInt par -> debugline Dbgdev (Printf.sprintf "int param '%s'=%d (%d) %d < %d" 
					name par.pvalue par.pdefault par.pmin par.pmax)
      | ParFloat par -> debugline Dbgdev (Printf.sprintf "float param '%s'=%f (%f) %f < %f" 
					name par.pfvalue par.pfdefault par.pfmin par.pfmax)
      | ParBool par -> debugline Dbgdev (Printf.sprintf "bool param '%s'=%s (%s)" 
					name (showbool par.pbvalue)
					(showbool par.pbdefault))
      | ParString par -> debugline Dbgdev (Printf.sprintf "string param '%s'=%s (%s)" 
					name par.psvalue par.psdefault)
      | ParChoice par -> debugline Dbgdev (Printf.sprintf "choice param '%s'=%s %s" 
					   name par.pcvalue (showstrlist par.pchoices))
    ); x


let param_is_file a name =
  let rec doit = function
    [] -> false
    | t::q -> (fst t = name && snd t = Filechooser) || doit q
  in 
    doit a.parameters.metaparams

let param_is_text a name =
  let rec doit = function
    [] -> false
    | t::q -> (fst t = name && snd t = Multiline) || doit q
  in 
    doit a.parameters.metaparams

let doc_about a name =
  let rec doit = function
    [] -> None
    | (n,Doc s)::q -> if  n = name then Some s else doit q
    | t::q -> doit q
  in 
    doit a.parameters.metaparams

(* Domain names for plugin parameters: a parameter name inside a
   plugin is prefixed by the plugin name to avoid name collision with
   regular aprameters. This is made transparent to the user via a
   global "domain name" variable. 
*)
let (enter_domain_name, exit_domain_name, current_domain_name, current_domain_prefix, clean_name) =
  let domain = ref "" in
  let prefix_from_domain s = if s="" then "" else s^"::" in
    ((fun name -> neutralize_myrandom (); domain := name),
     (fun () -> std_myrandom (); if !domain <> "" then (
	  domain := "")
     ),
     (fun () -> !domain),
     (fun () -> prefix_from_domain !domain),
     (fun name -> let reg = Str.regexp "\\(.*\\)::" in
	Str.global_replace reg "<i><small>(\\1) </small></i>" name
     )
    )
      
(* Each operation on parameters using a reference name must use the
   following name conversion to properly handle domain names *)
let real_param_name x = (current_domain_prefix ())^x
  

let (param_alive_list, param_is_alive, reset_param_alive_list) =
  let lili = ref [] in
    ((fun () -> !lili), 
     (fun p -> lili:= p::(!lili)),
     (fun () -> lili := [])
    )

let clear_param ac name =
  let name = real_param_name name in
  let rec doit = function
      [] -> []
    | (a,b)::q -> if a = name then doit q else (a,b)::(doit q)
  in ac.parameters.params <- doit ac.parameters.params

let filter_params ac filter =
  let rec doit = function
      [] -> []
    | (a,b)::q -> if filter a then (a,b)::(doit q) else doit q
  in ac.parameters.params <- doit ac.parameters.params

let param_bool ac name x =
  let name = real_param_name name in
    try 
      param_is_alive name;
      let pa = List.assoc name ac.parameters.params in
	match pa with
	    ParBool e -> ( (* Parameter exists with correct type so do not override user set value *)
	      e.pbdefault <- x;
	      debugline Dbgdev (Printf.sprintf "param %s=%s" name (showbool e.pbvalue));
	      e.pbvalue)
	  | ParString e -> (* Try conversion from string *)
	      (
		try
		  let v = bool_of_string e.psvalue in
		    debugline Dbgdev ("Param '"^name^"' existed before");
		    ac.parameters.params <- (show_param (name, 
							 ParBool {pbvalue = v; 
								  pbdefault = x}))::(List.remove_assoc name ac.parameters.params);
		    debugline Dbgdev ("Param '"^name^"' converted to bool");
		    v
		with _ -> ac.parameters.params <- List.remove_assoc name ac.parameters.params; raise Not_found
	      )
 	  |_ -> (* Parameter with incorrect type so override *) raise Not_found
    with Not_found -> (
      ac.parameters.params <- (show_param (name, ParBool {pbdefault = x; pbvalue = x}))::(ac.parameters.params);
      x
    )

let param_choice ac ?choice name x =
  let name = real_param_name name in
  try 
    param_is_alive name;
    let pa = List.assoc name ac.parameters.params in
      match pa with
	  ParChoice e -> ( (* Parameter exists with correct type so do not override user set value *)
	    e.pchoices <- x;
	    debugline Dbgdev (Printf.sprintf "param %s='%s'" name e.pcvalue);
	    e.pcvalue)
	| ParString e -> (* Try conversion from string *)
	    if List.mem e.psvalue x then (
	      debugline Dbgdev ("Param '"^name^"' existed before");
	      ac.parameters.params <- (show_param (name, ParChoice {pcvalue = e.psvalue; 
 								    pchoices = x}))::(List.remove_assoc name ac.parameters.params);
	      debugline Dbgdev ("Param '"^name^"' converted to choice");
	      e.psvalue
	    ) else (debugline Dbgdev ("Overriding parameter '"^name^"'"); raise Not_found)
	      
	|_ -> (* Parameter with incorrect type so override *) 
	   ac.parameters.params <- List.remove_assoc name ac.parameters.params; raise Not_found
  with Not_found -> 
        let defaultval = (match x with [] -> "" | t::q -> t) in
        let valu = (match choice with Some c -> (if List.mem c x then c else defaultval) | None -> defaultval) in 
      (
	ac.parameters.params <- (show_param (name, ParChoice {pchoices = x; 
							      pcvalue = valu}))::(ac.parameters.params);
	valu
      )

let param_set_text ac name =
  let name = real_param_name name in
  ac.parameters.metaparams <- (name, Multiline)::ac.parameters.metaparams

let param_string is_file ac name x =
  let name = real_param_name name in
  try 
    param_is_alive name;
    let pa = List.assoc name ac.parameters.params in
      match pa with
	  ParString e -> ( 
	    (* Parameter exists with correct type so do not override user set value *)
	    e.psdefault <- x;
	    debugline Dbgdev (Printf.sprintf "param %s='%s'" name e.psvalue);
	    (* however put the flag because it is erased at each eval *)
	    if is_file then ac.parameters.metaparams <- (name, Filechooser)::ac.parameters.metaparams;
	    e.psvalue
	  )
 	|_ -> (* Parameter with incorrect type so override *) 
	   ac.parameters.params <- List.remove_assoc name ac.parameters.params; raise Not_found
  with Not_found -> (
    ac.parameters.params <- (show_param (name, ParString {psdefault = x; psvalue = x}))::(ac.parameters.params);
    if is_file then ac.parameters.metaparams <- (name, Filechooser)::ac.parameters.metaparams;
    x
  )

let param_float ac name mi ma x =
  let name = real_param_name name in
  try 
    param_is_alive name;
    let pa = List.assoc name ac.parameters.params in
      match pa with
	  ParFloat e -> ( (* Parameter exists with correct type so do not override user set value *)
	    e.pfdefault <- x;
	    e.pfmin <- mi;
	    e.pfmax <- ma;
	    debugline Dbgdev (Printf.sprintf "param %s=%f (%f) %f < %f" name e.pfvalue x mi ma);
	    max (min e.pfvalue ma) mi)
	| ParString e -> (* Try conversion from string *)
	    (
	      try
		debugline Dbgdev ("Param '"^name^"' existed before");
		let v = float_of_string e.psvalue in
		  ac.parameters.params <- (show_param (name, ParFloat {pfvalue = max (min v ma) mi; 
					     pfdefault = x; pfmin = mi; pfmax = ma}))::(List.remove_assoc name ac.parameters.params);
		  debugline Dbgdev ("Param '"^name^"' converted to float");
		  max (min v ma) mi
	      with _ -> raise Not_found
	    )
	|_ -> (* Parameter with incorrect type so override *) 
	   ac.parameters.params <- List.remove_assoc name ac.parameters.params; raise Not_found	    
  with Not_found -> (
    ac.parameters.params <-   (show_param (name, 
					   ParFloat {pfvalue = x; pfdefault = x; pfmin = mi; pfmax = ma}))::(ac.parameters.params);
    x
  )

let param ac name mi ma x =
  let name = real_param_name name in
  try 
    param_is_alive name;
    let pa = List.assoc name ac.parameters.params in
      match pa with
	  ParInt e -> ( (* Parameter exists with correct type so do not override user set value *)
	    e.pdefault <- x;
	    e.pmin <- mi;
	    e.pmax <- ma;
	    debugline Dbgdev (Printf.sprintf "param %s=%d (%d) %d < %d" name e.pvalue x mi ma);
	    max (min e.pvalue ma) mi)
	| ParString e -> (* Try conversion from string *)
	    (
	      try
		debugline Dbgdev ("Param '"^name^"' existed before");
		let v = int_of_string e.psvalue in
		  ac.parameters.params <- (show_param (name, ParInt {pvalue = max (min v ma) mi; 
					     pdefault = x; pmin = mi; pmax = ma}))::(List.remove_assoc name ac.parameters.params);
		  debugline Dbgdev ("Param '"^name^"' converted to int");
		  max (min v ma) mi
	      with _ -> raise Not_found
	    )
	|_ -> (* Parameter with incorrect type so override *) 
	   ac.parameters.params <- List.remove_assoc name ac.parameters.params; raise Not_found	    
  with Not_found -> (
    ac.parameters.params <-   (show_param (name, 
					   ParInt {pvalue = x; pdefault = x; pmin = mi; pmax = ma}))::(ac.parameters.params);
    x
  )

let get_param ac name =
  let name = real_param_name name in
  match List.assoc name ac.parameters.params with
      ParInt e -> e.pvalue
    | _ -> raise (Badparam name)

let get_float_param ac name =
  let name = real_param_name name in
  match List.assoc name ac.parameters.params with
      ParFloat e -> e.pfvalue
    | _ -> raise (Badparam name)

let get_bool_param ac name =
  let name = real_param_name name in
  match List.assoc name ac.parameters.params with
      ParBool e -> e.pbvalue
    | _ -> raise (Badparam name)
	    
let get_string_param ac name =
  let name = real_param_name name in
  match List.assoc name ac.parameters.params with
      ParString e -> e.psvalue
    | ParChoice e -> e.pcvalue
    | _ -> raise (Badparam name)
	    

let set_param ac name valu =
  let name = real_param_name name in
  try
    let pa = List.assoc name ac.parameters.params in
      match pa with
	  ParInt e -> e.pvalue <- valu
	    (* Parameter exists with incorrect type *)
	| _ -> raise (Badparam name)
	    
  with Not_found -> raise (Unknownparam name)

let set_float_param ac name valu =
  let name = real_param_name name in
  try
    let pa = List.assoc name ac.parameters.params in
      match pa with
	  ParFloat e -> e.pfvalue <- valu
	    (* Parameter exists with incorrect type *)
	| _ -> raise (Badparam name)
	    
  with Not_found -> raise (Unknownparam name)

let set_bool_param ac name valu =
  let name = real_param_name name in
  try
    let pa = List.assoc name ac.parameters.params in
      match pa with
	  ParBool e -> e.pbvalue <- valu
	    (* Parameter exists with incorrect type *)
	| _ -> raise (Badparam name)
	    
  with Not_found -> raise (Unknownparam name)
    
let set_string_param ac name valu =
  let name = real_param_name name in
  try
    let pa = List.assoc name ac.parameters.params in
      match pa with
	  ParString e -> e.psvalue <- valu
	    (* Parameter exists with incorrect type *)
	| _ -> raise (Badparam name)
	    
  with Not_found -> raise (Unknownparam name)

let set_choice_param ac name valu =
  let name = real_param_name name in
  try
    let pa = List.assoc name ac.parameters.params in
      match pa with
	  ParChoice e -> if List.mem valu e.pchoices then e.pcvalue <- valu
	    (* Parameter exists with incorrect type *)
	| _ -> raise (Badparam name)
	    
  with Not_found -> raise (Unknownparam name)
    
    
    
(* By default, all params given on command line are interpreted as ParString and conversions
   are done (if possible) when param is created in source file *)
let rec inject_params ac = function
    [] -> ()
  | t::q -> let u = snd t in
	    let cln = List.remove_assoc (fst t) ac.parameters.params in
	    ac.parameters.params <- (show_param ((fst t), (ParString {psdefault = u; psvalue = u})))::cln;
	    inject_params ac q
	
let (is_main_program, declare_main_program, declare_co_program) =
  let cmp = ref true in
    ((fun () -> !cmp),
     (fun () -> cmp := true),
     (fun () -> cmp := false))

let myargnb () = if is_main_program () then raise (Unimplemented "myarg")
else Array.length Sys.argv

let mygetarg n = if is_main_program () then raise (Unimplemented "myarg")
else Sys.argv.(n)

let putrandomparams ac =
  let singler p =
    let diff = p.pmax - p.pmin + 1 in
    let nval = neutralrandom diff + p.pmin in
      p.pvalue <- nval
  in
  let do_only_int = function
      (_, ParInt a) -> singler a
    | _ -> ()
  in
    List.iter do_only_int ac.parameters.params

let (set_mem_effort, get_mem_effort) =
  let size = ref (1 lsl 24) in
    (fun i -> size := max 0 i), (fun () -> !size)

let secure_intexp ?limit mem_flag b e = 
  let thepremax = if mem_flag then get_mem_effort () else max_int in
  let themax = match limit with
      None -> thepremax | Some l -> (min l thepremax)
  in
    if e > Sys.word_size then raise (RuleTooLarge (Big_int.big_int_of_int 0)) else
      let res = Big_int.power_big_int_positive_big_int (Big_int.big_int_of_int b) (Big_int.big_int_of_int e) in
	if Big_int.gt_big_int res (Big_int.big_int_of_int themax) then raise (RuleTooLarge res)
	else Big_int.int_of_big_int res
      
let intsqrt n =
  int_of_float (sqrt (float_of_int n))

let arity a =
  match a.neighbourhood with
      Moore n -> if a.dimension = 1 then (2*n+1) else (2*n+1)*(2*n+1)
    | VonNeumann n -> if a.dimension=1 then 2*n+1 else 2*n*n+2*n+1
    | Cross n -> 2*n*a.dimension+1

(************* INFO DISPLAY FUNCTIONS *********************)


let humanbiginttostring n  =
  let level n = Big_int.power_int_positive_int 10 (3*n) in
    if Big_int.ge_big_int n (level 5) 
    then (Num.approx_num_exp 3 (Num.Big_int n))
    else if Big_int.ge_big_int n (level 4) 
    then (Big_int.string_of_big_int (Big_int.div_big_int n (level 4)))^"T"
    else if Big_int.ge_big_int n (level 3) 
    then (Big_int.string_of_big_int (Big_int.div_big_int n (level 3)))^"G"
    else if Big_int.ge_big_int n (level 2) 
    then (Big_int.string_of_big_int (Big_int.div_big_int n (level 2)))^"M"
    else if Big_int.ge_big_int n (level 1) 
    then (Big_int.string_of_big_int (Big_int.div_big_int n (level 1)))^"K"
    else (Big_int.string_of_big_int n)

let infointtostring n  =
  if n >= 1024*1024 then (string_of_int (n/(1024*1024)))^"MB"
  else if n >= 1024 then (string_of_int (n/(1024)))^"KB"
  else (string_of_int n)^"B"

let infointnbtostring n  =
  if n >= 1024*1024 then (string_of_int (n/(1024*1024)))^"M"
  else if n >= 1024 then (string_of_int (n/(1024)))^"K"
  else (string_of_int n)


let multi_size virt a = 
  let rec doit = function
      Table rv -> fast_array_memsize rv
    | XTable rv -> bytes_per_word*(Array.length rv) (* Bigarray.Array1.dim rv *)
    | MultisetTable rv -> bytes_per_word*(Bigarray.Array1.dim rv)
    | FunArray _ | FunFun _-> (* if virt then humanbiginttostring (Big_int.power_int_positive_int a.cardinality (arity a)) else *) 0
    | MultisetFunArray _ | MultisetFunFun _ -> (* if virt then humanbiginttostring (Big_int.power_int_positive_int (arity a) a.cardinality) else *) 0
	(*     | Selfproduct x -> doit x *)
    | Sparse (x,t) -> doit t
    | _ -> raise (Unimplemented "rawsize")
  in
    doit a.transitionrule
    

let rawsize a = let s = multi_size true a in humanbiginttostring (Big_int.big_int_of_int s)
let memsize a = a.meminfo
let correct_meminfo a = a.meminfo <- !(a.refmeminfo)

let rndratio a =
  match a.transitionrule with
      FunArray _ | FunFun _ | MultisetFunArray _ | MultisetFunFun _ -> "unknown complexity"
    | _ -> (
	match a.randominfo with
	    None -> "single rule"
	  | Some x -> ("complexity is O("^(humanbiginttostring x)^")")
      )
	
let neighbinfo ac = 
    match (ac.neighbourhood) with
	Moore n -> "Moore("^(string_of_int n)^")"
      | VonNeumann n -> "vonNeumann("^(string_of_int n)^")"
      | Cross n -> "cross("^(string_of_int n)^")"

let ruleinfo ac =
  let power_info = match ac.binary_power with
      Vcard c -> (Printf.sprintf " [c%d]" c)
    | Varit k -> (Printf.sprintf " [a%d]" k)
    | Vboth (c,k) -> (Printf.sprintf " [c%da%d]" c k)
    | _ -> ""
  in
  let rec doit = function 
      Table t -> fast_array_type t
    | XTable _ -> "fast table"
    | MultisetTable _ -> "table (multiset)"
    | FunArray _ -> "function (array neighb.)"
    | FunFun _ -> "function (fun neighb.)"
    | MultisetFunArray _ -> "function (multiset array)"
    | MultisetFunFun _ -> "function (multiset fun)"
    | Sparse (_,r) -> ("sparse update upon "^(doit r))
    | _ -> "other"
  in (doit ac.transitionrule)^power_info

let typeinfo ac = 
  match ac.wanted with
      Newconf -> "init"
    | Newall -> "all"
    | Newac -> "ca"
    | NeitherNor -> "other"
    | Blocked x -> "blocked"

let info_mat = function
    (Bidim a, Bidim b) -> fast_matrix_type a
  | _ -> "?"

let shortacinfo mat a =
  "dim="^(string_of_int a.dimension)^
    (if a.dimension=1 then (Printf.sprintf " (%d cells)" (fst a.universe)) 
     else (Printf.sprintf " (%dx%d cells, %s)" (fst a.universe) (snd a.universe) (info_mat mat)))
  ^", card="^(string_of_int a.cardinality)^", ruletype="^(ruleinfo a)^", new="^(typeinfo a)^
    ", neighb="^(neighbinfo a)^", mem cost="
  ^(infointtostring (memsize a))^", "^(rndratio a)

let listacinfo a =
  [(string_of_int a.dimension)^"D"(* ^(if a.dimension=1 then (Printf.sprintf "%d cells" (fst a.universe))  *)
(* 					  else (Printf.sprintf "%dx%d cells" (fst a.universe) (snd a.universe)))^")" *);
   (string_of_int a.cardinality)^" states";
   (neighbinfo a);
(*    (ruleinfo a); *)
(*    (rawsize a); *)
(*    (rndratio a) *)
  ]
    
let paramval_to_string = function
    ParInt p -> string_of_int p.pvalue
  | ParFloat p -> string_of_float p.pfvalue
  | ParBool p -> showbool p.pbvalue
  | ParChoice p -> p.pcvalue
  | ParString p -> p.psvalue

let coreacinfo outputmethod mat a =
  let rec params = function
      [] -> ""
    | (n,p)::q -> ", "^n^"="^(paramval_to_string p)^(params q) in    
    outputmethod ((shortacinfo mat a)^(params a.parameters.params))




(* This must be in accordance with nextconf *)
let centercell a =
  match a.neighbourhood with
      Moore n -> if a.dimension = 1 then n else (2*n+1)*n+n
    | VonNeumann n -> n
    | Cross n -> n

(****** MUST MATCH WITH NEXTCONF ALGORITHMS *************)

let relativecell a x y =
  let bad () = debugline Dbgwarning "Local rule tries to access a cell outside neighbourhood!" in 
    if a.dimension = 1 && y<>0 then (bad (); 0) else (
      match a.neighbourhood with
	  Moore n -> if abs x <= n && abs y <= n then (
	    if a.dimension  = 2 then (2*n+1)*(n+x)+n+y 
	    else x+n
	  )
	  else (bad (); 0)
	| VonNeumann n ->
	    if y = 0 && abs x <= n then n+x 
	    else if y < 0 && y >= -n && abs x <= n + y then 2*n+1 + (n+y)*(n+y) + n+y+x
	    else if y > 0 && y <= n && abs x <= n - y then 2*n+1 + n*n + (y-1)*(3*n-y-n+1)+(x+n-y)
	      (* 2*n+1 + 2*(n-1)*(n-1) - (n-y)*(n-y) + n-y+x *)
	    else (bad (); 0)
	| Cross n -> 
	    if y = 0 && abs x <= n then n+x 
	    else if x = 0 && y >= (-n) && y < 0 then 2*n+1+n+y
	    else if x = 0 && y>0 && y <=n then 2*n+n+y
	    else (bad (); 0)
    )

let neighbors a =
  let rec h_line k acc l =
    if k=0 then (0,l)::acc else h_line (k-1) (((-k),l)::((k,l)::acc)) l
  in
  let rec repeat acc p f = if p = 0 then (f acc 0) else repeat (f acc p) (p-1) f
  in
  match a.neighbourhood with
      Moore n -> if a.dimension = 1 then h_line n [] 0
	else repeat (h_line n [] 0) (n-1) (fun acc i -> h_line n (h_line n acc (i+1)) (-1-i))
    | VonNeumann n -> if a.dimension = 1 then h_line n [] 0
      else repeat [] n (fun acc i -> h_line (n-i) (if i=0 then acc else h_line (n-i) acc (-i)) i)
    | Cross n -> if a.dimension = 1 then h_line n [] 0
      else repeat (h_line n [] 0) (n-1) (fun acc i -> (0,(i+1))::((0,(-i-1))::acc))
      
let freshbigarray kind dim elt = 
  let x = Bigarray.Array1.create 
    kind Bigarray.c_layout dim in
    Bigarray.Array1.fill x elt;
    x

let lastrawindex a =
  let c = a.cardinality in
    match (a.transitionrule, a.binary_power) with
	(Table _, _)  -> let cs = secure_intexp true c (arity a) in cs - 1
    | (MultisetTable _, Varit k)  -> let cs = secure_intexp true (1 lsl k) c in cs - 1
    | (MultisetTable _, Vboth (_,k))  -> let cs = secure_intexp true (1 lsl k) c in cs - 1
    | (XTable _, Vcard c)  -> let cs = secure_intexp true (1 lsl c) (arity a) in cs - 1
    | (XTable _, Vboth (c,_))  -> let cs = secure_intexp true (1 lsl c) (arity a) in cs - 1
(*     | Selfproduct x -> doit c x *)
    | _ -> raise (Unimplemented "lastrawindex") 
    
let getradius ac =
  match ac.neighbourhood with
      Moore n -> n
    | VonNeumann n -> n
    | Cross n -> n

let getdimension ac = ac.dimension

(* destructive function *)
let calibrate_table a =
  let rec doit c = function
    | Table _  -> 
	(* first check the worst case *)
	let cs = secure_intexp true c (arity a) in 
	  debugline Dbgdev (Printf.sprintf "slow table OK (%d)" cs);
	  (* now try power-of-2 optimization *)
	  (try 
	     let ns = secure_intexp ~limit:Sys.max_array_length true (round_power_two c) (arity a) in
	       (* For now, only Moore(1) neighb. is accepted *)
	       if a.neighbourhood <> Moore 1 && a.dimension > 1 then raise (RuleTooLarge (Big_int.big_int_of_int 3));
	       a.binary_power <- Vcard (exp_round_power_two (c-1));
	       debugline Dbgdev (Printf.sprintf "fast table OK (%d)" ns); 
	       XTable (Array.make ns 0) 
	   with RuleTooLarge _ -> Table (make_fast_array cs (c-1)))
    | MultisetTable _ -> (
	(* first check whether table rule could be OK *)
	(
	  (* first we check worst case (no power-of-2 optimization) *)
	  ignore (secure_intexp false c c);
	  a.binary_power <- QueDalle;
	  let base = secure_intexp true (arity a + 1) c in
	    (* Now we try the power-of-two rounding *)
	  let cs = try 
	    debugline Dbgdev (Printf.sprintf "slow multiset-table OK (%d)" base);
	    let r = secure_intexp true (round_power_two (arity a + 1)) c in
	      (* now try to round-up cardinality also *)
	      (try 
		 let vc = round_power_two (c-1) in
		   ignore (secure_intexp false vc vc);
		   a.binary_power <- Vboth ((exp_round_power_two (c-1)),(exp_round_power_two (arity a)))
	       with RuleTooLarge _ -> a.binary_power <- Varit (exp_round_power_two (arity a)));
	      debugline Dbgdev (Printf.sprintf "fast multiset-table OK (%d)" r); r
	  with RuleTooLarge _ -> base
	  in MultisetTable (freshbigarray Bigarray.int cs 0)
	)
      )
    | _ -> internalbug "calibrate table on non-table pre-rule"
  in (
      (* if a.cardinality > Sys.max_array_length then raise (AbortRule "too many states"); *)
      a.transitionrule <- doit a.cardinality a.transitionrule
    )


let defaultcells () = (200,200)

let inttointarray ac x i =
  let (length,packet) = match (ac.transitionrule, ac.binary_power) with
      (Table _, _) -> (arity ac, ac.cardinality)
    | (XTable _, Vcard c) -> (arity ac, 1 lsl c)
    | (XTable _, Vboth (c,_)) -> (arity ac, 1 lsl c)
    | (XTable _, _) -> (arity ac, ac.cardinality)
    | (MultisetTable _, Varit k) -> (ac.cardinality, 1 lsl k)
    | (MultisetTable _, Vboth (_,k)) -> (ac.cardinality, 1 lsl k)
    | (MultisetTable _, _) -> (ac.cardinality, (arity ac +1))
    | _ -> internalbug "inttointarray"
  in
  let tmp = ref i in
    for t = 0 to length-1 do
      x.(t) <- !tmp mod packet;
      tmp := !tmp / packet
    done
      

let ml_source_action_name = "ML source file"
let wysiwyg_png_action_name = "WYSIWYG as PNG"

let defaultactions = [(ml_source_action_name, Source) ; (wysiwyg_png_action_name, Wysiwyg)]
let fixeddefaultaction = [(SrcSave, ml_source_action_name) ; (SrcSnapshot, wysiwyg_png_action_name)]

let defaultparamschem () = 
  {
    params = [];
    actions = defaultactions;
    metaparams = [];
    colormaps = thecolormaps;
    statesnames = None;
  }

let defaultinfodraw () = {under = []; over = []}

let defaultmatrice () = (Zerodim (ref 0), Zerodim (ref 0))

let defaultseed () = let s = Random2.get_state () in 
  {
    rs_rule = Random2.State.copy s;
    rs_conf = Random2.State.copy s;
    ordinal = 0;
  }

let defaultcolorcache () = Array.make 1 Int32.zero

let defaulttablerule () = Table (make_fast_array 8 1)
let defaultmultisettablerule () = MultisetTable (freshbigarray Bigarray.int 8 0)

let defaultrule = defaulttablerule

let defaultac () = 
  { 
    seed = defaultseed ();
    undo_seed = defaultseed ();
    transitionrule = defaultrule ();
    dimension = 1;
    universe = defaultcells ();
    cardinality = 1;
    neighbourhood = Moore 1;
    initconfs = ["Random uniform", Random(Uniform)];
    theinitconf = "Random uniform";
    spots = SpotSet.empty;
    modified = SpotSet.empty;
    complexconf = true;
    colormap = fst (List.hd thecolormaps);
    name = None;
    nick_name = None;
    obj_name = None;
    radius = None;
    parameters = defaultparamschem ();
    action = fixeddefaultaction;
    randominfo = None;
    infotodraw = defaultinfodraw ();
    colorcache = defaultcolorcache ();
    version = theversion;
    srcfile = None;
    dbg = [];
    description = [];
    wanted = Newall;
    determinization = Deterministic;
    binary_power = QueDalle;
    meminfo = 0;
    refmeminfo = ref 0;
  }

let defaultac_nocompute () =
  let a = defaultac () in
    a.wanted <- NeitherNor;
    a

let allow_cache ac = match ac.determinization with
    Stochastic -> false
  | _ -> true

let allow_table ac = match ac.determinization with
    Stochastic | DeterministicButSlow -> false
  | _ -> true

let dont_memorize ac = match ac.determinization with
    WantToBeDeterministic -> false
  | _ -> true

let initial ac = (ac.seed.ordinal = 0)

let makeinitial ac = (
  ac.seed.ordinal <- 0;
  if needresetallparameters ac then ac.parameters <- defaultparamschem ()
  else debugline Dbgdev "Parameters not erased.";
  (* should be useless but kept until initconf save mechanism fixed *)
(*   if needresetconf ac then  ac.theinitconf <- "Random uniform"; *)
  if needresetactions ac then (
    ac.action <- fixeddefaultaction;
    debugline Dbgdev "Reset colormap";
    ac.colormap <- fst (List.hd thecolormaps);
  );
  match ac.transitionrule with
      Table _ | MultisetTable _ | XTable _-> ()
    | _ -> (ac.transitionrule <- defaultrule ())
)
  
(* In case of new ac, some initconfs might become invalid *)
let resetinitconf ac =
  debugline Dbgdev "Resetting init configurations";
  ac.initconfs <- ["Random uniform", Random(Uniform)]

let resetmisc ac =
  (
    debugline Dbgdev "Resetting misc components of the object";
    ac.parameters.metaparams <- [];
    ac.parameters.actions <- defaultactions;
    ac.description <- [];
  (* colormaps must be non-empty *)
    if neednewac_unblocked ac then (
      ac.obj_name <- None;
      ac.name <- None;
      debugline Dbgdev "resetting colormaps";
      ac.parameters.colormaps <- thecolormaps;
    ) 
  )
    
(* Reset some fields before making a new automaton *)
let pseudoreset ac =
  ac.determinization <- Deterministic;
  ac.complexconf <- true;
  if needresetactions ac then resetmisc ac;
  if needresetconf ac then resetinitconf ac;
  ac.binary_power <- QueDalle;
  ac.srcfile <- None;
  if needresetdescription ac then ac.description <- [];
  ac

let fakeac i = Acompletion (i mod 100)

let fakeacerror s =  Acerror s


let getunderdrawings a = a.infotodraw.under

let getoverdrawings a = a.infotodraw.over

let setoverdrawings a o = a.infotodraw.over <- o




(**********************************************)
(* Matrix ACCESS                              *)
(**********************************************)

exception NoValidMatrix

let getcurrentunidimmatrix mat =
  match mat with
     (Unidim matrix,Zerodim ts) -> (matrix,!ts)
   | _ -> raise NoValidMatrix


let getoldbidimmatrix ac mat curt =
  match mat with 
      (Bidim redmatrix, Bidim bluematrix) -> 
	if is_sparse ac then redmatrix
	else if curt mod 2 <> 0 then redmatrix else bluematrix
    | _ -> raise NoValidMatrix


let getcurrentbidimmatrix ac mat curt =
  match mat with 
      (Bidim redmatrix, Bidim bluematrix) -> 
	if is_sparse ac then redmatrix
	else if curt mod 2 = 0 then redmatrix else bluematrix
    | _ -> raise NoValidMatrix
	

(* (\* Add neighbouring cells in spots collection *\) *)
(* let enlarge_spot_2d r cells tott sp =  *)
(*   let (x,y) = cells in *)
(*   let addbipos a b = function *)
(*     | Plane (u,v) -> (u+a,v+b) *)
(*     | _ -> internalbug "add bidim position" *)
(*   in *)
(*   let lili =  *)
(*     let rec encore acc x y = if x = 0 then ( *)
(*       if y = 0 then  (0,0)::acc else encore ((x,-y)::((-x,-y)::((x,y)::((-x,y)::acc)))) r (y-1) *)
(*     ) *)
(*     else encore ((x,-y)::((-x,-y)::((x,y)::((-x,y)::acc)))) (x-1) y in *)
(*       encore [] r r in *)
(*   let valid_pos (i,j) = Plane ((x+2*r*tott+i) mod (x+2*r*tott), (y+2*r*tott+j) mod (y+2*r*tott)) in *)
(*     List.fold_left (fun acc (i,j) -> SpotSet.add (valid_pos (addbipos i j sp)) acc) (SpotSet.empty) lili *)
  
(* let enlarge_spot_1d r cells tott sp =  *)
(*   let (x,y) = cells in *)
(*   let addunipos u = function *)
(*     | Line b -> u+b *)
(*     | _ -> internalbug "add unidim position" in *)
(*   let lili =  *)
(*     let rec encore acc x = if x = 0 then 0::acc else encore (x::((-x)::acc)) (x-1) in  *)
(*       encore [] r in *)
(*   let valid_pos i = Line ((x+2*r*tott+i) mod (x+2*r*tott)) in *)
(*     List.fold_left (fun acc i -> SpotSet.add (valid_pos (addunipos i sp)) acc) (SpotSet.empty) lili *)
  


let universe_reader ac mice curt cells steps =
  let (x,y) = cells in
    try
      (match ac.dimension with
	   2 -> 
	     let m = getcurrentbidimmatrix ac mice curt in 
	       getter_from_fast_matrix m
	 | 1 -> 
	     let r = getradius ac in
	     let (thearray,timeshift) = getcurrentunidimmatrix mice in
	       (fun i j -> thearray.(i+steps*r).((y+j+timeshift) mod y))
	 | _ -> raise NoValidMatrix
      ) 
    with NoValidMatrix -> (fun i j -> 0)
	

(* Must take care of Sparse rules *)
let universe_writer ac mice curt cells steps =
  let (x,y) = cells in
    try
      (match ac.dimension with
	   2 -> let m = getcurrentbidimmatrix ac mice curt in
	     (match ac.transitionrule with
		  Sparse _ -> 
		    (fun i j c -> setter_from_fast_matrix m i j c; ac.spots <- SpotSet.add (Plane (i,j)) ac.spots)
		| _  -> setter_from_fast_matrix m
	     )
	       (* In dimension 1, the writer can only access the current conf. so second coordinate is ignored *)
	 | 1 -> (
 	     let r = getradius ac in
	     let (thearray,timeshift) = getcurrentunidimmatrix mice in
	       (match ac.transitionrule with
		    Sparse _ -> 
		      (fun i j c -> thearray.(i+steps*r).((timeshift) mod y) <- c; ac.spots <- SpotSet.add (Line i) ac.spots)
		  | _ -> (fun i j c -> thearray.(i+steps*r).((timeshift) mod y) <- c)
	       )
	   )
	 | _ -> raise NoValidMatrix
      )
    with NoValidMatrix -> (fun i j k -> ())

let do_single measure ac mat curt steps =
  let reader = universe_reader ac mat curt ac.universe steps in
  let writer = universe_writer ac mat curt ac.universe steps in
  let x,y = ac.universe in
    measure curt x y reader writer

let iterate_on_matrix ac mice curt cells steps howto =
  let (x,y) = cells in
    (match ac.dimension with
	 2 -> let reader = getter_from_fast_matrix (getcurrentbidimmatrix ac mice curt) in
	   howto x y reader
       | 1 -> (
	   let r = getradius ac in
	   let (thearray,timeshift) = getcurrentunidimmatrix mice in
	   let nmat = Array.make_matrix x y 0 in
             for i = 0 to x-1 do
               for j = 0 to y-1 do
		 nmat.(i).(j) <- thearray.(i+steps*r).((j+timeshift+1) mod y)
	       done
	     done; howto x y (fun i j -> nmat.(i).(j))
	 )
       | _ -> assert false
    )


(************* CORE CA FUNCTIONS *****************)


(* Preparation fonctions called from main process just before computing a new CA and/or conf *)

let prepareseed ac com =
  mixwanted ac com;
  if neednewac ac then ac.seed.rs_rule <- getcount ();
  if neednewconf ac then ac.seed.rs_conf <- getnocount ();
  debugline Dbgdev "Seed prepared ";
  ac

let checkCompatibility ext cur =
  if fst ext <> fst cur then 
    if snd ext > snd cur then 
      debugline Dbgwarning "your ACML version is older than the one used to save the source file!"
	
let (register_sig, keep_universe, getseed, deviateseed, restorestate, get_saved, force_restore) =
  let savedstate = ref [] in
  let override_universe = ref true in
  let sig_fun = ref (fun (x:string) -> ("",0.)) in
    ((fun f -> sig_fun := f),
     (fun () -> override_universe:=false),
     (fun  ac -> 
	let rec doit = function
	    [] -> ()
	      (************* mutable structure everywhere: must make a copy before use **********************)
	  | (SavObjname s)::q -> (ac.obj_name <- Some s; doit q)
	  | (SavSeed s)::q -> (ac.seed <- (cloneseed s); doit q)
 	  | (SavParams p)::q -> (ac.parameters.params <- (List.map cloneparam p); doit q)
	  | (SavUniverse u)::q -> (if !override_universe then ac.universe <- u else override_universe:=false; doit q)
	  | (SavKernelSig (a,b))::q -> (checkCompatibility (a,b) (!sig_fun "kernel"); doit q)
	  | (SavAcmlSig (a,b))::q -> (checkCompatibility (a,b) (!sig_fun "acml"); doit q)
	  | (SavRandom2Sig (a,b))::q -> (checkCompatibility (a,b) (!sig_fun "random2"); doit q)
	in
(* 	  ac.name <- None; *)
(* 	  ac.obj_name <- None; *)
	  (* restore saved state only at first time *)
	  if initial ac then (
	    if (!savedstate <> []) then debugline Dbgdev "Restoring saved state...";
	    doit !savedstate
	  );
	  (* copy states in seed of the ca to PRNG states *)
 	  if neednewac ac then (debugline Dbgdev "Importing rule PRNG state...";
				setcount ac.seed.rs_rule);
	  if neednewconf ac then (debugline Dbgdev "Importing conf PRNG state...";
				  setnocount ac.seed.rs_conf)
     ),
     (* Old saving format: marshalled odd combination of tuples to save seed and parameters plus useless stuff *)
     (fun s -> match !savedstate with 
	  [] -> (
	    try
	      let ((s,u,i,_),p) = (Marshal.from_string s 0) in 
		savedstate:= [SavSeed {rs_rule = s; rs_conf = u; ordinal = i}; SavParams p]
	    with e -> debugline Dbgwarning ("marshaling problem '"^(Printexc.to_string e) 
					    ^"' (possible cause: Ocaml version mismatch)")
	  )
	| _ -> ()
     ),
     (* Current saving format: list of saving items *)
     (fun s -> match !savedstate with
	  [] -> savedstate:=s
	    (* if already saved then don't reload previous saved states *)
	| _ -> ()
     ),
     (* Return currently saved state *)
     (fun () -> !savedstate
     ),
     (* Force state restoration *)
     (fun s -> savedstate:=s
     )
    )
      
let send (ac : acprototype) = 
  (Marshal.to_channel stdout ac []);
  flush stdout

let infopm = ref ignore

let sendinfo cur tot =
  let seuil = 1000 in
  if (tot > seuil) then
    if (cur mod (tot/20)) = 0 then 
      let i = int_of_float (((float_of_int cur)/.(float_of_int tot))*.100.) in
	(
	  if is_main_program () then !infopm i
	  else send (fakeac i)
	)


let nextstateunidim vgf ac rawnext rawget =
  let rec doit vgetfun t c =
    match (ac.neighbourhood, t) with
	      
      |	((Moore n|VonNeumann n|Cross n), (XTable _ | Table _ | FunArray _ | MultisetTable _ | MultisetFunArray _ 
	 | MultisetFunFun _ | FunFun _)) -> 
	  (
	    for i= -n to n do
	      rawget (vgetfun i) c
	    done;
	    rawnext (vgetfun 0)) 
      | (_,Sparse (x,t)) -> doit vgetfun t c
      | _ -> raise (Unimplemented "nextstate1D")
  in
    doit vgf ac.transitionrule ac.cardinality

let getunipos  = function
  | Line x -> x
  | _ -> internalbug "get unidim position"

let addunipos u = function
  | Line b -> u+b
  | _ -> internalbug "add unidim position"

let nextconfunidim ac matrix lastline x y tott =
  let r = getradius ac in 
  let l = !lastline in
  let a = arity ac in
  let c = ac.cardinality in
  let enlarge_spot sp = 
    let lili = 
      let rec encore acc x = if x = 0 then 0::acc else encore (x::((-x)::acc)) (x-1) in 
	encore [] r in
    let valid_pos i = Line ((x+2*r*tott+i) mod (x+2*r*tott)) in
      List.fold_left (fun acc i -> SpotSet.add (valid_pos (addunipos i sp)) acc) (SpotSet.empty) lili in
  let ta = Array.make a 0 in
  let ta_user = Array.make a 0 in
  let tc = Array.make c 0 in
  let tc_user = Array.make c 0 in
  let vgetfun = if r >= x || r >= y then
    (fun i px -> Array.unsafe_get (Array.unsafe_get matrix ((x+2*r*tott+((i+px) mod (x+2*r*tott))) mod (x+2*r*tott))) (l))
  else
    (fun i px -> Array.unsafe_get (Array.unsafe_get matrix ((x+2*r*tott+i+px) mod (x+2*r*tott))) (l))
  in 
  let myexp a b =
    let rec tothe acc k i = if i=0 then acc else if i mod 2 = 1 
    then tothe (k*acc) (k*k) (i/2) else tothe acc (k*k) (i/2) in
      tothe 1 a b in
  let rec rawupofac = function
      Table u -> 
	(let acc = ref 0 and base = ref 1 in 
 	 let gettable = getter_from_fast_array u in
 	   ((fun i c -> acc := !acc + i * !base; base := !base * c),(fun cc -> let b = !acc in
								       (acc:= 0; base:= 1; gettable b))))
    | XTable u -> 
	(let acc = ref 0 and base = ref 1 in
	 let modifbase = match ac.binary_power with
	     Vcard vc | Vboth (vc,_) -> (fun base c -> base := !base lsl vc)
	   | _ -> (fun base c -> base := !base * c)
	 in
	   ((fun i c -> acc := !acc + i * !base; modifbase base c),(fun cc -> let b = !acc in
								      (acc:= 0; base:= 1; Array.unsafe_get u b))))
    | MultisetTable u ->
	(let acc = ref 0 in
	   (
	     (match ac.binary_power with
		  Varit k | Vboth (_,k) -> (fun i c -> acc := !acc + (1 lsl (k*i)))
		| _ -> (fun i c -> 
			  let ou = myexp (a+1) i in
			  let (q,r) = ( !acc / ou, !acc mod ou) in
			    acc := ou*(q+1)+r)),
	     (match ac.binary_power with
		  Vcard vc | Vboth (vc, _) -> 
		    (fun ccell -> let b = !acc in
		       (acc:= 0;  ((u.{b} lsr (vc*ccell)) land (1 lsl vc - 1))))
		| _ -> 
		    (fun ccell -> let b = !acc in
		       (acc:= 0;  ((u.{b} / (myexp c ccell)) mod c))))
	   )
	)
    | FunArray f -> 
	let pos = ref 0 in
	((fun i _ -> (ta.(!pos) <- i; incr pos)),
	 (fun _ -> (
	    pos:= 0;
	    Array.blit ta 0 ta_user 0 a;
	    f ta_user)))
    | MultisetFunArray f -> 
	((fun i _ -> (tc.(i) <- tc.(i)+1)),
	 (fun cc -> (
	    Array.blit tc 0 tc_user 0 c;
	    for i=0 to c-1 do
	      tc.(i) <- 0
	    done;
	    f cc tc_user)))
    | _ -> raise (Unimplemented "nextconf 1D")
  in
    (* read_right : destination position given is far from bounds *)
  let big_loop read_init read_right putnext =
    let ord = (l+1) mod y in
      for i= 0 to r-1 do
	read_init i;
	putnext i ord
      done;    
      for i= r to x+2*r*tott-1-r do
	read_right i;
	putnext i ord
      done;    
      for i= x+2*r*tott-r  to x+2*r*tott-1 do
	read_init i;
	putnext i ord
      done;
      lastline:=(l+1) mod y
  in
  let rec doit qi = match qi with
      MultisetTable _ | MultisetFunArray _ -> let (rawget,rawupdate) = rawupofac qi in
	(
	  for i= 0 to x+2*r*tott-1 do
	    matrix.(i).((l+1) mod y) <- (nextstateunidim (vgetfun i) ac rawupdate rawget)
	  done;
	  lastline:=(l+1) mod y
	)
    | FunArray f -> 
	let pos = ref 0 in
	let read_new_val x = ta.(!pos) <- x; pos:= (!pos + 1) mod a in
	let read_right i = read_new_val (Array.unsafe_get (Array.unsafe_get matrix (i+r)) (l)) in
	let read_init i = (
	  pos:=0;
	  for j = -r to r do
	    read_new_val (vgetfun i j)
	  done
	) in
	let putnext wa wb = 
	  if !pos=0 then Array.blit ta 0 ta_user 0 a 
	  else (Array.blit ta !pos ta_user 0 (a- !pos); Array.blit ta 0 ta_user (a- !pos) !pos); 
	  Array.unsafe_set (Array.unsafe_get matrix wa) (wb) (f ta_user) 
	in
	  big_loop read_init read_right putnext	  
    | Table tb -> let acc = ref 0 in
      let gettable = getter_from_fast_array tb in
      let putnext a b = Array.unsafe_set (Array.unsafe_get matrix a) (b) (gettable !acc) in
      let read_init i = (
	acc := vgetfun i r;
	for j = 1 to 2*r do
	  acc:= c*(!acc) + (vgetfun i (r-j))
	done
      )
      in
      let max_pond = myexp c (a-1) in
      let read_right i =
	acc:= !acc / c + (Array.unsafe_get (Array.unsafe_get matrix (i+r)) (l))* max_pond
      in
	big_loop read_init read_right putnext
    | FunFun f ->
	let reader pos a b = vgetfun pos a in
	let nop (a:int) = () in
	let putnext a b = Array.unsafe_set (Array.unsafe_get matrix a) (b) (f (reader a)) in
	  big_loop nop nop putnext
    | MultisetFunFun f ->
	let reader pos a b = vgetfun pos a in
	let count = ref 0 in
	let howmany pos x = (
	  count := 0;
	  for i = -r to r do
	    if reader pos i 0 = x then incr count
	  done; !count
	) in
	let nop (a:int) = () in
	let putnext a b = Array.unsafe_set (Array.unsafe_get matrix a) (b) (f (reader a 0 0) (howmany a)) in
	  big_loop nop nop putnext
    | XTable tb ->  
	let acc = ref 0 in
	let shift = match ac.binary_power with
	    Vcard vc | Vboth (vc,_) -> vc
	  | _ -> internalbug "XTable 1D"
	in
	  (* let putnext a b = matrix.(a).(b) <- (Array.unsafe_get tb !acc) in *)
	let putnext a b = Array.unsafe_set (Array.unsafe_get matrix a) (b) (Array.unsafe_get tb !acc) in
	let read_init i = (
	  acc := vgetfun i r;
	  for j = 1 to 2*r do
	    acc:= ((!acc) lsl shift) + (vgetfun i (r-j))
	  done
	)
	in
	let read_right i =
	  acc:= (!acc lsr shift) + ((Array.unsafe_get (Array.unsafe_get matrix (i+r)) (l)) lsl (shift*(a-1)))
	in
	  big_loop read_init read_right putnext
    | Sparse (activ, wat) -> let (rawget,rawupdate) = rawupofac wat in
	(
	  for i= 0 to x+2*r*tott-1 do
	    matrix.(i).((l+1) mod y) <- matrix.(i).((l) mod y)
	  done;
	  let enlarged_spots = (SpotSet.fold (fun pos acc -> SpotSet.union (enlarge_spot pos) acc) 
				  ac.spots SpotSet.empty) in
	    SpotSet.iter (fun pos -> let i = getunipos pos in
			    (
			      matrix.(i).((l+1) mod y) <- 
				(nextstateunidim (vgetfun i) ac rawupdate rawget))) enlarged_spots;
	    ac.spots <- SpotSet.fold (fun pos acc -> let i = matrix.((x+2*r*tott+(getunipos pos)) mod (x+2*r*tott)).((l+1) mod y) in
					if (ActivSet.mem i activ) then 
					  SpotSet.union acc (enlarge_spot pos)
					else acc
				     )  enlarged_spots SpotSet.empty;
	    lastline:=(l+1) mod y
	)
	  (*     | Selfproduct f -> doit f *)
    | _ -> raise (Unimplemented "nextconf1D")
  in
    doit ac.transitionrule

let completematrix dom ac matrix ts x y t=
  assert (ac.dimension = 1);
  let read j = universe_reader ac (Unidim matrix,Zerodim ts) j (x,y) t in
  let write j = universe_writer ac (Unidim matrix,Zerodim ts) j (x,y) t in
    for j=0 to (y-2) do  
      if dom then do_all_measurements ac (j) x y (read j) (write j);
      nextconfunidim ac matrix ts x y t;
    done;
    if dom then do_all_measurements ac (y-1) x y (read (y-1)) (write (y-1))
     
let repetitivinitund distortion ac matrix x y t f =
  let r = getradius ac in
    (
      for i= 0 to x+2*r*t-1 do
	match distortion i with 
	    Some a -> (matrix.(a).(0) <- ((f i) mod ac.cardinality))
	  | _ -> ()
      done
    )


let singleinitund distortion ac matrix x c =
  match distortion x with
      Some a -> (matrix.(a).(0) <- (c mod ac.cardinality))
    | _ -> ()


let newinitconfunidim ac matrix activeline x y t =
  let r = getradius ac in
  let rec xofpos = function
      Plane (x,y) -> x
    | Center ->  x/2+r*t
(*     | Shift (p,q) -> (xofpos p) + (xofpos q) *)
    | _ -> raise (Unimplemented "new init conf")
  in    
  let distortionofzone olddistortion = function
      Rectangular (p1,p2) -> 
	let xa = xofpos p1
	and xb = xofpos p2 in
	(fun x ->
	   if (x < xa) || (x > xb) 
	   then None
	   else olddistortion x)
    | Randomrectangular (p1,hmax,vmax) -> 
	let xa = xofpos p1 in
	let xb = xa + myrandomnocount hmax in
	  (fun x ->
	     if (x < xa) || (x > xb)
	     then None
	     else olddistortion x)
  in  
  let rec do_it distortion = function
      Random(Uniform) -> repetitivinitund distortion ac matrix x y t (fun i -> myrandomnocount ac.cardinality)
    | Random(Bernouilli a) -> repetitivinitund distortion ac matrix x y t (fun i -> a.(myrandomnocount (Array.length a)))
    | Monochrome c -> repetitivinitund distortion ac matrix x y t (fun i -> c)
    | Repeat (n,conf) -> 
	for i = 1 to n do
	  do_it distortion conf
	done
    | Layers l -> List.iter (do_it distortion) l
    | Sized (w,h,c) -> do_it distortion c
    | Finite l -> List.iter (fun (p, c) -> (singleinitund distortion ac matrix (xofpos p) c)) l
    | Randomtranslate conf -> let rx = myrandomnocount (x+2*r*t-1) in
      let newdistortion x = distortion (x-rx) in
	do_it newdistortion conf
    | Subzone (zone, conf) -> do_it (distortionofzone distortion zone) conf
    | Modulo (Udim at) -> let s = Array.length at in 
	repetitivinitund distortion ac matrix x y t (fun i -> at.(i mod s)) 
    | Modulo (_) -> raise (BadACML "1D ac and 2D init conf")
    | Custom genfun -> repetitivinitund distortion ac matrix x y t (fun i -> genfun i 0)
    | CustomArray genarrayfun -> 
      let ar = genarrayfun () in
      let w = Array.length ar in
      let h = Array.length ar.(0) in
      let genfun x y = ar.(x mod w).(y mod h) in
	repetitivinitund distortion ac matrix x y t (fun i -> genfun i 0)
    | Masked (m,conf) -> do_it (function p -> if m (p) (0) then distortion p else None) conf

  in
  let distortion a = Some ((a+x+2*r*t) mod (x+2*r*t)) in
   (
     do_it distortion (get_initconf ac);
     (* Collect spots in case of sparse update rule *)
     (match ac.transitionrule with
	 Sparse (ki, _) -> (
	   debugline Dbgdev "Collecting activity spots...";
	   ac.spots <- SpotSet.empty;
	   for i = 0 to (x+2*r*t - 1) do
	     if ActivSet.mem matrix.(i).(0) ki then ac.spots <- SpotSet.add (Line i) ac.spots
	   done
	 )
       | _ -> ());
     activeline:=0;
     completematrix true ac matrix activeline x y t
   )


let repetitivinit distortion ac writematrix x y t f =
  let r = getradius ac in
    (
      for i= 0 to x+2*r*t-1 do
	for j = 0 to y+2*r*t-1 do
	  match distortion i j with 
	      Some (a,b) -> (writematrix (a) (b) ((f i j) mod ac.cardinality))
	    | _ -> ()
	done
	done
    )


let singleinit distortion ac writematrix x y c =
  match distortion x y with
      Some (a,b) -> (writematrix (a) (b) (c mod ac.cardinality))
    | _ -> ()


let newinitconfbidim ac writematrix readmatrix x y t =
  let r = getradius ac in
  let rec xofpos = function
      Plane (x,y) -> x
    | Center ->  x/2+r*t
(*     | Shift (p,q) -> (xofpos p) + (xofpos q) *)
    | _ -> raise (Unimplemented "init conf")
  in      
  let rec yofpos = function
      Plane (x,y) -> y
    | Center -> y/2 +r*t
(*     | Shift (p,q) -> (yofpos p) + (yofpos q) *)
    | _ -> raise (Unimplemented "position")
  in    
  let distortionofzone olddistortion = function
      Rectangular (p1,p2) -> 
	let (xa,ya) = (xofpos p1, yofpos p1)
	and (xb,yb) = (xofpos p2, yofpos p2) in
	(fun x y ->
	   if (x < xa) || (x > xb) || (y < ya) || (y > yb) 
	   then None
	   else olddistortion x y)
    | Randomrectangular (p1,hmax,vmax) -> 
	let (xa,ya) = (xofpos p1, yofpos p1) in
	let xb = xa + myrandomnocount hmax and yb = ya + myrandomnocount vmax in
	  (fun x y ->
	     if (x < xa) || (x > xb) || (y < ya) || (y > yb) 
	     then None
	     else olddistortion x y)
  in  
  let rec do_it distortion = function
      Random(Uniform) -> repetitivinit distortion ac writematrix x y t (fun i j -> myrandomnocount ac.cardinality)
    | Random(Bernouilli a) -> repetitivinit distortion ac writematrix x y t (fun i j -> a.(myrandomnocount (Array.length a)))
    | Monochrome c -> repetitivinit distortion ac writematrix x y t (fun i j -> c)
    | Repeat (n,conf) -> 
	for i = 1 to n do
	  do_it distortion conf
	done
    | Layers l -> List.iter (do_it distortion) l
    | Sized (w,h,c) -> do_it distortion c
    | Finite l -> List.iter (fun (p, c) -> (singleinit distortion ac writematrix (xofpos p) (yofpos p) c)) l
    | Randomtranslate conf -> let rx = myrandomnocount (x+2*r*t-1) and ry = myrandomnocount (y+2*r*t-1) in
      let newdistortion x y = distortion (x-rx) (y-ry) in
	do_it newdistortion conf
    | Subzone (zone, conf) -> do_it (distortionofzone distortion zone) conf
    | Modulo (Bdim at) -> let s = Array.length at and u = Array.length at.(0) in 
	repetitivinit distortion ac writematrix x y t (fun i j -> at.(i mod s).(j mod u)) 
    | Modulo (_) -> raise (BadACML "2D ac and 1D init conf")
    | CustomArray genarrayfun -> 
      let ar = genarrayfun () in
      let w = Array.length ar in
      let h = Array.length ar.(0) in
      let genfun x y = ar.(x mod w).(y mod h) in
      repetitivinit distortion ac writematrix x y t genfun
    | Custom genfun -> repetitivinit distortion ac writematrix x y t genfun
    | Masked (m,conf) -> do_it (fun p q  -> if m (p) (q) then distortion p q else None) conf
  in
  let distortion a b = Some ((a+x+2*r*t) mod (x+2*r*t), (b+y+2*r*t) mod (y+2*r*t)) in
    (
      do_it distortion (get_initconf ac);
     (* Collect spots in case of sparse update rule *)
     (match ac.transitionrule with
	 Sparse (ki, _) -> (
	   debugline Dbgdev "Collecting activity spots...";
	   ac.spots <- SpotSet.empty;
	   for i = 0 to x - 1 do
	     for j = 0 to y - 1 do
	       if ActivSet.mem (readmatrix (i) (j)) ki then ac.spots <- SpotSet.add (Plane (i,j)) ac.spots
	     done
	   done
	 )
       | _ -> ())
    )



let nextconfbidim ac before next x y curt tott =
  (* some say it is better compiled when eta-expanded *)
  let readbefore  = getter_from_fast_matrix before  in
  let writebefore  = setter_from_fast_matrix before in
  let readnext  = getter_from_fast_matrix next  in
  let writenext  = setter_from_fast_matrix next  in
  let r = getradius ac in

  (* OPTZ *)
  let vgetfun =
    if r >= x || r>= y then
      (fun i j px py -> readbefore ((x+2*r*tott+((px+i) mod (x+2*r*tott))) mod (x+2*r*tott)) ((y+2*r*tott+((py+j) mod (y+2*r*tott))) mod (y+2*r*tott)))
    else
      (fun i j px py -> readbefore ((x+2*r*tott+i+px) mod (x+2*r*tott)) ((y+2*r*tott+j+py) mod (y+2*r*tott)))
  in
  let pos = ref 0 in
  let a = arity ac in
  let c = ac.cardinality in
  let t = Array.make a 0 in
  let cleanstart = ref 0 in
    (* Big update loop *)
  let big_loop minit vinit cinit mright vright cright putnext mfright vfright cfright =
    let (init, right, fast_right) =
      (match ac.neighbourhood with
	   Moore n -> (minit, mright, mfright)
	 | VonNeumann n -> (vinit, vright, vfright)
	 | Cross n -> (cinit, cright, cfright)
      ) in
    let (na, nb) = (y+2*r*tott-1, x+2*r*tott-1) in
       if r>=x || r>=y then (
      	 for j = 0 to na do
      	   init 0 j;
      	   putnext 0 j;
      	   for i= 1 to nb do
      	     right i j;
      	     putnext i j;
      	   done;
      	 done;
       )
       else (
	 (* Vertical Top *)
	for j = 0 to r-1 do
	  init 0 j;
	  putnext 0 j;
	  for i= 1 to nb do
	    right i j;
	    putnext i j;
	  done;
	done;
	(* Vertical Middle *)
 	for j = r to na-r do
 	  init 0 j;
	  putnext 0 j;
	  for i= 1 to r do
	    right i j;
	    putnext i j;
	  done;
	  (* TRUE HEART OF THE BIG LOOP *)
	  (* No boundary condition to check *)
	  for i= r+1 to nb-r do
	    (* fast computation of  next neighb. context *)
	    fast_right i j;
  	    putnext i j;
	  done;
	  (* OUT OF HEART *)
	  for i= nb-r+1 to nb do
	    right i j;
 	    putnext i j;
	  done;
	done;
 	(* Vertical Bottom  *)
	for j = na-r+1 to na do
 	  init 0 j;
	  putnext 0 j;
	  for i= 1 to nb do
	    right i j;
	    putnext i j;
	  done
	done;
      )
  in    
    (* Read Cross neihgbourhood in tab *)
  let cinit x y =
    pos:= 0;
    cleanstart := 0;
    for i= -r to r do
      Array.unsafe_set t (!pos) (vgetfun x y i 0);
      incr pos
    done;
    for j= -r to r do
      if j <> 0 then 
	(
	  Array.unsafe_set t (!pos) (vgetfun x y 0 j);
	  incr pos
	)
    done 
  in
    (* Read von Neumann neihgbourhood in tab (compatible with cross neihgb for radius 1) *)
  let vinit x y =
    pos:= 0;
    cleanstart := 0;
    for i= -r to r do
      Array.unsafe_set t (!pos) (vgetfun x y i 0);
      incr pos
    done;
    for j= -r to r do
      if j <> 0 then 
	let w = r - (abs j) in
	  for i = -w to w do
	    Array.unsafe_set t (!pos) (vgetfun x y i j);
	    incr pos
	  done
    done 
  in
    (* Read Moore neihgbourhood in tab *)
  let minit x y =
    pos := 0;
    cleanstart := 0;
    for i= -r to r do
      for j= -r to r do
	Array.unsafe_set t (!pos) (vgetfun x y i j);
	incr pos;
      done
    done 
  in
  let rec doit = function
      (* SPARSE update *)
      Sparse (activ, wat) -> 
	let getbipos = function
	    Plane (a,b) -> (a,b)
	  | _ -> internalbug "getbipos"
	in
	  (* Convert rule to a function *)
	let reponse getter tablo = 
	  let n = match ac.binary_power with
	      Vcard vc | Vboth (vc,_) -> 1 lsl vc
	    | _ -> ac.cardinality 
	  in
	  let acc = ref 0 in
	  let base = ref 1 in
	    for i=0 to Array.length tablo - 1 do
	      acc := !acc + tablo.(i)*(!base);
	      base := !base * n
	    done;
	    getter !acc
	in
	let rule_as_fun_array = match wat with
 	    FunArray f -> f
	  | Table t -> reponse (getter_from_fast_array t)
	  | XTable t -> reponse (fun acc -> Array.unsafe_get t acc)
	  | FunFun f -> let r = relativecell ac in
	      (fun x -> f (fun i j -> x.(r i j)))
	  | _ -> raise (Unimplemented "Sparse update above odd rule type")
	in
	  (* Compute next state for a single cell *)
	let nextstate = match ac.neighbourhood with
	    Moore _ -> (fun x y -> minit x y; rule_as_fun_array t)
	  | VonNeumann _ -> (fun x y -> vinit x y; rule_as_fun_array t)
	  | Cross _ -> (fun x y -> cinit x y; rule_as_fun_array t)
	in
	let addbipos a b = function
	  | Plane (u,v) -> (u+a,v+b)
	  | _ -> internalbug "add bidim position"
	in
	  (* Add neighbouring cells in spots collection *)
	let enlarge_spot sp = 
 	  let lili = 
	    let rec encore acc x y = if x = 0 then (
	      if y = 0 then  (0,0)::acc else encore ((x,-y)::((-x,-y)::((x,y)::((-x,y)::acc)))) r (y-1)
	    )
	    else encore ((x,-y)::((-x,-y)::((x,y)::((-x,y)::acc)))) (x-1) y in
	      encore [] r r in
	  let valid_pos (i,j) = Plane ((x+2*r*tott+i) mod (x+2*r*tott), (y+2*r*tott+j) mod (y+2*r*tott)) in
	    List.fold_left (fun acc (i,j) -> SpotSet.add (valid_pos (addbipos i j sp)) acc) (SpotSet.empty) lili in
	  (
	    let enlarged_spots = (SpotSet.fold (fun pos acc -> SpotSet.union (enlarge_spot pos) acc) 
				    ac.spots SpotSet.empty) in
	      SpotSet.iter (fun pos -> let (i,j) = getbipos pos in
			      writenext (i) (j) (nextstate i j)
			   ) enlarged_spots;
	      ac.modified <- enlarged_spots;
	      ac.spots <- SpotSet.fold (fun pos acc -> let (i,j) = getbipos pos in 
					let sta = readnext (i) (j) in (
					    writebefore (i) (j) sta;
					    if (ActivSet.mem sta activ) then 
					      SpotSet.union acc (enlarge_spot pos)
					    else acc)
				       )  enlarged_spots SpotSet.empty		
	  )
    | MultisetFunArray f ->
	let tgiven = Array.make c 0 in
	let tintern = Array.make c 0 in
	let up i = Array.unsafe_set tintern (i) ((Array.unsafe_get tintern i)+1) in
	let pu i = Array.unsafe_set tintern (i) ((Array.unsafe_get tintern i)-1) in
	  (* Read multiset cross neihgbourhood in tab *)
	let cinit x y =
	  for i = 0 to c-1 do
	    tintern.(i)<-0
	  done;
	  for i= -r to r do
	    up (vgetfun x y i 0);
	  done;
	  for j= -r to r do
	    if j <> 0 then up (vgetfun x y 0 j)
	  done 
	in
	  (* Read multiset von Neumann neihgbourhood in tab *)
	let vinit x y =
	  for i = 0 to c-1 do
	    tintern.(i)<-0
	  done;
	  for j= -r to r do
	    let w = r - (abs j) in
	      for i = -w to w do
		up (vgetfun x y i j)
	      done
	  done 
	in
	  (* Read multiset Moore neihgbourhood in tab *)
	let minit x y =
	  for i = 0 to c-1 do
	    tintern.(i)<-0
	  done;
	  for i= -r to r do
	    for j= -r to r do
	      up (vgetfun x y i j)
	    done
	  done 
	in
	  (* Shift right multiset cross neihgbourhood in tab *)
	let cright x y =
	  for j = -r to r do
	    if j<>0 then (
	      pu (vgetfun x y (-1) j);
	      up (vgetfun x y 0 j)
	    )
	  done;
	  up (vgetfun x y r 0);
	  pu (vgetfun x y (-r-1) 0)
	in
	let cfright x y =
	  for j = y-r to y+r do
	    if j<>y then (
	      (* Warning: must remove before adding to avoid overflow!!! *)
 	      pu (readbefore (x-1) (j)) (* (vgetfun x y (-1) j) *);
	      up (readbefore (x) (j)) (* (vgetfun x y 0 j) *)
	    )
	  done;
	  (* Warning again! *)
	  pu (readbefore (x-r-1) (y)) (* (vgetfun x y (-r-1) 0) *);
 	  up (readbefore (x+r) (y)) (* (vgetfun x y r 0) *)
	in
	  (* Shift right multiset von Neumann neihgbourhood in tab *)
	let vright x y =
	  for j = -r to r do
	    let w = r - (abs j) in
		pu (vgetfun x y (-w-1) j);
		up (vgetfun x y w j)
	  done
	in
	let vfright x y =
	  for j = -r to r do
	    let w = r - (abs j) in
	      (* Warning: must remove before adding to avoid overflow!!! *)
 	      pu (readbefore (x-w-1) (y+j)) (* (vgetfun x y (-1) j) *);
	      up (readbefore (x+w) (y+j)) (* (vgetfun x y 0 j) *)
	  done
	in
	  (* Shift right multiset Moore neihgbourhood in tab *)
	let mright x y =
	  for j = -r to r do
	    up (vgetfun x y r j);
	    pu (vgetfun x y (-r-1) j)
	  done
	in
	let mfright x y =
	  for j = y-r to y+r do
	    up (readbefore (x+r) (j));
	    pu (readbefore (x-r-1) (j))
	  done
	in
	  (* multiset local rules use central cell as first argument *) 
	let putnext x y =
	  Array.blit tintern 0 tgiven 0 c;
	  writenext (x) (y) (f (readbefore (x) (y)) tgiven)
	in
	  big_loop minit vinit cinit mright vright cright putnext mfright vfright cfright
    | MultisetFunFun f ->
	let count = ref 0 in
	  (* howmany 'sta' in cross neihgbourhood of cell (x,y) *)
	let chowmany x y sta =
	  count := 0;
	  for i= -r to r do
	    if (vgetfun x y i 0) = sta then incr count;
	  done;
	  for j= -r to r do
	    if j <> 0 && (vgetfun x y 0 j) = sta then incr count;
	  done; !count
	in
	  (* howmany 'sta' in von Neumann neihgbourhood of cell (x,y) *)
	let vhowmany x y sta =
	  count := 0;
	  for j= -r to r do
	    let w = r - (abs j) in
	      for i= -w to w do
		if (vgetfun x y i j) = sta then incr count;
	      done
	  done; !count
	in
	  (* howmany 'sta' in Moore neihgbourhood of cell (x,y) *)
	let mhowmany x y sta =
	  count := 0;
	  for i= -r to r do
	    for j= -r to r do
	      if (vgetfun x y i j) = sta then incr count;
	    done
	  done; !count
	in
	  (* multiset local rules use central cell as first argument *) 
	let nop (a:int) (b:int) = () in
	let putnext  =       
	  match ac.neighbourhood with
	      Moore n -> (fun posx posy -> writenext posx posy (f (vgetfun posx posy 0 0) (mhowmany posx posy)))
	    | VonNeumann n -> (fun posx posy -> writenext posx posy (f (vgetfun posx posy 0 0) (vhowmany posx posy)))
	    | Cross n -> (fun posx posy -> writenext posx posy (f (vgetfun posx posy 0 0) (chowmany posx posy)))
	in
	  big_loop nop nop nop nop nop nop putnext nop nop nop
    | MultisetTable t ->
	let acc = ref 0 in
	let dividors = Array.make c 0 in	
	let statecomp = Array.make c 0 in
	let (up,pu) = match ac.binary_power with
	    Varit k | Vboth (_,k) -> (fun i -> acc := !acc + (1 lsl (k*i))), (fun i -> acc := !acc - (1 lsl (k*i)))
	  | _ -> (fun i ->
		    let knife = Array.unsafe_get dividors (i) in 
		    let (a,b) = (!acc / knife, !acc mod knife) in acc:= ((a+1) * knife + b)),
	      (fun i ->
		 let knife = Array.unsafe_get dividors (i) in 
		 let (a,b) = (!acc / knife, !acc mod knife) in acc:= ((max 0 (a-1)) * knife + b)
	      )
	in	  
	  (* Read multiset cross neihgbourhood in acc *)
	let cinit x y =
	  acc:=0;
	  for i= -r to r do
	    up (vgetfun x y i 0);
	  done;
	  for j= -r to r do
	    if j <> 0 then up (vgetfun x y 0 j)
	  done 
	in
	  (* Read multiset von Neumann neihgbourhood in tab *)
	let vinit x y =
	  acc:=0;
	  for j= -r to r do
	    let w = r - (abs j) in
	      for i = -w to w do
		up (vgetfun x y i j)
	      done
	  done 
	in
	  (* Read multiset Moore neihgbourhood in acc *)
	let minit x y =
	  acc:=0;
	  for i= -r to r do
	    for j= -r to r do
	      up (vgetfun x y i j)
	    done
	  done 
	in
	  (* Shift right multiset cross neihgbourhood in acc *)
	let cright x y =
	  for j = -r to r do
	    if j<>0 then (
	      (* Warning: must remove before adding to avoid overflow!!! *)
 	      pu (vgetfun x y (-1) j);
	      up (vgetfun x y 0 j)
	    )
	  done;
	  (* Warning again! *)
	  pu (vgetfun x y (-r-1) 0);
 	  up (vgetfun x y r 0)
	in
	  (* Fast shift right *)
	let cfright x y =
	  for j = y-r to y+r do
	    if j<>y then (
	      (* Warning: must remove before adding to avoid overflow!!! *)
 	      pu (readbefore (x-1) (j)) (* (vgetfun x y (-1) j) *);
	      up (readbefore (x) (j)) (* (vgetfun x y 0 j) *)
	    )
	  done;
	  (* Warning again! *)
	  pu (readbefore (x-r-1) (y)) (* (vgetfun x y (-r-1) 0) *);
 	  up (readbefore (x+r) (y)) (* (vgetfun x y r 0) *)
	in
	let v1fright x y =
 	  pu (readbefore (x-1) (y-1));
	  up (readbefore (x) (y-1));
 	  pu (readbefore (x-1) (y+1));
	  up (readbefore (x) (y+1));
	  pu (readbefore (x-2) (y));
 	  up (readbefore (x+1) (y))
	in
	  (* Shift right multiset von Neumann neihgbourhood in tab *)
	let vright x y =
	  for j = -r to r do
	    let w = r - (abs j) in
		pu (vgetfun x y (-w-1) j);
		up (vgetfun x y w j)
	  done
	in
	let vfright x y =
	  for j = -r to r do
	    let w = r - (abs j) in
	      (* Warning: must remove before adding to avoid overflow!!! *)
 	      pu (readbefore (x-w-1) (y+j)) (* (vgetfun x y (-1) j) *);
	      up (readbefore (x+w) (y+j)) (* (vgetfun x y 0 j) *)
	  done
	in
	  (* Shift right multiset Moore neihgbourhood in acc *)
	let mright x y =
	  for j = -r to r do
	    (* Warning: must remove before adding to avoid overflow!!! *)
	    pu (vgetfun x y (-r-1) j);
	    up (vgetfun x y r j)
	  done
	in
	  (* Fast shift right *)
	let mfright x y =
	  for j = y-r to y+r do
	    pu (readbefore (x-r-1) (j));
	    up (readbefore (x+r) (j));
	  done
	in
	  (* Faster shift right for radius 1 *)
	let m1fright x y =
	  pu (readbefore (x-2) (y-1));
	  up (readbefore (x+1) (y-1));
	  pu (readbefore (x-2) (y));
	  up (readbefore (x+1) (y));
	  pu (readbefore (x-2) (y+1));
	  up (readbefore (x+1) (y+1));
	in
	  (* multiset local rules use central cell as first argument *) 
	let putnext =
	  (match ac.binary_power with
	       Vcard vc | Vboth (vc, _) -> 
		 (fun x y -> 
		    writenext (x) (y) (((Bigarray.Array1.unsafe_get t !acc) lsr (vc*(readbefore (x) (y))))
				       land ((1 lsl vc) - 1))
		 )
	     | _ -> (fun x y ->
		       writenext (x) (y) (((Bigarray.Array1.unsafe_get t !acc) / statecomp.(readbefore (x) (y))) mod c)
		    )
	  )
	in
	  statecomp.(0) <- 1;
	  dividors.(0) <- 1;
	  for i = 1 to c-1 do
	    statecomp.(i) <- statecomp.(i-1) * c;
	    dividors.(i) <- dividors.(i-1) * (a+1)
	  done;
	  big_loop minit vinit cinit mright vright cright putnext (if r=1 then m1fright else mfright) 
	    (if r=1 then v1fright else vfright) (if r=1 then v1fright else cfright)

    | FunFun f ->
	let nop (a:int) (b:int) = () in
	let putnext posx posy = writenext posx posy (f (vgetfun posx posy)) in
	  big_loop nop nop nop nop nop nop putnext nop nop nop
    | FunArray f ->
	let tclean = Array.make a 0 in
	let putnext x y =
 	  (* debugline Dbgdev (Printf.sprintf "%dx%d a=%d cleanstart=%d pos=%d" x y a !cleanstart !pos); *)
 	  Array.blit t !cleanstart tclean 0 (a- !cleanstart);
 	  if !cleanstart > 0 then Array.blit t 0 tclean (a- !cleanstart) !cleanstart;
 	  writenext (x) (y) (f tclean)
	in
	let mright x y =
	  pos := !cleanstart;
	  for j= -r to r do
	    Array.unsafe_set t (!pos) (vgetfun x y r j);
	    pos := (!pos + 1) mod a
	  done;
	  cleanstart := !pos
	in
	let mfright x y =
	  pos := !cleanstart;
	  for j= y-r to y+r do
	    Array.unsafe_set t (!pos) (readbefore (x+r) (j));(*  vgetfun x y r j; *)
	    pos := (!pos + 1) mod a
	  done;
	  cleanstart := !pos
	in
	let vright x y =
	  cleanstart:= (!cleanstart + 1) mod a;
	  pos := (!cleanstart+2*r) mod a;
	  Array.unsafe_set t (!pos) (vgetfun x y r 0);
(* 	  pos := (!pos + 1) mod a; *)
	  for j= -r to -1 do
	    let w = r - (abs j) in
	      pos := (!pos + 1 + 2*w) mod a;
	      Array.unsafe_set t (!pos) (vgetfun x y w j);
	  done;
	  for j= 1 to r do
	    let w = r - (abs j) in
	      pos := (!pos + 1 + 2*w) mod a;
	      Array.unsafe_set t (!pos) (vgetfun x y w j);
	  done
	in
	let vfright x y =
	  cleanstart:= (!cleanstart + 1) mod a;
	  pos := (!cleanstart+2*r) mod a;
	  Array.unsafe_set t (!pos) (readbefore (x+r) (y)) (* vgetfun x y r 0 *);
(* 	  pos := (!pos + 1) mod a; *)
	  for j= -r to -1 do
	    let w = r - (abs j) in
	      pos := (!pos + 1 + 2*w) mod a;
	      Array.unsafe_set t (!pos) (readbefore (x+w) (y+j)) (* vgetfun x y 0 j *);
	  done;
	  for j= 1 to r do
	    let w = r - (abs j) in
	      pos := (!pos + 1 + 2*w) mod a;
	      Array.unsafe_set t (!pos) (readbefore (x+w) (y+j)) (* vgetfun x y 0 j *);
	  done
	in
	let cright x y =
	  cleanstart:= (!cleanstart + 1) mod a;
	  pos := (!cleanstart+2*r) mod a;
	  Array.unsafe_set t (!pos) (vgetfun x y r 0);
	  pos := (!pos + 1) mod a;
	  for j= -r to r do
	    if j<>0 then (
	      Array.unsafe_set t (!pos) (vgetfun x y 0 j);
	      pos := (!pos + 1) mod a
	    )
	  done
	in
	let cfright x y =
	  cleanstart:= (!cleanstart + 1) mod a;
	  pos := (!cleanstart+2*r) mod a;
	  Array.unsafe_set t (!pos) (readbefore (x+r) (y));
	  pos := (!pos + 1) mod a;
	  for j= y-r to y+r do
	    if j<>y then (
	      Array.unsafe_set t (!pos) (readbefore (x) (j));
	      pos := (!pos + 1) mod a
	    )
	  done
	in
	  big_loop minit vinit cinit mright vright cright putnext mfright vfright cfright
    | XTable t -> 
	let acc = ref 0 in
	let vc = match ac.binary_power with
	    Vcard q | Vboth (q,_) -> q
	  | _ -> internalbug "Xtable without power-of-two rounding"
	in
	  (* OPTZ *)
	let putnext x y = writenext x y (Array.unsafe_get t !acc) in

	let minit = match ac.neighbourhood with
	    Moore 1 ->
	      (fun px py -> acc:=((((((((vgetfun (1) (1) px py) lsl vc + (vgetfun (1) (0) px py)) lsl vc
				  + (vgetfun 1 (-1)  px py)) lsl vc + (vgetfun 0 (1) px py)) lsl vc
				+ (vgetfun 0 0 px py)) lsl vc + (vgetfun 0 (-1) px py)) lsl vc
			      + (vgetfun (-1) 1 px py)) lsl vc
			     + (vgetfun (-1) 0 px py)) lsl vc + (vgetfun (-1) (-1) px py))
	       | _ -> internalbug "XTable without Moore 1 neighbourhood"
	in
	let doright = 
 	  (fun a b c ->
	     acc := !acc lsr (3*vc) + (((c lsl vc) + b) lsl vc + a) lsl (6*vc)
	  )
	in
	let mright x y = doright (vgetfun 1 (-1) x y) (vgetfun 1 (0) x y) (vgetfun 1 (1) x y) in
	  (* OPTZ *)
	let mfright x y = doright (readbefore (x+1) (y-1)) (readbefore (x+1) (y)) 
	  (readbefore (x+1) (y+1)) in
	let forbid x y = internalbug "call forbiden in XTable" in
	  (match (before, next) with
	       (MatNative b, MatNative n) ->
		 let putnext x y = Array.unsafe_set (Array.unsafe_get n x) y (Array.unsafe_get t !acc) in
		 let mfright x y = doright (Array.unsafe_get (Array.unsafe_get b (x+1)) (y-1)) 
		   (Array.unsafe_get (Array.unsafe_get b (x+1)) (y)) 
		   (Array.unsafe_get (Array.unsafe_get b (x+1)) (y+1)) in
		   big_loop minit forbid forbid mright forbid forbid putnext mfright forbid forbid
	     | (MatMixed (nfo,b), MatMixed (_,n)) ->
		 let y_mask = (1 lsl (nfo.log_blocs_per_int)) - 1 in
		 let bloc_mask = (1 lsl (nfo.bloc_size)) - 1 in
		 let bpi = nfo.log_blocs_per_int in
		 let bs = nfo.bloc_size in
		 let putnext x y = 
		   let where = (y lsr bpi) in
		   let shift = (y land y_mask) * bs in
		   let ith = Array.unsafe_get n x in
		     Array.unsafe_set ith where (((Array.unsafe_get ith where) land (lnot (bloc_mask lsl shift))) lor 
						   ((Array.unsafe_get t !acc) lsl shift))
		 in
		 let mfright x y = 
		   let line = Array.unsafe_get b (x+1) in
		     doright  (((Array.unsafe_get line ((y-1) lsr bpi)) lsr 
				  (((y-1) land y_mask) * bs)) land bloc_mask)
		       (((Array.unsafe_get line ((y) lsr bpi)) lsr 
			   (((y) land y_mask) * bs)) land bloc_mask)
		       (((Array.unsafe_get line ((y+1) lsr bpi)) lsr 
			   (((y+1) land y_mask) * bs)) land bloc_mask)
		 in big_loop minit forbid forbid mright forbid forbid putnext mfright forbid forbid
	     | _ -> big_loop minit forbid forbid mright forbid forbid putnext mfright forbid forbid
	  )
	   
    | Table t -> 
	let a = arity ac in
	let q = ac.cardinality in
	let neibmult = Array.make a 0 in
	let k = 2*r+1 in
	let base = ref 1 in
	let acc = ref 0 in
	let row = secure_intexp false q k in
	let mgetxy i j = Array.unsafe_get neibmult (k*i+j) in
	let vgetxy i j = 
	  if j=r then Array.unsafe_get neibmult (i) 
	  else if j < r then Array.unsafe_get neibmult (k+j*j+(i-r+j)) 
	  else Array.unsafe_get neibmult (k + r*r+(j-r-1)*(3*r-j+1)+(i+r-j)) in
	let cgetxy i j = if j=r then Array.unsafe_get neibmult (i) else if j<r then Array.unsafe_get neibmult (k+j) else Array.unsafe_get neibmult (k+j-1) in
	let ponderationinit () =
	  base := 1;
	  for i=0 to Array.length neibmult - 1 do
	    Array.unsafe_set neibmult (i) !base;
	    base := !base * q
	  done
	in
	let minit px py =
	  (
	    acc := 0;
	    for i = 0 to k-1 do
	      for j = 0 to k-1 do
		acc := !acc + (mgetxy i j)*(vgetfun (i-r) (j-r) px py)
	      done
	    done
	  ) in
	let vinit px py =
	  (
	    acc := 0;
	    for i = 0 to k-1 do
	      acc := !acc + (vgetxy i r)*(vgetfun (i-r) (0) px py)
	    done;
	    for j = 0 to r-1 do
 	      for i = r-j to r+j do
		acc := !acc + (vgetxy i j)*(vgetfun (i-r) (j-r) px py)
	      done
	    done;
	    for j = 0 to r-1 do
	      for i = j+1 to 2*r-1-j do
		acc := !acc + (vgetxy i (j+r+1))*(vgetfun (i-r) (j+1) px py)
	      done
	    done;
	  ) in 
	let cinit px py =
	  (
	    acc := 0;
	    for i = 0 to k-1 do
	      acc := !acc + (cgetxy i r)*(vgetfun (i-r) (0) px py)
	    done;
	    for j = 0 to k-1 do
	      if j <> r then (acc := !acc + (cgetxy r j)*(vgetfun (0) (j-r) px py))
	    done
	  ) in 
	let gettable = getter_from_fast_array t in
	let putat px py =
	  let g = (gettable !acc) in
	    (* if g >= q or g<0 then failwith (Printf.sprintf "%d acc=%d %dx%d" g !acc px py) else *)
	    writenext (px) (py) g
	in
	  (* The current value of acc reflects the neighb. of  px,py and we want to update to the neighb. of (px+1),py *)
	let vri px py =
	  (
	    (* first we substract the two left sides of the diamond *)
	    for j = 0 to r-1 do 
 	      acc := !acc - (vgetxy (r-j) j)*(vgetfun (-j) (j-r) px py)
	    done;
	    for j = r-1 downto 0 do
	      acc := !acc - (vgetxy (r-j) (k-1-j))*(vgetfun (-j) (r-j) px py)
	    done;
	    (* then we 'shift' everything and add the new rightmost position *)
	    acc:= !acc /q + (vgetxy (k-1) r)*(vgetfun (r+1) 0 px py);
	    (* then we add the two right sides of the new diamond *)
	    for j = 0 to r-1 do 
 	      acc := !acc + (vgetxy (j+r) j)*(vgetfun (j+1) (j-r) px py)
	    done;
	    for j = r-1 downto 0 do
	      acc := !acc + (vgetxy (j+r) (k-1-j))*(vgetfun (j+1) (r-j) px py)
	    done;
	  ) in
	let cri px py =
	  (
	    acc:= (!acc mod row)/q + (cgetxy (k-1) r)*(vgetfun (r+1) 0 px py);
	    for j = 0 to r-1 do 
	      acc := !acc + (cgetxy r j)*(vgetfun 1 (j-r) px py)
	    done;
	    for j = r+1 to k-1 do
	      acc := !acc + (cgetxy r j)*(vgetfun 1 (j-r) px py)
	    done
	  ) in
	let mri px py =
	  (
	    acc := !acc / row;
	    for j = 0 to k-1 do
	      acc := !acc + (mgetxy (k-1) j)*(vgetfun (r+1) (j-r) px py)
	    done
	  ) in
	  (
	    match ac.neighbourhood with
		Moore _ -> 
		  let (na, nb) = (y+2*r*tott-1, x+2*r*tott-1) in
		    (
		      ponderationinit ();
		      minit 0 0;
		      (* Top: we must be careful with boundary conditions! *)
		      for j = 0 to r-1 do
			minit 0 j;
			for i= 0 to nb-1 do
			  putat i j;
			  mri i j;
			done;
			putat nb j;
		      done;
		      (* Middle for vertical positionning *)
		      for j = r to na-r do
 			minit 0 j;
			for i= 0 to r-1 do
			  putat i j;
			  mri i j;
			done;
			(* TRUE HEART OF THE BIG LOOP *)
			(* No boundary condition to check *)
			for i= r to nb-r-1 do
  			  putat i j;
			  (* Compute next neighb. context without modulo *)
			  acc := !acc / row;
 			  let farx = i+r+1 in
			    for u = 0 to k-1 do
			      acc := !acc + (mgetxy (k-1) u)*(readbefore (farx) (j+u-r))
			    done
			done;
			for i= nb-r to nb-1 do
			  putat i j;
			  mri i j;
			done;
			putat nb j;
		      done;
		      (* Bottom: we must be careful with boundary conditions!  *)
		      for j = na-r+1 to na do
 			minit 0 j;
			for i= 0 to nb-1 do
			  putat i j;
			  mri i j;
			done;
			putat nb j;
		      done
		    )
	      | VonNeumann _ -> 
		  let (na, nb) = (y+2*r*tott-1, x+2*r*tott-1) in
		    (
		      ponderationinit ();
		      (* Top: we must be careful with boundary conditions! *)
		      for j = 0 to r-1 do
			vinit 0 j;
			for i= 0 to nb-1 do
			  putat i j;
			  vri i j;
			done;
			putat nb j;
		      done;
		      (* Middle for vertical positionning *)
		      for j = r to na-r do
 			vinit 0 j;
			for i= 0 to r-1 do
			  putat i j;
			  vri i j;
			done;
			(* TRUE HEART OF THE BIG LOOP *)
			(* No boundary condition to check *)
			for i= r to nb-r-1 do
  			  putat i j;
			  (* Compute next neighb. context without modulo *)
			  for p = 0 to r-1 do 
 			    acc := !acc - (vgetxy (r-p) p)*(readbefore (-p+i) (p-r+j))
			  done;
			  for p = r-1 downto 0 do
			    acc := !acc - (vgetxy (r-p) (k-1-p))*(readbefore (-p+i) (r-p+j))
			  done;
			  acc:= !acc /q + (vgetxy (k-1) r)*(readbefore (i+r+1) j);
			  for p = 0 to r-1 do 
			    acc := !acc + (vgetxy (p+r) p)*(readbefore (p+i+1) (p-r+j))
			  done;
			  for p = r-1 downto 0 do
			    acc := !acc + (vgetxy (p+r) (k-1-p))*(readbefore (p+i+1) (r-p+j))
			  done
			done;
			for i= nb-r to nb-1 do
			  putat i j;
			  vri i j;
			done;
			putat nb j;
		      done;
		      (* Bottom: we must be careful with boundary conditions!  *)
 		      for j = na-r+1 to na do
 			vinit 0 j;
			for i= 0 to nb-1 do
			  putat i j;
			  vri i j;
			done;
			putat nb j;
		      done
		    )
	      | Cross _ -> 
		  let (na, nb) = (y+2*r*tott-1, x+2*r*tott-1) in
		    (
		      ponderationinit ();
		      (* Top: we must be careful with boundary conditions! *)
		      for j = 0 to r-1 do
			cinit 0 j;
			for i= 0 to nb-1 do
			  putat i j;
			  cri i j;
			done;
			putat nb j;
		      done;
		      (* Middle for vertical positionning *)
		      for j = r to na-r do
 			cinit 0 j;
			for i= 0 to r-1 do
			  putat i j;
			  cri i j;
			done;
			(* TRUE HEART OF THE BIG LOOP *)
			(* No boundary condition to check *)
			for i= r to nb-r-1 do
  			  putat i j;
			  (* Compute next neighb. context without modulo *)
			  acc:= (!acc mod row)/q + (cgetxy (k-1) r)*(readbefore (i+r+1) (j));
			  for p = j-r to j-1 do 
			    acc := !acc + (cgetxy r (p+r-j))*(readbefore (i+1) (p))
			  done;
			  for p = j+1 to j+r do
			    acc := !acc + (cgetxy r (p+r-j))*(readbefore (i+1) (p))
			  done
			done;
			for i= nb-r to nb-1 do
			  putat i j;
			  cri i j;
			done;
			putat nb j;
		      done;
		      (* Bottom: we must be careful with boundary conditions!  *)
		      for j = na-r+1 to na do
 			cinit 0 j;
			for i= 0 to nb-1 do
			  putat i j;
			  cri i j;
			done;
			putat nb j;
		      done
		    )
	  )

    (*     | Selfproduct f -> doit f *)
    | _ -> raise (Unimplemented "next conf 2D")
  in
    doit ac.transitionrule
      
let getfilebitsize f = 8*(Unix.stat f).Unix.st_size

(* let addfileinfo ac f =  *)
(*   let fs = getfilebitsize f in *)
(*     (match ac.randominfo with *)
(* 	 None -> (ac.randominfo <- Some (Big_int.big_int_of_int fs)) *)
(*        | Some i -> (ac.randominfo <- Some (Big_int.add_int_big_int fs i))); *)
(*     if ac.srcfile = None then ac.srcfile <- Some f *)
      
let get_src_file ac = ac.srcfile

(* Decide what is kept from old CA/conf *)
let conforac oldac newac =
  let rec doit = function
	Newall -> (newac)
      | Newac -> (newac.theinitconf <- oldac.theinitconf; newac.initconfs <- oldac.initconfs; newac)
      | Newconf -> (oldac.theinitconf <- newac.theinitconf; oldac.initconfs <- newac.initconfs; oldac)
      | NeitherNor -> (
	  (* newac.initconfs <- oldac.initconfs;  *)
	  newac.theinitconf <- oldac.theinitconf;
	  newac.transitionrule <- oldac.transitionrule;
	  newac.binary_power <- oldac.binary_power;
	  newac
	)
      | Blocked x -> doit x
  in doit newac.wanted

(************* CA INTERFACE FUNCTIONS *****************)

let needversion n = if n > theversion then raise (BadACML ("Need version "^(string_of_int n)^
						    " of ACML and this one is "^(string_of_int theversion)))
let giveaname a n =
  a.name <- Some n

let whatsyourname ac = match ac.name with
     None -> "No name"
  | Some n -> n
 
let giveanobjname a n =
  a.obj_name <- Some n

let whatsyourobjname ac = match ac.obj_name with
     None -> "Anonymous"
  | Some n -> n

let whatsyourfile ac = match ac.srcfile with
     None -> "Buit-in"
  | Some n -> "file: "^n

let iscomplex a = a.complexconf

let setcomplex a = a.complexconf <- true

let iscomplete = function
    Acfinished _ -> true
  | _ -> false

let iserror = function
    Acerror _ -> true
  | _ -> false


let geterror = function
    Acerror e -> e
  | _ -> assert false

let percent  = function 
    Acompletion n -> n mod 100
  | _ -> assert false

let getcomplete = function
    Acfinished a -> a
  | _ -> assert false

let check a =
  let rec doit c t =
    match (t, a.dimension, a.binary_power) with
	(Table rv, d, _)  -> (
	  if (secure_intexp false c (arity a) > fast_array_size rv) then
	    raise (BadACML "Bad arity/cardinality or rule definition");
	  a
	)
      | (XTable rv, d, Vcard vc)  -> (
	  if (secure_intexp false (1 lsl vc) (arity a) <> Array.length rv) then
	    raise (BadACML "Bad arity/cardinality or rule definition");
	  a
	)
      | (XTable rv, d, Vboth (vc,_))  -> (
	  if (secure_intexp false (1 lsl vc) (arity a) <> Array.length rv) then
	    raise (BadACML "Bad arity/cardinality or rule definition");
	  a
	)
      | (XTable rv, d, _)  -> (
	  if (secure_intexp false c (arity a) <> Array.length rv) then
	    raise (BadACML "Bad arity/cardinality or rule definition");
	  a
	)
      | (MultisetTable rv, d, Varit k)  -> (
	  if (secure_intexp false (1 lsl k) c <> Bigarray.Array1.dim rv) then
	    raise (BadACML "Bad arity/cardinality or rule definition");
	  a
	)
      | (MultisetTable rv, d, Vboth (_,k))  -> (
	  if (secure_intexp false (1 lsl k) c <> Bigarray.Array1.dim rv) then
	    raise (BadACML "Bad arity/cardinality or rule definition");
	  a
	)
      | (MultisetTable rv, d, _)  -> (
	  if (secure_intexp false (arity a + 1) c <> Bigarray.Array1.dim rv) then
	    raise (BadACML "Bad arity/cardinality or rule definition");
	  a
	)
      | _ -> a
  in 
    debugline Dbgdev "Checking rule size";
    if (arity a) >= Sys.max_array_length  then raise (BadACML "rule with too many neighbours!");
    (* if  a.cardinality >= Sys.max_array_length then raise (BadACML "rule with too many states!"); *)
    debugline Dbgdev "Checking actions";
    purge_actions a;
    debugline Dbgdev "Checking table size (if needed)";
    if (neednewac a) then doit a.cardinality a.transitionrule
    else a
	  
let putinfo mat a = coreacinfo (debugline Dbginteract) mat a

let outputrule ac =
  let rec doit = function
      (* Table v -> (  let y = Array.make (arity ac) 0 in *)
      (* 		    for i = 0 to lastrawindex ac do *)
      (* 		      inttointarray ac y i; *)
      (* 		      Array.iter (fun x -> Printf.fprintf stderr "%d " x) y; *)
      (* 		      Printf.fprintf stderr "--> %d\n" v.{i} *)
      (* 		    done; *)
      (* 		    flush stderr *)
      (* 		  ) *)
    | _ -> raise (Unimplemented "outputrule")
  in doit ac.transitionrule
    
    
let acstringinfo mat a = coreacinfo (fun x -> x) mat a

let start a =
    resetrandom ();
    getseed a;
    if neednewac_unblocked a then (
      a.obj_name <- None;
      a.name <- None;
    );
    if neednewac a then (
      a.meminfo <- 0;
      a.refmeminfo <- ref 0;
    );
    a
      

(* Finalization of the generated CA *)

let finalize subprocess ac =
  (* ( *)
  (*   (\* in case of buit-in evaluation, we restore the previous state of the unwanted component *\) *)
  (*   match ac.wanted with *)
  (* 	Newac -> setnocount ac.seed.rs_conf  *)
  (*     | Newconf -> setcount ac.seed.rs_rule *)
  (*     | _ -> () *)
  (* ); *)
  ac.seed.ordinal <- (ac.seed.ordinal+1);
  (* Check bounds *)
  let check_single = function
      ParInt p -> (p.pvalue <- min p.pmax (max p.pmin p.pvalue))
    | ParChoice p -> if not (List.mem p.pcvalue p.pchoices) then (
	match p.pchoices with
	    [] -> p.pcvalue <- ""
	  |t::q -> p.pcvalue <- t
      )
    | _ -> ()
	  
  in
    (* check that the colormap still exists *)
    if not (List.mem_assoc ac.colormap ac.parameters.colormaps) then (
      debugline Dbgdev "Changing unknown colormap";
      ac.colormap <- fst (List.hd (ac.parameters.colormaps));
    );
    List.iter (fun (n,p) -> check_single p) ac.parameters.params;
    (* Suppress unused params *)
    let pal = param_alive_list () in
    let rec singleocc chk = function
	[] -> []
      | t::q -> if List.mem (fst t) chk then singleocc chk q
	else t::(singleocc ((fst t)::chk) q)
    in
      ac.parameters.params <- singleocc [] (List.filter (fun (n,p) -> List.mem n pal) ac.parameters.params);
      reset_param_alive_list ()
	
(*  Restore/save undo version of random states in the given CA *)

let apply_undo_seeds ac =
  debugline Dbgdev "Restoring conf PRNG state...";
  setnocount ac.undo_seed.rs_conf;
  debugline Dbgdev "Restoring rule PRNG state...";
  setcount ac.undo_seed.rs_rule

let prepare_undo_seeds ac =
  ac.undo_seed.rs_rule <- getcount ();
  ac.undo_seed.rs_conf <- getnocount ()
  
    
(**************** SHORTCUTS *************)
    
let (make_shortcut, try_shortcuts, list_shortcuts, remove_shorcuts) =
  let short_list = ref [] in
    (
      (fun rex nick param_gen -> 
	 debugline Dbgdev ("Adding shorcut '"^rex^"' for '"^nick^"'");
	 short_list := (rex, nick, param_gen)::(!short_list)
      ),
      (fun text -> List.iter (fun (rex,nick,gen) -> if Str.string_match (Str.regexp_case_fold rex) text 0 then (
				debugline Dbgdev ("Shortcut success on regexp '"^rex^"'");
				raise (Shortcut (true, nick, gen text))
			      )
			     ) !short_list
      ),
      (fun () -> !short_list),
      (fun nick -> short_list := List.filter (fun (_,x,_) -> x<>nick) !short_list)
    )


(********************** FAMILY OR PLUGIN ******************)
(* when loading a file, we don't know if it is a plugin or a family until the invocation of a "valid*" command *)

let (last_load_is_family, last_load_is_plugin, set_last_load_to_family,reset_last_load) =
  let last = ref true in
  let fresh = ref true in
    (
      (fun () -> !last),
      (fun () -> not !last),
      (fun b -> if !fresh then (
	 last:= b; fresh:=false
       )),
      (fun () -> fresh:=true)
    )


(**************** BUILTIN OBJECTS MANAGEMENT *************)

type status_type = Human | Zombie

type builtin_rule = {
  mutable nickname : string;
  (* name of file for future save *)
  filename : string;
  mutable fullname : string;
  mutable maker_fun : actype -> actype;
  src : string;
  mutable status : status_type;
  mutable saved_state : saving list;
  mutable objects : string list;
}

let currentTransform:(actype->actype) option ref = ref None

let unik_name tentative liste =
  let name n = if n=1 then tentative else tentative^"_"^(string_of_int n) in
  let rec doit n =
    if List.mem (name n) liste then doit (n+1)
    else name n
  in doit 1
    
let clean_entry x lref =
  let ni,fu = (unik_name x.nickname (List.map (fun x -> x.nickname) lref), 
	       unik_name x.fullname (List.map (fun x -> x.fullname) lref))
  in (
      x.nickname <- ni;
      x.fullname <- fu;
      x
    )
     
let nicks what = Array.map (fun x -> x.nickname) what
let names what = Array.map (fun x -> x.fullname) what

let (is_preface,preset_builtin,in_place_builtin, erase_builtin, current_builtin,
     postset_builtin, add_shortcut, valid, getcurrentbuiltin, get_builtin_source, 
     get_builtin_filename, set_builtin, set_builtin_bynick, 
     set_family_from_nick,
     apply_builtin,
     get_all_combo,
     get_all_nicks, get_all_names,
     get_builtin_nicks, get_builtin_names,
     get_ext_nicks, get_ext_names,
     declare_object, get_objects, get_objects_deluxe,
     default_builtin
    ) =
  let i = ref 0 in
  let ext = ref 0 in
  let preface = ref true in
  let builtin_table = ref [||] in
  let orphan_objects = ref [] in
    (
      (fun () -> !preface),
      (************ preset_builtin **************)
      (fun x -> builtin_table := Array.append [|clean_entry x (Array.to_list !builtin_table)|] !builtin_table; 
	 if not !preface then (i:= 0; incr ext)),
      (************ in_place_builtin **************)
      (fun x -> (!builtin_table).(!i) <- clean_entry x (Array.to_list 
							  (Array.sub !builtin_table 1 (Array.length !builtin_table -1)));
      ),
      (************* erase last entry *********************)
      (fun () -> 
	 debugline Dbgdev ("Removing builtin slot '"^(!builtin_table).(0).nickname^"'");
	 (* First, remove shortcuts if any *)
	 remove_shorcuts (!builtin_table).(0).nickname;
	 (* Then remove table entry *)
	 builtin_table := Array.sub !builtin_table 1 (Array.length !builtin_table - 1); 
  	 if not !preface then decr ext
      ),
      (************* get last entry *********************)
      (fun () -> (!builtin_table).(0)
      ),
      (************ postset_builtin **************)
      (fun () -> i:= 0; preface := false; debugline Dbgdev "Built-in rules prepared."; 
	 Array.iter (fun x -> debugline Dbgdev ("Rule "^x.fullname)) (!builtin_table)
      ),
      (************ add_shortcut ***************)
      (fun rex param_gen ->
	 (
	   debugline Dbgdev ("Adding shortcut: "^(String.escaped rex));
	   if !i < Array.length !builtin_table then 
	     make_shortcut rex (!builtin_table).(!i).nickname param_gen      
	   else debugline Dbgwarning "Shortcut declaration not allowed here"
	 )
      ),
      (************ valid ***************)
      (fun maker ->
 	 ((!builtin_table).(!i).maker_fun <- maker;
	  (* keep the saved state for future reloading *)
	  (!builtin_table).(!i).saved_state <- get_saved ();
	  (!builtin_table).(!i).status <- Human;	  
 	  let a = maker (defaultac_nocompute ()) in
	    (match a.name with
		 None -> ()
	       | Some x -> (!builtin_table).(!i).fullname <- x
	    );
	    reset_last_load ();
	    set_last_load_to_family true;
	 )
      ),       
      (**************** getcurrentbuiltin *****************)
      (fun () -> (!i< !ext, (!builtin_table).(!i))
      ),
      (fun () -> (!builtin_table).(!i).src),
      (fun () -> Filename.basename (!builtin_table).(!i).filename),
      (**************** set_builtin *************************)
      (fun name -> 
 	 let t = !builtin_table in
	 let rec doit p =
	   if p>= 0 then (if t.(p).fullname = name then i:= p else doit (p-1))
	   else debugline Dbgdev ("Builtin selection by fullname failed on '"^name^"'");
	 in doit (Array.length t - 1);
	   force_restore (!builtin_table).(!i).saved_state
      ),
      (**************** set_builtin by nick *************************)
      (fun name -> 
	 let t = !builtin_table in
	 let rec doit p =
	   if p>= 0 then (if t.(p).nickname = name then i:= p else doit (p-1))
	   else debugline Dbgdev ("Builtin selection by nickname failed on '"^name^"'");
	 in doit (Array.length t - 1);
	   force_restore (!builtin_table).(!i).saved_state
      ),
      (**************** set family from nick *************************)
      (fun ac -> 
	 (match ac.nick_name with
	      None -> debugline Dbgdev ("Object without nickname: family unchanged!")
	    | Some name -> (
		let t = !builtin_table in
		let rec doit p =
		  if p>= 0 then (if t.(p).nickname = name then i:= p else doit (p-1))
		  else debugline Dbgdev ("Nickname '"^name^"' not found: family unchanged!");
		in doit (Array.length t - 1)
	      )
	 )
      ),
      (******************** Apply builtin ******************)
      (fun name x -> 
	 let t = !builtin_table in
	 let rec doit p =
	   if p>= 0 then (if t.(p).nickname = name then (
			    force_restore (!builtin_table).(p).saved_state;
			    ignore (t.(p).maker_fun (start x));
			  ) else doit (p-1))
	   else debugline Dbgdev ("Apply builtin by nickname failed on '"^name^"'");
	 in doit (Array.length t - 1)
      ),
      (************* Get ALL combo ******************)
      (fun () -> Array.to_list (Array.mapi (fun n x -> x.nickname, x.fullname, (n>= !ext),
					      (x.saved_state <> []))
				  !builtin_table)
      ),
      (************* ALL nicks/names ******************)
      (fun () -> Array.to_list (nicks (!builtin_table))),
      (fun () -> Array.to_list (names (!builtin_table))),
      (************* BUILTIN nicks/names ******************)
      (fun () -> Array.to_list (nicks (Array.sub !builtin_table !ext (Array.length !builtin_table - !ext)))),
      (fun () -> Array.to_list (names (Array.sub !builtin_table !ext (Array.length !builtin_table - !ext)))),
      (************* EXT nicks/names ******************)
      (fun () -> Array.to_list (nicks (Array.sub !builtin_table 0 !ext))),
      (fun () -> Array.to_list (names (Array.sub !builtin_table 0 !ext))),
      (************** Declare object ********************)
      (fun ?family (s:string) -> match family with
	   None -> (!builtin_table).(!i).objects <- s::((!builtin_table).(!i).objects)
	 | Some f -> orphan_objects := (s,f)::(!orphan_objects)
      ),
      (************** Get objects ********************)      
      (fun () -> Array.fold_left (fun acc x -> List.rev_append x.objects acc) (List.map fst !orphan_objects) !builtin_table),
      (************** Get objects DELUXE ********************)      
      (fun () -> (List.rev_append (Array.fold_left (fun acc x -> if x.saved_state<>[] then 
				      (x.nickname,x.fullname,false)::acc else acc) 
		     [] !builtin_table) (Array.fold_left 
					    (fun acc x -> List.rev_append (List.map (fun a -> (a,x.fullname,(x.saved_state=[]))) 
									     x.objects) acc) 
					    (List.map (fun (a,b) -> (a,b,true)) !orphan_objects)
					    !builtin_table))
      ),
      (************** Default builtin ****************)
      (fun () -> {fullname = ""; nickname = ""; src = ""; filename = ""; 
		  maker_fun = (fun x -> x); saved_state = []; objects=[]; status = Zombie})
    )
      

(***************** PLUGINS ***********************)

type apply_repetition = Apresinitialisation | Apreschaqueobjet | Avantchaqueobjet | Apreschaqueetape | Alademande 
			| Aucrochet of string

type one_plug = {
  pg_name : string;
  pg_src : string;
  pg_function : actype -> unit;
  pg_hook_fun : string -> actype -> int *int -> (int->int->int) -> (int->int->int->unit)-> (bool*bool*bool)*(string list);
  mutable pg_active : bool;
  pg_repetition : apply_repetition;
  pg_description : string;
  pg_start: unit -> unit;
  pg_finish: unit -> unit;
}

let (plugin_mode, enter_pg_mode, leave_pg_mode) =
  let pg =ref false in
    ((fun () -> !pg),
     (fun () -> pg:=true),
     (fun () -> pg:=false)
    )
  
let string_from_repetition = function
    Apresinitialisation -> "at program startup"
  | Apreschaqueobjet -> "after each new object creation"
  | Avantchaqueobjet -> "before each new object creation"
  | Apreschaqueetape -> "after each step"
  | Alademande -> "on user demand"
  | Aucrochet c -> ("on hook regexp '"^c^"'")

let (valid_plugin, valid_hook_plugin, apply_plugins, plugin_hooks, list_plugins, list_hook_plugins, plugin_on, plugin_off) =
  let plugs = ref [] in
    (
      (* valid_plugin *)
      (fun ?activate ?name ~repetition ?description ?start ?finish (maker_fun:actype->unit) ->
	 let builtin = current_builtin () in
	   (* Ensure that we work on a fresh builtin object not already 'validated' as a family *)
	   if builtin.status = Zombie then (
	     let active = (match activate with Some t -> t | _ -> false) in
 	     let nm = (match name with Some n -> n | _ -> builtin.nickname) in
	     let start = (match start with Some s -> s | _ -> (fun _ -> ())) in
	       plugs:= {
		 pg_name = nm;
		 pg_src = builtin.src;
		 pg_function = (fun ac -> 
				  debugline Dbgdev ("Applying plugin '"^nm^"'...");
				  enter_domain_name nm;
				  maker_fun ac;
				  exit_domain_name ()
			       );
		 pg_hook_fun = (fun s ac (h,w) read write -> ((false,false,false),[]));
		 pg_active = active;
		 pg_repetition = repetition;
		 pg_start = start;
		 pg_finish = (match finish with Some f -> f | _ -> (fun _ -> ()));
		 pg_description = (match description with Some d -> d | _ -> "")
	       } :: (List.filter (fun x -> x.pg_name <> nm) !plugs);
	       reset_last_load ();
	       set_last_load_to_family false;
	       if active then start ();
 	       debug Dbgdev ("Plugin '"^nm^"' loaded")
	   ) else debug Dbgwarning "Plugin validation in a file with a family validation!"
      ),
      (* valid_hook_plugin *)
      (fun  ?activate ?name ~pattern ?description ?start ?finish maker_fun ->
	 let builtin = current_builtin () in
	   (* Ensure that we work on a fresh builtin object *)
	   if builtin.status = Zombie then (
	     let active = (match activate with Some t -> t | _ -> false) in
 	     let nm = (match name with Some n -> n | _ -> builtin.nickname) in
	     let start = (match start with Some s -> s | _ -> (fun _ -> ())) in
	       plugs:= {
		 pg_name = nm;
		 pg_src = builtin.src;
		 pg_hook_fun = (fun s ac (h,w) read write -> 
				  debugline Dbgdev ("Applying plugin '"^nm^"'...");
				  enter_domain_name nm;
				  let res = maker_fun s ac (h,w) read write in
				    exit_domain_name ();
				    res
			       );
		 pg_function = (fun ac -> ());
		 pg_active = active;
		 pg_repetition = Aucrochet pattern;
		 pg_start = start;
		 pg_finish = (match finish with Some f -> f | _ -> (fun _ -> ()));
		 pg_description = (match description with Some d -> d | _ -> "")
	       } :: (List.filter (fun x -> x.pg_name <> nm) !plugs);
	       reset_last_load ();
	       set_last_load_to_family false;
	       if active then start ();
 	       debug Dbgdev ("Plugin hook '"^nm^"' loaded")
	   ) else debug Dbgwarning "Plugin validation in a file with a family validation!"
      ),
      (* apply plugins *)
      (fun ac (context:apply_repetition) ->
	 List.iter (fun x -> (if x.pg_repetition = context && x.pg_active then
				(
				  x.pg_function ac;
				)
			     )) !plugs
      ),
      (* plugin hooks *)
      (fun mat curt steps ac txt ->
	 let atleastone = ref false in
	 let reader = universe_reader ac mat curt ac.universe steps in
	 let writer = universe_writer ac mat curt ac.universe steps in
	 let w,h = ac.universe in
	 let regen_rule = ref false in
	 let regen_conf = ref false in
	   enter_pg_mode ();
	   let res = List.fold_left (fun acc x -> if x.pg_active then (
				       match x.pg_repetition with
					   Aucrochet st ->
					     let regexp = Str.regexp st in if Str.string_match regexp txt 0 then (
						 let ((ok,rb,cb),lili) = x.pg_hook_fun txt ac (w,h) reader writer in
						   atleastone := !atleastone || ok;
						   regen_rule:= !regen_rule || rb;
						   regen_conf:= !regen_conf || cb;
						   lili@acc) else acc
					 | _ -> acc
				     ) else acc
				    ) [] !plugs
	   in (leave_pg_mode (); (!atleastone,(!regen_rule, !regen_conf), res))
      ),
      (* list 'true' plugins *)
      (fun () -> List.filter (fun x -> match x.pg_repetition with Aucrochet _ -> false | _ -> true) !plugs),
      (* list hook plugins *)
      (fun () -> List.filter (fun x -> match x.pg_repetition with Aucrochet _ -> true | _ -> false) !plugs),
      (* activate plugin *)
      (fun name -> List.iter (fun x -> if String.lowercase_ascii x.pg_name = String.lowercase_ascii name then (x.pg_active <- true; x.pg_start ())) !plugs),
      (* desactivate plugin *)
      (fun name -> List.iter (fun x -> if String.lowercase_ascii x.pg_name = String.lowercase_ascii name then (x.pg_active <- false; x.pg_finish ())) !plugs)
    )



(*************************************************)
(* OBJECT GENERATION                             *)
(*************************************************)


let getcompiledac pm a =  
  infopm := pm;
  let inside,buin = getcurrentbuiltin () in 
  let s = start a in
    apply_plugins s Avantchaqueobjet;
    let x = buin.maker_fun s in
      apply_plugins x Apreschaqueobjet;
      setrandomcount x;
      (* finalize CA *)
      finalize false x;
      if inside then  x.srcfile <- Some buin.filename;
      x.nick_name <- Some buin.nickname;
      x
      
(*************************************************)

let (setinfoprinter, infomsg) = 
  let printer = ref (debugline Dbguser) in
    ((fun f -> printer:= f),
     (fun s -> if not (is_preface ()) then (!printer) s))

module OrdTab=
  struct
    type t = int * (int array)
    let compare = compare
  end

module TabMap = Map.Make(OrdTab)

exception Stop

let (add_meminfo, set_current) =
  let ac = ref (defaultac ()) in
    (
      (fun n -> (!ac).meminfo <- (!ac).meminfo + n; (!ac).refmeminfo := !((!ac).refmeminfo) + n),
      (fun a -> ac:= a)
    )

(* Add a TabMap cache on functional rules  *)
let addcache ac =
  let funcanswer local =
    let h= ref (TabMap.empty) in
    let up = fun p -> add_meminfo (p + 1) in
      (fun c x -> try
	 TabMap.find (c,x) !h
       with Not_found -> let r = local c x in
	 (h := TabMap.add (c, Array.copy x) r !h; up (Array.length x); r)
      )
  in
    match ac.transitionrule with
	FunArray local  -> ac.transitionrule <- (FunArray (funcanswer (fun c x -> local x) 0))
      | MultisetFunArray local -> ac.transitionrule <- (MultisetFunArray (funcanswer local))
      | _ -> raise NoNeedCache



let is_unsafe a = false


(* generic heuristic cache mechanism for speed *)
(* f can possibly be called two times with same arguments *)
module ParamHT=
  struct
    type t = int * (int array)
    let equal = (=)
    (* this is really arbitrary but seems slightly better *)
    let hash = Hashtbl.hash_param 50 50
  end

module MyHT = Hashtbl.Make(ParamHT)

let speed_cache f =
  (* this is really arbitrary but seems slightly better *)
  let nb = 2 in
  let size = bytes_in_cache in
  let harray = Array.init nb (fun _ -> MyHT.create size) in
  let usage = ref 0 in
  let pos = ref 0 in
  let _ = debugline Dbgdev (Printf.sprintf "adding %d hashtables of size %d" nb size) in
  let _ = add_meminfo (nb*size*bytes_per_word) in
  let rec find i c x =
    try 
      MyHT.find harray.(i) (c,x)
    with Not_found -> (
      if i = ((nb + !pos - 1) mod nb) then (
	let r = f c x in
	incr usage;
	MyHT.add harray.(!pos) (c,(Array.copy x)) r;
	if !usage >= size then (
	  pos := (!pos + 1) mod nb;
	  MyHT.clear harray.(!pos);
	  (* harray.(!pos) <- MyHT.create size; *)
	  usage:= 0;
	);
	r
      ) else find ((i+1) mod nb) c x
    )
  in (fun c x -> find !pos c x)

(* cache ensuring that local rule is never called twice on same arguments *)
let determinization_cache loc =
    let h= ref (TabMap.empty) in
    let up = fun p -> add_meminfo (p + 1) in
    (fun c x -> try
		  TabMap.find (c,x) !h
      with Not_found -> let r = loc c x in
			(h := TabMap.add (c, Array.copy x) r !h; up (Array.length x); r)
    )


(* first argument of local function is the central cell and is not used except for multiset rules *)
let table_from_fun_array suggested_cache ac (loc: int -> int array -> int) =
  let _ = set_current ac in
  let local = ref (if is_unsafe ac then
		     (fun x t -> try
			loc x t
		      with e -> raise (BadACML ("The evaluation of local function of the current object raised this: "^(Printexc.to_string e))))
		   else loc)
  in
  let kernel_cache loc =
    speed_cache (determinization_cache loc)
  in
  let functional_answer () =
    let newlocal = if dont_memorize ac then (
      if allow_cache ac then (match suggested_cache with
				| Some cache -> (debugline Dbgdev "ready to use suggested cache on local rule (if needed)"; cache !local)
				| None -> (debugline Dbgdev "using local rule directly"; !local)
			     )
      else (debugline Dbgdev "using (stochastic) local rule";
	    !local
	   ) 
    )
    else (
      debugline Dbgdev "using determinized local rule";
      kernel_cache !local
    )
    in
      (match ac.transitionrule with
	   Table _ | XTable _ -> ac.transitionrule <- FunArray (newlocal 0)
	 | MultisetTable _ -> ac.transitionrule <- MultisetFunArray newlocal
	 | _ -> internalbug "not a table rule as input to table_from_fun_array"
      )
  in
    if neednewac ac then (
(*       ac.meminfo <- 0; *)
(*       ac.refmeminfo <- ref 0; *)
      if  allow_table ac then
	(
	  if (is_main_program ()) then debugline Dbgdev "Computing transition table from built-in rule..."
	  else debugline Dbgdev "Computing transition table...";
	  let next max_digit base tabsize tb ct =
	    let rec prop i cut = if i = tabsize then raise Stop 
	    else if (Array.unsafe_get tb i) = max_digit then (tb.(i)<-0; 
							      ct:= !ct - ((!ct / cut) mod base) * cut; 
							      prop (i+1) (cut*base))
	    else (Array.unsafe_set tb (i) (Array.unsafe_get tb i + 1); ct := !ct + cut)
	    in prop 0 1
	  in
	  let theend max_digit base tabsize = let rec go_baby_go acc i = if i = tabsize then acc else go_baby_go (max_digit+base*acc) (i+1) in
	    go_baby_go 0 0 in
	  let do_computation setrawindex next theend tabsize truetrans =
	    debugline Dbgdev (Printf.sprintf "computation (end=%d tabsize=%d)" theend tabsize);
	    let x = Array.make tabsize 0 in
	    let i = ref 0 in
	      try
		while true do
 		  if truetrans x then setrawindex !i (fun ccell -> !local ccell x) x;
		  sendinfo !i theend;
		  next x i
 		done;
	      with Stop -> () in
	    try
	      (* first try to promote a multiset rule to a classic rule *)
	      (* IMPORTANT: in this case, the order in which local fun is called on successive possible inputs MUST BE the same as multiset-table *)
	      (match ac.transitionrule with
	      	   MultisetTable _ when ac.cardinality < 6 && getradius ac = 1 -> (
		     try ignore (secure_intexp true (ac.cardinality) (arity ac));
			 (* The promotion is valid, so first we compute fake calls to local function as if it was a multi-set table*)
			 (* IMPORTANT2: in the case of a rule which requires determinism, add a cache to the local rule before any call *)
			 debugline Dbgdev "using determinized local rule";
			 local:= kernel_cache !local;
			 let (truetrans,tabsize,max_digit,base) =
			   ((fun x -> 
			     (arity ac = Array.fold_left (fun acc i -> acc+i) 0 x)
			    ), 
			    ac.cardinality, arity ac, (arity ac +1)) in
			 let setrawindex i f tb = 
			   let rec doit p =
			     if p = -1  then ()
			     else if tb.(p) > 0 then (f p; doit (p-1))
			     else doit (p-1)
			   in
			   doit (ac.cardinality-1)
			 in
			 do_computation setrawindex (next max_digit base tabsize) (theend max_digit base tabsize) tabsize truetrans;
			 debugline Dbgdev "Using non-multiset table...";
			 (* Now, we do the promotion by changing table type and local rule appropriately *)
 			 ac.transitionrule <- (Table (make_fast_array 1 1));
 			 local :=
  	      		   let old_local = !local in
	      		   let cards = Array.make (ac.cardinality) 0 in
	      		     (fun x t ->
	      			for i=0 to (ac.cardinality)-1 do
	      			  Array.unsafe_set cards i 0
	      			done;
	      			for i=0 to (arity ac)-1 do
	      			  Array.unsafe_set cards (Array.unsafe_get t (i)) (Array.unsafe_get cards (Array.unsafe_get t (i)) + 1)
	      			done;
	      			old_local (Array.unsafe_get t (centercell ac)) cards
	      		     )
	      	     with _ -> ()
	      	   )
	         | _ -> ()
	      );
	      (* old table is destroyed if the calibration terminates *)
	      debugline Dbgdev "Calibrating table...";
	      calibrate_table ac;
	      debugline Dbgdev "calibration done";
	      let (truetrans,tabsize) = match ac.transitionrule with 
		  Table _ -> ((fun x -> true), arity ac)
		| XTable _ -> ((fun x -> true), arity ac)
		| MultisetTable _ -> ((fun x -> 
					 (arity ac = Array.fold_left (fun acc i -> acc+i) 0 x)
				      ), 
				      ac.cardinality)
		| _ -> internalbug "not a table rule as input to table_from_fun_array"
	      in
	      let (max_digit,base) =  match (ac.transitionrule, ac.binary_power) with
		  (Table _, _) -> (ac.cardinality-1, ac.cardinality)
		| (XTable _, Vcard vc) -> (ac.cardinality-1, 1 lsl vc)
		| (XTable _, Vboth (vc,_)) -> (ac.cardinality-1, 1 lsl vc)
		| (MultisetTable _, Varit k) -> (arity ac, 1 lsl k)
		| (MultisetTable _, Vboth (_,k)) -> (arity ac, 1 lsl k)
		| (MultisetTable _, _) -> (arity ac, (arity ac +1))
		| _ -> internalbug "incoherent table in table_from_fun_array"
	      in
	      let verif i x = if 0<= x && x<ac.cardinality then x 
	      else raise (BadACML ("Transition "^(string_of_int i)^" is out of range: "
				   ^(string_of_int x)
				   ^" for cardinality "^(string_of_int ac.cardinality)))
	      in
		(* For multiset rules *)
		(* in case of power_two optimization *)
	      let multi_base = match ac.binary_power with
		  Vcard vc | Vboth (vc, _) -> (1 lsl vc)
		| _ -> ac.cardinality
	      in
	      let multi_comput tb i f =
		let rec doit acc p =
		  if p = -1  then acc
		  else if tb.(p) > 0 then doit ((verif i (f p)) + multi_base * acc) (p-1)
		  else doit (multi_base*acc) (p-1)
		in
		  doit 0 (ac.cardinality-1)
	      in 
	      let setrawindex =
		match ac.transitionrule with
		    Table v -> ( fun i f tb ->  setter_from_fast_array v i (verif i (f 0)))
		  | XTable v -> ( fun i f tb ->  Array.unsafe_set v i (verif i (f 0)))
		  | MultisetTable v -> ( fun i f tb ->  v.{i}<- (multi_comput tb i f))
		  | _ -> internalbug "setrawindex"
	      in
	      let next = next max_digit base tabsize  in
	      let theend = theend max_digit base tabsize in
		debugline Dbgdev (Printf.sprintf "max_digit=%d base=%d" max_digit base);
		do_computation setrawindex next theend tabsize truetrans;
 		let calcul = multi_size false ac in
		  ac.meminfo <- calcul;
		  ac.refmeminfo := calcul;
		  debugline Dbgdev "rule computed"
	    with RuleTooLarge x -> (
	      if is_main_program () then (
		functional_answer ();
		debugline Dbgdev ("Rule too large: not using fully-cached table")
	      ) else send (fakeacerror ("Huge CA: " ^(Big_int.string_of_big_int x)^" transitions!"))
	    )
	      | e -> if is_main_program () then raise e else send (fakeacerror (Printexc.to_string e))
	)
      else if is_main_program () then functional_answer ()
      else send (fakeacerror "No functional object can be transmitted when using external toplevel")
    )
      
let table_from_fun_fun ac (loc: (int -> int-> int) -> int) =
  (* first test whether we make a table *)
  let size_ok_for_table a = try ignore (secure_intexp true a.cardinality (arity a)); true with _ -> false in
  if (neednewac ac && allow_cache ac && size_ok_for_table ac) then (
    (* OK, translate into fun_array *)
    debugline Dbgdev "Translating fun_fun to fun_array";
    let new_loc c x = loc (fun a b -> x.(relativecell ac a b)) in
      table_from_fun_array None ac new_loc
  ) 
    (* Not OK, so keep fun_fun *)
  else ac.transitionrule <- (FunFun loc)

let table_from_multiset_fun_fun ac (loc: int -> (int-> int) -> int) =
  (* first test whether we make a table *)
  let c = ac.cardinality in
  let size_ok_for_table a = try 
    	  ignore (secure_intexp false c c);
    true with _ -> false 
  in
  if (neednewac ac && allow_cache ac && size_ok_for_table ac) then (
    (* OK, translate into multiset_fun_array *)
    debugline Dbgdev "Translating (multiset) fun_fun to fun_array";
    let new_loc c x = loc c (fun i -> x.(i)) in
      table_from_fun_array None ac new_loc
  ) 
    (* Not OK, so keep fun_fun *)
  else ac.transitionrule <- (MultisetFunFun loc)

let ordupdate c x =
  c.universe <- (fst c.universe, x)

let absupdate c x =
  c.universe <- (x, snd c.universe)

let cellsinfo c = (string_of_int (fst c)) ^ "x" ^ (string_of_int (snd c))
  
let bidimcells c = c

(* WARNING! the color of a cell does not uniquely depend on its state in the general case *)
let (colorofstate, colorseedreset) = 
  let newshift () =  (neutralrandom 255, neutralrandom 255) in
  let light = 10 in
  let lightstep = 16 in
  let (rop,gop,bop) = (ref (neutralrandom 3), ref (neutralrandom 3), ref (neutralrandom 3)) in
  let shift = ref (newshift ()) in
  let attfun op x = match op with
      0 -> x
    | 1 -> lightstep - x
    | _ -> min x (lightstep/2)
  in
    ((fun ac state reader posx posy -> 
	let rec doit t i =
	  match (get_colormap ac, t) with
	      (Auto, _) -> 
		let att = (i)*lightstep /((ac.cardinality-1)) in 
		let v = (i*255)/(ac.cardinality-1) in 
		  Uni ((attfun !rop att)*(max light ((fst !shift + v) mod (256/lightstep))),
		       (attfun !gop att)*(max light ((snd !shift + v) mod (256/lightstep))),
		       (attfun !bop att)*(max light ((fst !shift + snd !shift + v) mod (256/lightstep))))
	    | (Tabscale t, _) -> 
		let n = Array.length t in
		let where = ((n-1)*i)/(ac.cardinality) in
		let i = ((n-1)*i - (where*(ac.cardinality))) in
		let ((ar,av,ab),(br,bv,bb)) = (t.(where),t.((where+1) mod n)) in
		let vr = (i*(br-ar))/(ac.cardinality-1) in 
		let vv = (i*(bv-av))/(ac.cardinality-1) in 
		let vb = (i*(bb-ab))/(ac.cardinality-1) in Uni (ar+vr,av+vv,ab+vb)
	    | (Graycycle, _) -> let v = ((abs (i-(ac.cardinality/2)))*255)/(ac.cardinality/2+1) in Uni (v,v,v)
	    | (Simple tab, _) -> Uni tab.(i mod (Array.length tab))
	    | (Genscale ((ar,av,ab),(br,bv,bb)), _) -> 
		 let vr = (i*(br-ar))/(ac.cardinality-1) in 
		 let vv = (i*(bv-av))/(ac.cardinality-1) in 
		 let vb = (i*(bb-ab))/(ac.cardinality-1) in Uni (ar+vr,av+vv,ab+vb)
	    | (General f, _) -> f i
	    | (Viewer f,_) -> f posx posy reader
	in
	  if ac.cardinality = 1 then Uni (255,255,255)
	  else doit ac.transitionrule state),
     (fun () -> shift := newshift (); rop := neutralrandom 3; gop := neutralrandom 3; bop := neutralrandom 3))
      
let get_universe_size a =
  match  get_initconf a with
      Sized (w,h,_) -> a.universe <- (w,h); (w,h)
    | _ -> a.universe

let preparematrix force mat ac t =
  let (x,y) = get_universe_size ac in
  let forcing () =
    match ac.dimension with
	2 -> let r = getradius ac in 
	let m1 = (make_fast_matrix (x+2*r*t) (y+2*r*t) (ac.cardinality - 1)) in (
	  mat := (Bidim m1,
		  Bidim (make_fast_matrix (x+2*r*t) (y+2*r*t) (ac.cardinality - 1)));
	  debugline Dbgdev ("universe prepared ("^(fast_matrix_type m1)^")")
	)
      | 1 -> let r = getradius ac in 
        let lastline = ref 0 in
          mat := (Unidim (Array.make_matrix (x+2*r*t) (y) 0),
		  Zerodim lastline)
	    
      | _ -> raise (Unimplemented "higher dimension")
  in
  let getmat = function
    | (Unidim x, _) -> x
    | _ -> internalbug "bizzare matrix"
  in
    if force then forcing ()
    else match (ac.dimension, !mat) with
	(2, (Bidim a, Bidim _)) ->
	  if (x,y)= fast_matrix_size a then () else forcing ()	  
      | (1, (Unidim _, Zerodim _)) -> 
	  if (x=Array.length (getmat !mat)) && (y=Array.length (getmat !mat).(0)) then () else forcing ()
      | _ -> forcing ()

let run ac mat time tott =
  let (x,y) = ac.universe in
  let curt = !time in 
    (    
      incr time;
      match ac.dimension with
	  2 -> (match mat with 
		    (Bidim redmatrix, Bidim bluematrix) -> 
		      (if curt = -1 then newinitconfbidim ac (setter_from_fast_matrix redmatrix) 
			 (getter_from_fast_matrix redmatrix) x y tott
		       else if is_sparse ac then 
			 nextconfbidim ac (redmatrix) (bluematrix) x y (curt+1) tott
		       else 
			 nextconfbidim ac (if curt mod 2 = 1 then bluematrix else redmatrix) (if curt mod 2 = 0 then bluematrix else redmatrix) x y (curt+1) tott;
		       let reader = universe_reader ac mat !time (x,y) tott in
		       let writer = universe_writer ac mat !time (x,y) tott in
		         do_all_measurements ac (!time) x y reader writer
		      )
		  | _ -> assert false)
	| 1 -> (match mat with 
		    (Unidim matri, Zerodim i) -> 

	              (if curt = -1 then (
			 newinitconfunidim ac matri i x y tott;
			 time:= y-1
		       ) else (
			 nextconfunidim ac matri i x y tott;
			 let reader = universe_reader ac mat !time (x,y) tott in
			 let writer = universe_writer ac mat !time (x,y) tott in
			   do_all_measurements ac (!time) x y reader writer
		       ))
		  | _ -> assert false)
	| _ -> raise (Unimplemented "higher dim")
    )

let altermatrix zone c read write a b = function
    Cycle -> write a b ((read a b + 1) mod c)
  | Force x -> write a b (x mod c)
  | Funkmod f -> write a b (f (read a b) zone a b mod c)
  | _ -> debugline Dbgdev "Modification action not allowed here"
  
    
let copymatrix m n =
  let (x,y) = Array.length m, Array.length m.(0) in
    for i = 0 to x - 1 do
      for j = 0 to y - 1 do
	n.(i).(j) <- m.(i).(j)
      done
      done

let copyarray s d pos =
  for i = 0 to Array.length d - 1 do
    d.(i).(pos) <- s.(i)
  done

let clonematrix m =
  let (x,y) = Array.length m, Array.length m.(0) in
  let n = Array.make_matrix x y 0 in
    copymatrix m n; n

let getoption = function
    Some x -> x
  | _ -> assert false


let (memorize,load) =
  let membi = ref None in
  let memsi = ref None in
    ((fun ac mat curt ->
	      match ac.dimension with
		  2 -> (match mat with 
			    (Bidim redmatrix, Bidim bluematrix) -> 
			      (if curt = -1 || is_sparse ac then  membi := Some (clone_fast_matrix redmatrix)
			       else (if curt mod 2 = 0 then membi := Some (clone_fast_matrix redmatrix)
				     else membi := Some (clone_fast_matrix bluematrix)))
			  | _ -> assert false)
		    
		| 1 -> (match mat with
			    (Unidim matrix, Zerodim ts) -> 
			      let cells = ac.universe in
			      let (x,y) = cells in
			      let newmat = Array.make x 0 in (
				  for i=0 to x-1 do
				    newmat.(i) <- matrix.(i).((!ts+1) mod y)
				  done;
				  memsi := Some newmat
				)
			  | _ -> assert false)  
		| _ -> raise (Unimplemented "mem/load Higher dim")
     ),
     (fun ac mat curt tott ->	  
	match ac.dimension with
	    2 -> if !membi <> None then 
		let newmat = getoption !membi in
		  (match mat with 
		       (Bidim redmatrix, Bidim bluematrix) -> 
			 (if curt = -1 || is_sparse ac then  copy_fast_matrix newmat redmatrix
			  else (if curt mod 2 = 0 then copy_fast_matrix newmat redmatrix
				else copy_fast_matrix newmat bluematrix))
		     | _ -> assert false)
		    
	  | 1 -> if !memsi <> None then 
		let newmatsi = getoption !memsi in
		  (match mat with
		       (Unidim matrix, Zerodim ts) -> 
			 let cells = ac.universe in
			 let (x,y) = cells in
			   (
			     ts:= (!ts+1) mod y;
			     copyarray newmatsi matrix !ts;
			     completematrix false ac matrix ts x y tott)
		     | _ -> assert false)  
	  | _ -> raise (Unimplemented "higher dim")
     )
    )
      
let modifspacetime action zone ac mat curt tott =
  let cells = ac.universe in
  let (x,y) = cells in
  let c = ac.cardinality in
  let r = getradius ac in
  let (ax,ay,bx,by) = zone in
  let dynread () = universe_reader ac mat curt cells tott in
  let dynwrite () = universe_writer ac mat curt cells tott in
  let read = dynread () in
  let write = dynwrite () in
  let xma,xmi,yma,ymi = (max ax bx),(min ax bx),(max ay by),(min ay by) in
  let secure g x y = g (min (max xmi x) xma) (min (max ymi y) yma) in
    match ac.dimension with
      | 2 -> (match action with
		  Freemod f -> f (ax,ay,bx,by) (secure read) (secure write)
		| _ -> 
		    for i = xmi to xma do
		      for j = ymi to yma do
			altermatrix zone c read write i j action
		      done
		    done
	     )
      | 1 -> (match mat with
		  (Unidim matrix, Zerodim ts) -> (
		    ts:= (!ts + 1) mod y;
		    (match ac.transitionrule with
			 Sparse (ki, _) -> (
			   debugline Dbgdev "Collecting activity spots...";
			   ac.spots <- SpotSet.empty;
			   for i = 0 to (x+2*r*tott - 1) do
			     if ActivSet.mem ((dynread ()) i 0) ki then ac.spots <- SpotSet.add (Line i) ac.spots
			   done
			 )
			   | _ -> ()
		    );
		    (match action with
			 Freemod f -> f (ax,ay,bx,by) (secure (dynread ())) (secure (dynwrite ()))
		       | _ -> 
			   for i = xmi to xma do
			     altermatrix zone c (dynread ()) (dynwrite ()) i 0 action
			   done
		    );
		    completematrix false ac matrix ts x y tott;
		  )
		| _ -> assert false
	     )
	  
      | _ -> raise (Unimplemented "modifspacetime")
	  

let recomputefuture ac mat tott =
  let (x,y) = ac.universe in
    match (ac.dimension, mat) with
	(1, (Unidim matrix, Zerodim ts)) -> 
	  (ts := 0;
           completematrix false ac matrix ts x y tott)
      | _ -> ()


let whatsyoursource ac = match ac.srcfile with
     None -> get_builtin_source ()
  | Some n -> 
      let rd = open_in n in
      let rec gob s =
	try 
	  let sn = input_line rd in
	    gob (s^"\n"^sn)
	with End_of_file -> s
      in gob ""



	 
