open Debug

exception FatalSubProcessError
exception NothingDone
exception SubProcessError of Kernel.actype * float
exception GotFinalResult of Kernel.actype
exception Secondchance

type execution = OK | ERROR of (string)

(* compilation environment *)

type compilation_environment = 
{ 
  compilation_libs : string;
  compilation_dir : string
}

let default_comp_env () = {compilation_dir = Filename.get_temp_dir_name (); compilation_libs = ""}

let temp_file_env env ext = Filename.temp_file ~temp_dir:(env.compilation_dir) "acml_" ext

let libs_command_line env =
  env.compilation_libs
  (* if env.compilation_libs = "" then "" *)
  (* else "-package "^env.compilation_libs *)

let update_env_from_file env file =
  let reading = open_in file in
  let reg = Str.regexp ".*(\\*[ ]*compilation_options?:\\(.*\\)\\*)" in
  let n = input_line reading in 
    if Str.string_match reg n 0 then (
      let theopt = Str.global_replace reg "\\1" n in
	debugline Dbgdev ("using compilation option '"^theopt^"'");
	{env with compilation_libs = theopt} 
    ) else env
      

let todo_list env = 
  let pref x = Filename.concat env.compilation_dir x in
    [ 
      (pref "debug.mli"), Signature.debug_mli;
      (pref "random2.mli"), Signature.random2_mli;
      (pref "kernel.mli"), Signature.kernel_mli;
      (pref "acml.mli"), Signature.acml_mli;
      (pref "keymap.mli"), Signature.keymap_mli;
    ]
      
(* let std_includes = "str.cmxa"  *)

(* let cmd_mli = List.fold_left (fun acc x -> acc^(Filename.chop_extension (fst x))^".cmi ") std_includes todo_list  *)

  
let tmp_exec env what =
  let dir = Sys.getcwd () in
    try
      Sys.chdir env.compilation_dir;
      let out,fin,err = Unix.open_process_full what (Unix.environment ()) in
      let reg = Str.regexp ".*\\(line [0-9]*\\).*\\(charact.*[0-9]*-[0-9]*\\).*" in
      let rec pump s = try
	  let n = input_line err in 
	  let newn = if Str.string_match reg n 0 then 
	               Str.global_replace reg "At \\1 \\2:" n
	             else n in
	  pump (if String.length s > 0 then s^"\n"^newn else newn)
        with End_of_file -> s
          in
          let errput = pump "" in
          let  res = Unix.close_process_full (out,fin,err) in (
	      Sys.chdir dir;
              (* DEBUG INFO *)
              debugline Dbgdev errput;
              (* DEBUG INFO *)
	      match res with
	        Unix.WEXITED 0 -> OK
	      | _ -> ERROR errput
	    )
    with e -> (Sys.chdir dir; raise e)
            
let produce_mli =
  let generated = ref [] in
    ( fun env name string_gen ->
	if not (List.mem name !generated) then 
	  let fout = open_out name in (
	      debugline Dbgdev ("Writing interface source: "^name);
	      output_string fout (string_gen ());
	      close_out fout;
	    )
    )

let string_of_src_file f =
  let fread = open_in f in
  let rec add_line acc =
    try
      let l = input_line fread in
	add_line (acc^l^"\n")
    with End_of_file -> acc
  in 
  let res = add_line "" in (close_in fread; res)


let (last_file,set_last_file) =
  let last = ref "" in
    (
      (fun () -> assert (String.length !last > 0); !last),
      (fun s -> last:=s)
    )

let rec dynlinker env make_preset file =
  try
    let src = Filename.check_suffix file ".ml" in
    let dynlinkable = if src then (temp_file_env env ".cmxs") else file in
      if not Dynlink.is_native then raise (Kernel.Unimplemented "bytecode compilation/dynlink");
      set_last_file file;
      if  src then
	(
	  debugline Dbgdev ("'"^file^"' considered as a source file");
	  let src_file = if Filename.is_relative file then Filename.concat (Sys.getcwd ()) file else file in
	  let tmp_src = (temp_file_env env ".ml") in
	  let env = update_env_from_file env src_file in
	  let libs = libs_command_line env in
	  let cmd="cp \""^src_file^"\" "^tmp_src^" && ocamlfind ocamlopt -package sdl "^libs
	    ^" -o "^(dynlinkable)^" -shared  debug.mli random2.mli kernel.mli acml.mli keymap.mli "^tmp_src in (
	      List.iter (fun (n,g) -> produce_mli env n g) (todo_list env);
	      debugline Dbgdev ("Invoking ocaml compilation: "^cmd);
	      if !Interactive.interactive then Interactive.setworkingstatus "Compiling...";
		match tmp_exec env cmd with
		  OK -> debugline Dbgdev "Compilation finished successfully"
		| ERROR (message) ->
		    if !Interactive.interactive &&
		      Interactive.retryorabort ("Compilation error on file '"^file^"'") message then (
			dynlinker env make_preset file; raise Secondchance)
		    else ( 
		      debugline Dbgwarning ("Compilation finished with an error");
		      debugline Dbgwarning ("Message:\n"^message);
 		      raise FatalSubProcessError   
		    )
	    )
	)
      else if Filename.check_suffix file ".cmxs" then 
	debugline Dbgdev ("'"^file^"' considered as a compiled dynlinkable file")
      else (
	debugline Dbgwarning ("Unknown file type: '"^file^"'");
	raise NothingDone
      );
      let nick = Filename.chop_extension (Filename.basename file) in
	try
	  (* this is uggly but we don't know in advance whether it is a family or a plugin *)
	  (if make_preset then Kernel.preset_builtin else Kernel.in_place_builtin) 
 	    {(Kernel.default_builtin ()) with
	       Kernel.nickname=nick; 
	       Kernel.filename=file; 
 	       Kernel.src= if src then string_of_src_file file else "no source given"
	    };
	  Kernel.force_restore [];
	  Dynlink.loadfile dynlinkable;
	  (* Remove from builtin list if it was a plugin *)
	  if Kernel.last_load_is_plugin () then Kernel.erase_builtin ();
	  (* Now we reset the plugin/family indicator for next time *)
	  Kernel.reset_last_load ()
	with
	    Dynlink.Error e->(debugline Dbgwarning ("Dynlink error: "^(Dynlink.error_message e)); 
			      (* Erase rule slot only if first compilation *)
			      if make_preset then Kernel.erase_builtin ();
  			      raise (Dynlink.Error e))
  with Secondchance -> ()

let loadDynlink env = dynlinker env true
let revert_from_file () = dynlinker (default_comp_env ()) false (last_file ())
