open Debug

exception SaveError
exception NoFile
exception RescueCA of Kernel.actype

(* let interactive = ref true *)
(* let _ = Interactive.interactive:= true *)

let desinteraction () = (
  (* interactive:=false; *)
  Interactive.interactive:=false;
)

let acmlwelcome = "ACML - Automates Cellulaires en caML"
let defaultshot = "screenshot.png"

let versioninfo () =
    debugline Dbginteract acmlwelcome;
    debugline Dbginteract HELP.compiled;    
    debugline Dbginteract "details on source code";
    List.iter (debugline Dbginteract) HELP.components

let licenceinfo () =
  debugline Dbgeneral LICENCE.text

let comboinfo cat s =
  debugline cat s;
  if !Interactive.interactive then (
    Interactive.setstatus s
  )
    

let getoption = function
    Some x -> x
  | _ -> assert false

let putwaitsighandler () =
  Sys.set_signal Sys.sigalrm (Sys.Signal_handle (fun x -> Interactive.waittick ()))

let putvoidsighandler () =
  Sys.set_signal Sys.sigalrm (Sys.Signal_handle (fun x -> ()))

let allow_siginterrupt () =
  Sys.set_signal Sys.sigint (Sys.Signal_handle (fun x -> debugline Dbgwarning "Interrupted!"; exit 0))

let interactiveinfo x y = debug Dbginteract ("Click at "^(string_of_int x)^"x"^(string_of_int y)^"\n")


let rec debuginfo e =
  try 
    let s = (input_line e) in
      debugline Dbgsrce s;
      debuginfo e
  with End_of_file -> ()

let rec outputinfo e =
  try 
    let s = (input_line e) in
      debugline Dbgsrco s;
      outputinfo e
  with End_of_file -> ()

let errorinfo s =
  debugline Dbgwarning ("Problem in CA/conf generation subprocess: '"^s^"'")
    
    
let readinfo () =
  comboinfo Dbgwarning "Reading aborted!"

let sadendinfo () =
  comboinfo Dbgwarning "Reading process still pending!"


let paramitem_to_string (s,par) =
  match par with
      Kernel.ParInt p ->  Printf.sprintf "\"%s\" => integer (default= %d; min= %d; max= %d; value= %d)"
	s p.Kernel.pdefault p.Kernel.pmin p.Kernel.pmax p.Kernel.pvalue
    | Kernel.ParFloat p ->  Printf.sprintf "\"%s\" => integer (default= %g; min= %g; max= %g; value= %g)"
	s p.Kernel.pfdefault p.Kernel.pfmin p.Kernel.pfmax p.Kernel.pfvalue
    | Kernel.ParBool p ->  Printf.sprintf "\"%s\" => boolean (default= %s; value= %s)" 
	s (Kernel.showbool p.Kernel.pbdefault) (Kernel.showbool p.Kernel.pbvalue)
    | Kernel.ParString p ->  Printf.sprintf "\"%s\" => string (default= %s; value= %s)" 
	s p.Kernel.psdefault p.Kernel.psvalue
    | Kernel.ParChoice p ->  Printf.sprintf "\"%s\" => list of choices (choices= %s; value= %s)" 
	s (Kernel.showstrlist p.Kernel.pchoices) p.Kernel.pcvalue
    
let param_append_doc a name = match (Kernel.doc_about a name) with
    None -> ""
  | Some s -> " --- "^s
  

let param_info a =
  debugline Dbginteract "Description: ";
  List.iter (function (title,explanation) -> debugline Dbgitem (title^": "^explanation)) a.Kernel.description;
  debugline Dbginteract "Parameters: ";
  List.iter (fun x -> debugline Dbgitem ((paramitem_to_string x)^(param_append_doc a (fst x))))
    a.Kernel.parameters.Kernel.params;
  debugline Dbginteract "Views: ";
  List.iter (fun x -> debugline Dbgitem (fst x)) a.Kernel.parameters.Kernel.colormaps;
  debugline Dbginteract "Actions: ";
  List.iter (fun x -> debugline Dbgitem ((fst x)^" ("^(Kernel.kind_of_action (snd x))^")")) a.Kernel.parameters.Kernel.actions

let identify a =
  let koi = (Kernel.listacinfo a) in
  (List.fold_left (fun acc x -> acc^" / "^x) (List.hd koi) (List.tl koi))^" / "^
    (match a.Kernel.determinization with
      Kernel.Stochastic -> "stochastic"
    | _ -> "deterministic"
    )^"\n"^
    "Family: "^(Kernel.whatsyourname a)^"\n"^
    "Object: "^(Kernel.whatsyourobjname a)^"\n"^
    (List.fold_left (fun acc desc -> acc^(fst desc)^": "^(snd desc)^"\n") "" a.Kernel.description)


let giveinfo mat ac =
  let sa = match (Kernel.neednewac ac, Kernel.neednewconf ac) with
      (true,true) -> ""
    | (true,false) -> "(same init conf)"
    | (false,true) -> "(same rule)"
    | (false,false) -> "(same rule + init conf)" in
    comboinfo Dbginteract ("new object "^sa);
    Kernel.putinfo mat ac


(******************************************)
(* UNDO mechanism                         *)
(******************************************)

let (pushsos, popsos, clean_history, history_memsize, history_size, history_set_keep, history_keep) =
  let soslist = ref [] in
  let keep = ref false in
  let history_size = ref 0 in
    ((fun  ac ->
	let add = Kernel.memsize ac in
	  (* Suppress old items if no keep mode  *)
 	  (if (not !keep) && (List.length !soslist) > 4 then
	      let rec purge i acc l = if i = 0 then List.rev acc else match l with 
		  [] -> List.rev acc
		| t::q -> purge (i-1) (t::acc) q
	      in soslist:= purge 4 [] !soslist;
	      history_size := List.fold_left (fun acc x -> acc + Kernel.memsize (fst x)) 0 !soslist
	  );
	  (* second argument for timestamp but not implemented yet *)
	  soslist := (Kernel.cloneac ac, 0)::(!soslist);
 	  history_size := !history_size + add;
	  Kernel.set_current ac		
     ),
     (fun  destructive ->
	(match !soslist with 
	    [] ->  None
	  | [x] -> if destructive then None else (soslist:=[];
						  history_size := !history_size - (Kernel.memsize (fst x));
						  Kernel.correct_meminfo (fst x);
						  Some (x))
	  | a::(b::q) -> if destructive then (soslist := q; 
					      history_size := !history_size - (Kernel.memsize (fst a))
					      - (Kernel.memsize (fst b));
					      Kernel.correct_meminfo (fst b);
					      Some (b))
	    else (soslist := b::q; 
		  history_size := !history_size - (Kernel.memsize (fst a));
		  Kernel.correct_meminfo (fst a);
		  Some (a)))
     ),
     (fun () -> soslist := []; history_size := 0),
     (fun () -> (List.length !soslist, !history_size)),
     (fun () -> List.length !soslist),
     (fun b -> keep:=b),
     (fun () -> !keep)
    )
      


(**********************************************************************************)
(* Eval transformation function to generate en acml object                        *)
(**********************************************************************************)



let rec evalacml curac newtype progressmeter =
  let doit automaton =
    let _ = Kernel.set_current automaton in
      (* Apply the currently used transformation function *)
    let x = Kernel.check (Kernel.getcompiledac progressmeter automaton) in (
	debugline Dbgdev "New object generated.";
	let x = Kernel.conforac curac x in
	  debugline Dbgdev "Saving seeds for future undo...";
	  Kernel.unblock x;
	  Kernel.prepare_undo_seeds x;
	  x
      )
  in
  let automaton = Kernel.prepareseed curac newtype
  in
    (
      if !Interactive.interactive then (
	Interactive.setworkingstatus "Computation...";
	Interactive.working ()
      );
      try 
	doit (Kernel.pseudoreset (Kernel.cloneac automaton))
      with Load.SubProcessError (x,mt) -> (sadendinfo (); let nca = (Kernel.check x) in 
					   let nca = Kernel.conforac curac nca in
					     (* pushsos nca; *) nca)
    )

(******************************************)
(* Save ACml file                         *)
(******************************************)

let savedinfo name ac =
  debugline Dbginteract ("File "^name^" successfully saved!");
  if !Interactive.interactive then Interactive.setstatus ("File "^ name ^ " successfully saved!")


(* save as a new source file with preamble containing: 
   1/ parameters
   2/ PRNG seed
   3/ Universe size
   4/ Some info about program version to anticipate backward compatibility
*)
let saverule b sv filename ac =
  let (name,dep) = sv in
  let today = Unix.gmtime (Unix.time ()) in
  let gmdate = Printf.sprintf "y%d m%02d d%02d %02dh%02d:%02d UTC" 
    (1900 + today.Unix.tm_year) (today.Unix.tm_mon + 1) today.Unix.tm_mday 
    (today.Unix.tm_hour) (today.Unix.tm_min) (today.Unix.tm_sec) in
  let seedstring = "[\nKernel.SavParams ["^(List.fold_left (fun acc x -> acc^(Kernel.param_to_string x)^"; ") ""
				       ac.Kernel.parameters.Kernel.params)
    ^"];\nKernel.SavSeed "^(Kernel.seed_ordinal_zero_to_string ac.Kernel.seed)
    ^";\nKernel.SavUniverse "^(Kernel.universe_to_string ac.Kernel.universe)
    ^";\nKernel.SavObjname \"SAVED "^gmdate^"\""
    ^";\nKernel.SavKernelSig (\""^(fst (Signature.sig_fun "kernel"))^"\","
    ^(string_of_float (snd (Signature.sig_fun "kernel")))^")"
    ^";\nKernel.SavRandom2Sig (\""^(fst (Signature.sig_fun "random2"))^"\","
    ^(string_of_float (snd (Signature.sig_fun "random2")))^")"
    ^";\nKernel.SavAcmlSig (\""^(fst (Signature.sig_fun "acml"))
    ^"\","^(string_of_float (snd (Signature.sig_fun "acml")))^")"
    ^"\n]\n"
  in
  let ithname i = if i < 0 then name else (
    if Filename.check_suffix name ".ml" then ((Filename.chop_suffix name ".ml")^"_"^(string_of_int i)^".ml")
    else (name^"."^(string_of_int i))
  ) in
  let newname = 
    let i = ref dep in 
      (
	if !Interactive.interactive then putwaitsighandler ();
	while Sys.file_exists (ithname !i) do
	  incr i
	done;
	(ithname !i)) in
  let flux = open_out newname in
    Printf.fprintf flux ("(*** SAVED %s -- modify at your own risk ***)\n\nlet _ = Kernel.restorestate %s;;\n(**************************************************)\n")
      gmdate seedstring;
    flush flux;
    if b then output_string flux (Kernel.get_builtin_source ());
    close_out flux;
    if not b then (
      let i = (Unix.open_process_in ("cat "^filename^" >> "^newname)) in
	match Unix.close_process_in i with
	    Unix.WEXITED 0 -> putvoidsighandler ()
	  | _ -> putvoidsighandler (); raise SaveError
    );
    savedinfo newname ac
    


(******************************************)
(* Save screenshot                        *)
(******************************************)


let imgsavedinfo name ac =
  debugline Dbginteract ("Screenshot successfully saved in file "^name^"!");
  if !Interactive.interactive then Interactive.setstatus ("Screenshot successfully saved in file " ^ name ^"!")

(* let rgbify r g b =  {Images.r=r; Images.g=g; Images.b=b} *)
(* let maker = Rgb24.make *)
(* let setter = Rgb24.set *)

    
let saveimg shooting filename ac mice curt steps = 
  let lenom =
    if !Interactive.interactive then putwaitsighandler ();
    let base = Filename.chop_extension filename
    and pos = String.rindex filename '.'  in
    let ext = String.sub filename pos ((String.length filename) - pos) in
      if shooting then (base^"_"^(Printf.sprintf "%06u" curt)^ext)
      else if Sys.file_exists filename then
	try
	  let i = ref 0 in
	    (
	      while Sys.file_exists (base^"_"^(string_of_int (!i))^ext) do
		incr i
	      done;
	      (base^"_"^(string_of_int !i)^ext))
	with _ -> raise SaveError
      else filename in
    (
      (if !Interactive.interactive then 
	Interactive.wysiwyg lenom
      else let (w,h)= ac.Kernel.universe in
	Interactive.raw_gray lenom (Kernel.universe_reader ac mice curt ac.Kernel.universe steps) ac w h;
      );
      imgsavedinfo lenom ac;
      putvoidsighandler ();
    )
    

(******************************************)
(* Speed meter                            *)
(******************************************)

        
let (dofpsinfo, resetfps) =
  let base = ref (1 lsl 4) in
  let cnt = ref !base in
  let slow = 1000 in
  let fast = 500 in
  let getthetime  = Interactive.sdlticks in
  let prectime = ref (getthetime ()) in
    ((fun theref ->  if !cnt = 1 then (let newtime = getthetime () in
					(if newtime - !prectime < fast then (
					   cnt := !base;
					   base := !base lsl 1
					 ) 
					 else (
					  theref := (float_of_int (!base*1000)) /. (float_of_int (newtime - !prectime));
					  prectime := newtime;
					  if newtime - !prectime > slow then (base := max 1 (!base lsr 1));
					  cnt := !base
					))
				      )
      else decr cnt
     ),
     (fun theref -> base := 1 lsl 4; theref := 0. ; cnt := !base; prectime := getthetime ())
    )

let (timetrap, testtimetrap) =
  let getthetime () = Interactive.sdlticks () in
  let trap = ref false in
  let timeref = ref 0 in
  let todo = ref (fun () -> ()) in
    ((fun t newtodo -> (trap := true; todo := newtodo; timeref := getthetime () + t)),
     (fun () -> 
	if !trap then (
	  let g = getthetime () in
	  if g >= !timeref then (trap := false; !todo ()))
     ))

let timerwait () = Unix.pause ()

let sysdeprandominit () =
  let tim = (Unix.localtime (Unix.time ())) in
  let pid = Unix.getpid () in
  let x = [|tim.Unix.tm_sec; pid; tim.Unix.tm_min; tim.Unix.tm_mday|] in
    Random2.full_init x;
    Kernel.hasard ();
    Kernel.initneutralrandom [|x.(1); x.(2); x.(0); x.(3)|];
    debugline Dbgdev (Printf.sprintf "random state initialized (%d,%d,%d,%d)" x.(0) x.(1) x.(2) x.(3))





(******************************************)
(* main                                   *)
(******************************************)

let main () =
  (*   let markcolor = ref (200,50,50) in *)
  let savfile = ref None in
  let imgfile = ref None in
  let steps = ref 0 in
  let fps = ref 0. in
  let exit_not_asked = ref true in
  let time = ref 0 in
  let showpercent = ref true in
  let fake_sdl = ref false in
  let run = ref false in
  let resetmatonnewac = ref false in
  let onestep = ref false in
  let clearonce = ref false in
  let autosave = ref false in
  let popupsave = ref false in
  let shooting = ref false in
  let intervalshoot = ref 0 in
  let initconfset = ref false in
  let shootcnt = ref 0 in
  let fullscree = ref false in
  let params = ref [] in
  let something_loaded = ref false in
  let cmdline_params = ref [] in
  let ac = ref (Kernel.defaultac ()) in
  let percentprogression =
    let last = ref (-1) in (fun () ->
        if !Kernel.stoptime >0 && !showpercent then
          let pct = (100 * !time) / !Kernel.stoptime in
          if pct <> !last then 
            (
              last := pct;
              debugline Dbginteract (Printf.sprintf "step %d/%d (%d%%)" !time !Kernel.stoptime pct)
            )
      )
  in
  let theprogressmeter () = if !Interactive.interactive then (fun i -> ignore (Interactive.pollInteraction !ac true); 
    Interactive.progress i) else (textprogressmeter ())
  in
  let matrice = ref (Kernel.defaultmatrice ()) in
  
  (* All params command-line values are taken as strings and conversions are done after *)
  let flipflopparams params =
    let setval = ref false in
    (function s -> 
      if !setval then (
	setval := not !setval;
	params := (fst (List.hd !params),s)::(List.tl !params)
      )
      else (setval := not !setval; 
	    params := (s,"")::(!params))
    ) in
  let newstart () =
    (* Generate new init conf *)
    time := -1;
    shootcnt:=0;
    Kernel.run !ac !matrice time !steps;    
  in

  (* Compute the filename and call the saving procedure *)
  let dosave sav = 
    let fcpl f = (match sav with 
	Some x -> (x, -1) 
      | None ->  (f, -1)) in
    saverule true (fcpl (Kernel.get_builtin_filename ())) "" !ac
  in

  let dostartstop () = 
    (
      run := not !run;
      resetfps fps;
      if !run then (
	debugline Dbgdev "START";
	Interactive.setstatus "Started";
	Interactive.startstopwilldostop ();
	debugline Dbgdev (Kernel.acstringinfo !matrice !ac);
	Interactive.quickrefresh !ac !matrice !time !steps !run
      )							      
      else (
	debugline Dbgdev "STOP";
	Interactive.setstatus "Stopped";
	Interactive.startstopwilldostart ();
	(* 	Interactive.quickrefresh !ac !matrice !time !steps !run *)
	Interactive.refresh true !ac !matrice !time !fps !steps !run;		    
      )
	
    )
  in let doundo x =
       (
	 resetfps fps;
	 ac := fst x;
	 pushsos !ac;
	 Kernel.apply_undo_seeds !ac;
	 Kernel.set_family_from_nick !ac;
	 Kernel.preparematrix true matrice !ac !steps; 
	 if !autosave then dosave !savfile;
	 newstart ();
	 if !Interactive.interactive then
	   (
	     Interactive.geometry !ac;
	     Interactive.colorcache !ac;
	     Interactive.puttitle (history_memsize ()) !ac;
	     Interactive.show_params !matrice !ac;
	     Interactive.refresh true !ac !matrice !time !fps !steps !run;		    
	     Interactive.unworking ()
	   );
       )      
     in let donewneno () =
	  (
	    resetfps fps;
	    try
	      ac := evalacml !ac Kernel.NeitherNor (theprogressmeter ());
	      pushsos !ac;
	      (* Kernel.preparematrix !resetmatonnewac matrice !ac !steps;  *)
	      (* debugline Dbgdev "Matrix prepared"; *)
	      (* (\* For 1D case only: future steps must be computed according to new CA*\) *)
	      (* Kernel.recomputefuture !ac !matrice !steps; *)
	      (* debugline Dbgdev "Future recomputed"; *)
	      (* if !autosave then dosave !savfile; *)
	      if !Interactive.interactive then
		(
		  Interactive.colorcache !ac;
		  Interactive.puttitle (history_memsize ()) !ac;
		  Interactive.show_params !matrice !ac;
		  Interactive.refresh true !ac !matrice !time !fps !steps !run;		    
		  Interactive.unworking ()
		);
	      giveinfo !matrice !ac
	    with e -> (
	      comboinfo Dbgwarning ("Nothing new was generated: "^(Printexc.to_string e)); 
	      debugline Dbgdev "Reverting to previous state";
	      match popsos false with
 		  None ->  internalbug "nothing to revert to after a fatal eval error" 
		| Some x -> doundo x
	    )
      )
  in let donewinit obeyinterface =
      (
	resetfps fps;
	if !Interactive.interactive  && obeyinterface then Interactive.get_universe !ac;
	try
	  if Kernel.iscomplex !ac then 
 	    ac := evalacml !ac Kernel.Newconf (theprogressmeter ())
	  else (if !Interactive.interactive then
		  Interactive.working ());
	  pushsos !ac;
	  Kernel.preparematrix false matrice !ac !steps; 
	  debugline Dbgdev "Matrix prepared";
	  newstart ();
	  if !Interactive.interactive then (
	    Interactive.geometry !ac;
	    Interactive.puttitle (history_memsize ()) !ac;
	    Interactive.show_params !matrice !ac;	    
	    Interactive.refresh true !ac !matrice !time !fps !steps !run;		    
	    Interactive.unworking ();
	    if not (Kernel.iscomplex !ac) then 
	      Interactive.setstatus "New init conf successfully generated!";
	  );
	  giveinfo !matrice !ac
	with e -> (
	  comboinfo Dbgwarning ("Nothing new was generated: "^(Printexc.to_string e)); 
	  debugline Dbgdev "Reverting to previous state";
	  match popsos false with
 	      None -> internalbug "nothing to revert to after a fatal eval error"
	    | Some x -> doundo x
	)
      )
  in let donewboth  obeyinterface =
      (
	resetfps fps;
	if !Interactive.interactive  && obeyinterface then Interactive.get_universe !ac;
	try 
	  ac := evalacml !ac Kernel.Newall (theprogressmeter ());
	  pushsos !ac;
	  Kernel.preparematrix !resetmatonnewac matrice !ac !steps; 
	  debugline Dbgdev "Matrix prepared";
	  if !autosave then dosave !savfile;
	  newstart ();
	  if !Interactive.interactive then
	    (
	      Interactive.geometry !ac;
	      Interactive.colorcache !ac;
	      Interactive.puttitle (history_memsize ()) !ac;
	      Interactive.show_params !matrice !ac;	    
	      Interactive.refresh true !ac !matrice !time !fps !steps !run;		    
	      debugline Dbgdev "refresh done";
	      Interactive.unworking ()
	    );
	  giveinfo !matrice !ac
	with e -> (
	  comboinfo Dbgwarning ("Nothing new was generated: "^(Printexc.to_string e)); 
	  debugline Dbgdev "Reverting to previous state";
	  match popsos false with
 	      None -> internalbug "nothing to revert to after a fatal eval error"
	    | Some x -> doundo x
	)
      )
  in let do_addselect env is_file bynick name =
      if is_file then (
	Load.loadDynlink env name;
	if !Interactive.interactive then Interactive.show_plugins !ac
      )
      else (
	(if bynick then Kernel.set_builtin_bynick else Kernel.set_builtin) name;
	Kernel.set_last_load_to_family true; Kernel.reset_last_load ();
      )
	
  in let doopen obeyinterface =
      if Kernel.last_load_is_plugin () then (
	if !Interactive.interactive then Interactive.show_plugins !ac;
	donewneno ();
	comboinfo Dbginteract "plugin OK"
      ) else
	(
	  (* reset ordinal counter, parameters, etc *) 
	  Kernel.makeinitial !ac;
	  (* inject parameters if needed - reverse order because inject_param destroys previously set values *)
	  debugline Dbgdev (Printf.sprintf "%d parameters to be injected" (List.length !params));
	  Kernel.inject_params !ac (List.rev !params);
 	  resetfps fps;
	  if !Interactive.interactive && obeyinterface then Interactive.get_universe !ac;
	  try
	    ac := evalacml !ac Kernel.Newall (theprogressmeter ());
	    pushsos !ac;
	    Kernel.preparematrix !resetmatonnewac matrice !ac !steps; 
	    debugline Dbgdev "Matrix prepared";
	    if !autosave then dosave !savfile;
	    newstart ();
	    if !Interactive.interactive then
	      (
		Interactive.geometry !ac;
		Interactive.colorcache !ac;
		Interactive.puttitle (history_memsize ()) !ac;
		Interactive.show_params !matrice !ac;	    
		Interactive.show_plugins !ac;
		Interactive.refresh true !ac !matrice !time !fps !steps !run;		    
		debugline Dbgdev "refresh";
		Interactive.unworking ()
	      );
	    giveinfo !matrice !ac
	  with e -> (
	    comboinfo Dbgwarning ("Nothing new was generated : "^(Printexc.to_string e));
	    debugline Dbgdev "Reverting to previous state";
	    match popsos false with
 		None -> raise (Kernel.AbortRule "aborting.")(* internalbug "nothing to revert to after a fatal eval error" *)
	      | Some x -> doundo x
	  )
	)

  in let donewcolormap frominterface =
      (
	Interactive.working ();
	if frominterface then (
	  Interactive.gtk_get_colormap !ac
	) else (
	  let thecolormaps = List.map fst (!ac).Kernel.parameters.Kernel.colormaps in
	  let cur = (!ac).Kernel.colormap in
	  let rec next_colormap  = function
	    | [] -> List.hd thecolormaps
	    | t::[] -> List.hd thecolormaps
	    | t::q -> if t=cur then List.hd q else next_colormap q
	  in (!ac).Kernel.colormap <- next_colormap thecolormaps;
	  if frominterface then Interactive.gtk_set_colormap !ac;
	);
	Interactive.colorcache !ac;
	Interactive.refresh true !ac !matrice !time !fps !steps !run;		    
	(* 	comboinfo Dbginteract "New colors"; *)
	Interactive.unworking ()
      ) in
  let dosnapshot ac matrice curt steps =
    (
      match  Kernel.getaction ac Kernel.SrcSnapshot with
	  Kernel.Snapshot howto -> (
	    Kernel.iterate_on_matrix ac matrice curt ac.Kernel.universe steps howto;
	    (* 	    comboinfo Dbginteract "Snapshot done." *)
	  )
	    (* howto (Array.length m) (Array.length m.(0)) (fun i j -> m.(i).(j)) *)
	| Kernel.Wysiwyg -> saveimg !shooting (match !imgfile with None -> defaultshot | Some f -> f) ac matrice curt steps
	| _ -> comboinfo Dbgwarning "No valid method to do the snapshot!"
    ) in
    (*   let domeasurements ac matrice curt steps = *)
    (*     let (x,y) = ac.Kernel.universe in *)
    (*     let read = Kernel.universe_reader ac matrice curt ac.Kernel.universe steps in *)
    (*     let write = Kernel.universe_writer ac matrice curt ac.Kernel.universe steps in *)
    (*       Kernel.do_all_measurements ac curt x y read write *)
    (*   in *)
  let domodifzone src ax ay bx by =
    (
      debugline Dbgdev (Printf.sprintf "modif zone %dx%d->%dx%d" ax ay bx by);
      let nm = Kernel.getactionname !ac src in     
      let action = Kernel.getaction !ac src in
      Interactive.setstatus (Printf.sprintf "%s on zone %dx%d->%dx%d" nm ax ay bx by);
      Kernel.modifspacetime action (ax,ay,bx,by) !ac !matrice !time !steps;
      Interactive.refresh true !ac !matrice !time !fps !steps !run;
    ) in
  let domodifzonediag src ax ay bx by =
    (
      debugline Dbgdev (Printf.sprintf "modif diag %dx%d->%dx%d" ax ay bx by);
      let nm = Kernel.getactionname !ac src in     
      let action = Kernel.getaction !ac src in
      Interactive.setstatus (Printf.sprintf "%s on diagonal %dx%d->%dx%d" nm ax ay bx by);
      let dx = abs (bx-ax) in
      let dy = abs (by-ay) in
      let sx = if bx >= ax then 1 else -1 in
      let sy = if by >= ay then 1 else -1 in
      for i = 0 to min dx dy do
	let x = ax + sx*i in
	let y = ay + sy*i in
	Kernel.modifspacetime action (x,y,x,y) !ac !matrice !time !steps;
      done;
      Interactive.refresh true !ac !matrice !time !fps !steps !run;
    ) in
  let domodifzoneside src ax ay bx by =
    (
      debugline Dbgdev (Printf.sprintf "modif diag %dx%d->%dx%d" ax ay bx by);
      let nm = Kernel.getactionname !ac src in     
      let action = Kernel.getaction !ac src in
      Interactive.setstatus (Printf.sprintf "%s on diagonal %dx%d->%dx%d" nm ax ay bx by);
      let dx = abs (bx-ax) in
      let dy = abs (by-ay) in
      if dx >= dy then
	for i = min ax bx to max ax bx do
	  Kernel.modifspacetime action (i,ay,i,ay) !ac !matrice !time !steps;
	done
      else
	for i = min ay by to max ay by do
	  Kernel.modifspacetime action (ax,i,ax,i) !ac !matrice !time !steps;
	done;
      Interactive.refresh true !ac !matrice !time !fps !steps !run;
    ) in    
  let random_list l =
    let ki = Kernel.neutralrandom (List.length l) in
    let rec ith n l =
      if n = 0 then List.hd l else ith (n-1) (List.tl l)
    in ith ki l
  in

  let (group_shortcuts,get_shortcuts) =
    let at_least_one = ref false in
    let s = ref "" in
    let s_list = ref [] in
      ((fun a -> 
	  if a="," then (
	    if !at_least_one then s_list := (!s ^ "}")::(!s_list);
	    s:="";
 	    at_least_one := false;
	  )
 	  else if !at_least_one then s:= !s^"}#{"^a 
	  else (s:= "{"^a; at_least_one:=true;)
       ),
       (fun () -> if !at_least_one then (List.rev ((!s ^ "}")::(!s_list))) else (List.rev !s_list))
      )
  in      
  let do_shortcut matrice time steps recup_if_not_found text =
    let rec do_it env text =
      try 
	debugline Dbgdev ("Trying shortcut '"^text^"'");
	
	(* First expand group if needed *)
	let grp = Str.regexp "^{\\(.+\\)}$" in	  
	let (one, regen, liste) = 
	  if Str.string_match grp text 0 then (true, (false,false), Str.split (Str.regexp "}#{") (Str.matched_group 1 text))
	    (* Else try plugin hooks *)
 	  else (Kernel.plugin_hooks matrice time steps (!ac) text) in
	let mix_regen a b = match (a,b) with
	  | (None,_) | (_, None) -> None
	  | Some (a,b), Some (c,d) -> Some ((a || b), (c || d))
	in
	let final_regen = List.fold_left (fun acc x -> mix_regen acc (do_it env x)) (Some regen) liste in
	  (* continue process only if no hook was triggered *)
	  if one then final_regen
	  else (
	    (* Internal shortcuts *)
	    Kernel.try_shortcuts text;
	    List.iter (fun candidate -> if Str.string_match (Str.regexp_case_fold text) candidate 0 then (
			 debugline Dbgdev ("Shortcut success on full-name '"^candidate^"'");
			 raise (Kernel.Shortcut (false, candidate, [])))) (Kernel.get_all_names ());
	    List.iter (fun candidate -> if Str.string_match (Str.regexp_case_fold text) candidate 0 then (
			 debugline Dbgdev ("Shortcut success on short-name '"^candidate^"'");
			 raise (Kernel.Shortcut (true, candidate, [])))) (Kernel.get_all_nicks ());
	    (* External file *)
	    if Sys.file_exists text then (
	      try 
		debugline Dbgdev "Shortcut success on external file";
		(try
		  do_addselect env true false text;
		   comboinfo Dbginteract ("external file '"^text^"' loaded");
		with Load.NothingDone -> comboinfo Dbginteract ("external file '"^text^"' not loaded")
		); None
	      with e -> if recup_if_not_found then (
		comboinfo Dbgwarning ("Problem with external file ("^Printexc.to_string e^"), using randomly chosen builtin rule instead");	  
		raise (Kernel.Shortcut (true,(random_list (Kernel.get_builtin_nicks ())),[]))	  
	      ) else raise e
	    )
	    else if recup_if_not_found then (
	      comboinfo Dbgwarning "Shortcut not recognized, using randomly chosen builtin rule";
	      raise (Kernel.Shortcut (true,(random_list (Kernel.get_builtin_nicks ())),[]))
	    ) 
	    else raise Not_found(* (comboinfo Dbgwarning "Shortcut not recognized"; Some (false,false)) *)
	  )
      with Kernel.Shortcut (bynick, name, param_list) -> (
	do_addselect env false bynick name;
	params := List.append !params param_list;
	None
      )
    in
      let pre_init = List.rev (!ac).Kernel.initconfs in
      let pre_metap = List.rev (!ac).Kernel.parameters.Kernel.metaparams in
      let pre_act = List.rev (!ac).Kernel.parameters.Kernel.actions in
      let pre_color = List.rev (!ac).Kernel.parameters.Kernel.colormaps in
      let pre_description = (!ac).Kernel.description in
      let rec reinject r w = function
	| [] -> ()
	| t::q -> (if not (List.exists (fun x -> fst x = fst t) (r ())) then w (t::(r ())); reinject r w q)
      in
      let reinit = reinject (fun () -> (!ac).Kernel.initconfs) (fun x -> (!ac).Kernel.initconfs <- x) in
      let remetap = reinject (fun () -> (!ac).Kernel.parameters.Kernel.metaparams) (fun x -> (!ac).Kernel.parameters.Kernel.metaparams <- x) in
      let react = reinject (fun () -> (!ac).Kernel.parameters.Kernel.actions) (fun x -> (!ac).Kernel.parameters.Kernel.actions <- x) in
      let recolor = reinject (fun () -> (!ac).Kernel.parameters.Kernel.colormaps) (fun x -> (!ac).Kernel.parameters.Kernel.colormaps <- x) in
      let redescription what = (!ac).Kernel.description <- what in
      let re () = (
	reinit pre_init;
	remetap pre_metap;
	recolor pre_color;
	react pre_act;
	(* ca.Kernel.theinitconf <- pre_theinit *)
      ) in
	Kernel.resetinitconf (!ac);
	Kernel.resetmisc (!ac);
	if not !initconfset then (!ac).Kernel.theinitconf <- "Random uniform";
	Kernel.block (!ac); 
	(try
	   match do_it (Load.default_comp_env ()) text with
	     | Some (true,true) -> donewboth false
	     | Some (true,false) -> donewboth false
	     | Some (false,true) -> re (); donewinit false
	     | Some (false,false) -> re (); donewneno ()
	     | None -> if Kernel.last_load_is_family () then (something_loaded:=true; doopen true)
	 with Not_found -> (comboinfo Dbgwarning "Shortcut not recognized: aborting."; 	
			    redescription pre_description;
			    re()));
	params:=[];
	Kernel.unblock (!ac)
  in

  (******************************************)
  (* Initialization                         *)
  (******************************************)
  let execstack = ref (fun cont contafter -> (
			 ondebug Dbgdev versioninfo;
			 ondebug Dbgdev (fun () -> adddebugcat Dbgtk; adddebugcat Dbgsdl);
			 (!ac).Kernel.dbg <- (getdebugcats ());
			 debugline Dbgdev "Initialization...";
			 sysdeprandominit ();
			 if (!Interactive.interactive) then (
			   debugline Dbgdev "Interfaces...";
			   putvoidsighandler ();
			   Interactive.sdl_initialization !fake_sdl;
			   Keymap.check_multiple ();
			   Keymap.generate_help ();
			   if Interactive.use_gtk () then Interactive.releaseInterface Interactive.gtkWindow "GTK";
			   Interactive.set_universe !ac;
			   debugline Dbgdev "Interfaces done.";
			 );
			 allow_siginterrupt ();
			 (* ACML is now initialised, it's time to do some command-line driven stuff *)
			 (* First, the after-init stuff driven by command-line *)
 			 cont ();
			 (* Then, prepare to use parameters from command-line if any *)
			 debugline Dbgdev (Printf.sprintf "%d params set from command-line" (List.length !cmdline_params));
			 params := !cmdline_params;
			 (* Then execute all shortcuts from the command-line *)
			 List.iter (do_shortcut !matrice !time !steps false) (get_shortcuts ());
		 	 (* At this point, since command-line did not
			    result in program termination, it means
			    that we have to load some object if none is already loaded by command-line and
			    display it as usual *)
			 if not !something_loaded then (
			   let ch = (random_list (Kernel.get_builtin_nicks ())) in
			   debugline Dbginteract ("Loading a randomly chosen family... "^ch^"!");
			   do_addselect (Load.default_comp_env ()) false true ch;
			   ac := evalacml !ac Kernel.Newall (theprogressmeter ());
			   pushsos !ac;
			   Kernel.preparematrix !resetmatonnewac matrice !ac !steps; 
			   giveinfo !matrice !ac;
			   if !autosave then dosave !savfile;
			   if (!Interactive.interactive) then (
			     Interactive.geometry !ac;
			     Interactive.puttitle (history_memsize ()) !ac;
			     Interactive.show_params !matrice !ac;
			     Interactive.show_plugins !ac;
			     Interactive.colorcache !ac;
			     Interactive.unworking ();
			   );
			   newstart ();
			 );
			 debugline Dbgdev "Object loaded.";
			 if !Interactive.interactive then (
			   if !fullscree then (
			     Interactive.toggle_fullscreen (); 			     
			     Interactive.colorcache !ac;
			     Interactive.unworking ();
			   );
			 );
			 debugline Dbgdev "Init done.";
			 Kernel.apply_plugins !ac Kernel.Apresinitialisation;
			 (* Everything is done, it's time to do command-line finalization *)
			 contafter ();
			 if !Interactive.interactive && !run then (
			   run := not !run;
			   dostartstop ();
			 )
		       )
		      ) 
  in let dothisafterinit myfun =
      let before = !execstack in
	execstack := (fun cont contafter -> before (fun () -> myfun ();  cont ()) contafter)
  in let dothisafterload myfun =
      let before = !execstack in
	execstack := (fun cont contafter -> before cont (fun () -> myfun ();  contafter ()))
  in let show (la,lb) = 
      let mx = List.fold_left (fun m s -> max m (String.length s)) 0 la +1 in
      let pad s = s^(String.make (mx - (String.length s)) ' ') in
      let rec doit = function
	  (t::q, u::r) -> Printf.printf "%s---> %s\n" (pad t) u; doit (q,r)
	| _ -> ()
      in doit (la,lb)
  in let plugshow p =
      Printf.printf "--- %s (%s) ---\n %s\n\n" p.Kernel.pg_name (Kernel.string_from_repetition p.Kernel.pg_repetition) p.Kernel.pg_description
  in  
    (******************************************)
    (* Command line                           *)
    (******************************************)
    (* First argument of the uple is a list of alternatives for command name *)
  let mypreargs =
    [
      (["--shootfile"; "-g"], Arg.String (fun x -> imgfile := Some x), "STRING Ouput file name for screenshots") ; 
      (["--colormap"; "-c"], Arg.String (fun x -> dothisafterload (fun () ->
	debugline Dbgdev ("Setting colormap to: "^x);
	(!ac).Kernel.colormap <- x)), "STRING Set colormap to use (if wrong name silently fails)") ; 
      (["-x"], Arg.Int (fun x -> Kernel.keep_universe (); Kernel.absupdate !ac x), "INT Number of cells along horizontal axis") ;
      (["-y"], Arg.Int (fun y -> Kernel.keep_universe (); Kernel.ordupdate !ac (max 2 y)), "INT Number of cells along vertical axis (time in 1D), must be >1") ;
      (["--refresh_rate"; "-r"], Arg.Set_int Interactive.showrate, "INT Number of steps between 2 screen refresh") ;
      (* (["--preperiod"], Arg.Set_int steps, "INT Number of steps before periodic conditions apply") ; *)
      (["--video_x"; "-vx"], Arg.Set_int Interactive.dx, "INT Width in pixels of the view window");
      (["--video_y"; "-vy"], Arg.Set_int Interactive.dy, "INT Height in pixels of the view window");
      (["--fps"], Arg.Set_float Interactive.cadence, "FLOAT Frames per second") ;
      (["--outfile"; "-o"], Arg.String (fun x -> savfile := Some x), "STRING Output filename for rule saving") ;
      (["---"], Arg.Rest (flipflopparams cmdline_params), 
       "[...] read parameters as a sequence of type \"identifier1 value1 identifier2 value2...\"") ;
      (["--version"; "-v"], Arg.Unit (fun () -> versioninfo (); exit 0), " Print program version and exit") ;
      (["--documentation"; "-doc"], Arg.Unit (fun () -> List.iter (function a,b ->
								     debugline Dbginteract (a^"  ----------");
								     debugline Dbgeneral b;
								     print_newline ()
								  ) HELP.rubriks; exit 0), 
       " Print builtin documentation and exit") ;      
      (["--keyboard_help"; "-kh"], Arg.Unit (fun () -> Interactive.disable_gtk ();fake_sdl:=true;
					       dothisafterinit (fun () -> (
								  debugline Dbginteract "Keyboard shorcuts";
								  debugline Dbgeneral !Keymap.keymaptext; exit 0))), 
       " Print keyboard shortcuts and exit") ;
      (["--info"; "-n"], Arg.Unit (fun () ->  desinteraction (); (* interactive:=false; *)
				     dothisafterload (fun () -> (param_info !ac; exit 0))), 
       " Give extended info on current object and exit") ;
      (["--identify"; "-id"], Arg.Unit (fun () ->  desinteraction (); (* interactive:=false; *)
				     dothisafterload (fun () -> (print_string (identify !ac); exit 0))), 
       " Give short description of current object and exit") ;
      (["--licence"], Arg.Unit (fun () -> licenceinfo (); exit 0), " Print licence information and exit") ;
      (["--init_conf"; "-i"], Arg.String (fun x -> dothisafterinit (fun () -> initconfset := true; (!ac).Kernel.theinitconf <- x)), 
       "STRING Name of the initconf type to set") ;
      (["--debug"; "-d"], Arg.Unit (fun () -> adddebugcat Dbgdev), " Enable debug messages output");
      (["--guidebug"], Arg.Unit (fun () -> adddebugcat Dbgdevsdl; adddebugcat Dbgsdl; adddebugcat Dbgtk), 
       " Enable GUI-related debug messages");
      (["--quiet"; "--silent"], Arg.Unit (fun () -> setdebugcats []), " No info message output on stderr");
      (["--unlimited_history"; "-uh"], Arg.Unit (fun () -> history_set_keep true), " Unlimited history of objects");
      (["--plugin_on"; "-pon"], Arg.String (fun x -> dothisafterinit (fun () -> Kernel.plugin_on x)), 
       "STRING Switch on the named plugin") ;
      (["--plugin_off"; "-poff"], Arg.String (fun x -> dothisafterinit (fun () -> Kernel.plugin_off x)), 
       "STRING Switch off the named plugin") ;
      (["--batch"; "-b"], Arg.Unit (fun () -> run := true; desinteraction (); Interactive.disable_gtk (); Interactive.cadenced:= false), " Turn into batch mode (not interactive, auto-run, no delay)") ;
      (["--start_with"], Arg.String (fun s -> Kernel.set_builtin_bynick s), "STRING Name of builtin rule to load at startup");
      (["--list_families"; "--list_builtin"; "-l"], Arg.Unit (fun () -> 
					     show (Kernel.get_builtin_nicks (), Kernel.get_builtin_names ()); exit 0), 
       " Output the list of builtin families and exit");
      (["--list_plugins"; "-lp"], Arg.Unit (fun () -> let p = Kernel.list_plugins () and hp = Kernel.list_hook_plugins () in
					      Printf.printf "*** List of available plugins\n\n"; 
					      List.iter plugshow p;
					      Printf.printf "*** List of available 'hook' plugins\n\n"; 
					      List.iter plugshow hp;
					      exit 0), 
       " Output the list of plugins and exit");
      (["--list_objects"; "-lo"], Arg.Unit (fun () -> List.iter (fun s ->  Printf.printf "%s\n" s) 
					      ((Kernel.get_objects ())); exit 0), 
       " Output the list of builtin objects and exit (use short names to call a rule)");
      (["--list_shortcuts"; "-ls"], Arg.Unit (fun () -> 
						let show (rex,nick,_) =  Printf.printf "%s ==> %s\n" rex nick in
						  List.iter show (Kernel.list_shortcuts ()); exit 0), 
       " Output the list of shortcuts and exit");
      (["--shooting"], Arg.Set shooting, " Turn on shooting mode") ;
      (["--fullscreen"], Arg.Set fullscree, " Turn on fullscreen mode") ;
      (["--noshooting"], Arg.Clear shooting, " Turn off shooting mode (default)") ;
      (["--intervalshoot"], Arg.Int (fun x -> intervalshoot:=x), "INT Interval between two shots in shooting mode (default 0)") ;
      (["--stopandshoot"; "-sas"], Arg.Int (fun x -> Kernel.stoptime:= x; shooting:=true; intervalshoot:=x), "INT Time at which to stop and shoot once (implies stop and shooting and set intervalshoot accordingly)") ;
      (["--run"], Arg.Set run, " Turn on running mode at startup (don't by default, usefull only for interactive mode)") ;
      (["--autosave"], Arg.Set autosave, " Each time a new rule is generated save it (don't by default)" );
      (["--cadence"], Arg.Int (fun x -> Interactive.cadenced:=true; Interactive.cadence:=(float_of_int x)), "INT Try to display the specified number of frames per second" );
      (["--nogtk"], Arg.Unit (Interactive.disable_gtk), " Disable GTK interface (some features may become unreachable)");
      (["--nocontrolwindow"; "-nc"], Arg.Unit (Interactive.disable_control), " Disable GTK control window (other GTK features are still available)");
      (["--slideshow"], Arg.Unit (let rec progtrap () = donewboth false; timetrap 10000 (fun () -> (progtrap ())) in fun () -> dothisafterinit (fun () -> timetrap 1000 progtrap)),
       " Automatically change AC/init conf every 10 seconds" );
      (["--popupsave"], Arg.Set popupsave, " Interactively asks the user for a name each time something is to be saved" );
      (["--stop"; "-t"], Arg.Set_int Kernel.stoptime, "INT Set stopping time (-1 means never)") ;
    ]
  in 
  let alternative_expand (list, what, info) = 
    match list with
	[] -> internalbug "No alternative in command name"
      | t::q -> let see = " See '"^t^"'" in (t,what,info)::(List.map (fun nme -> (nme,what,see)) q)
  in
  let myargs = Arg.align (List.sort compare (List.flatten (List.map alternative_expand mypreargs))) in
  let react_to = function
    | Keymap.Stopstart -> dostartstop ()
    | Keymap.Stop -> if !run then dostartstop ()
    | Keymap.Newinit -> donewinit true
    | Keymap.Newconfsize (x,y) -> 
      (
	(!ac).Kernel.universe <- (x,y);
	donewinit false
      )
    | Keymap.Undo -> (
      match popsos true with
	  None -> comboinfo Dbginteract "nothing left to undo!" 
	| Some x -> doundo x; comboinfo Dbginteract "undo completed!" 
    )
    | Keymap.Revert -> (
      match popsos false with
	  None -> comboinfo Dbginteract "nothing to revert!" 
	| Some x -> doundo x; comboinfo Dbginteract "revert completed!" 
    )
    | Keymap.Randomparams -> (
      Kernel.putrandomparams !ac;
      if !Interactive.interactive then Interactive.show_params !matrice !ac
    )
    | Keymap.Builtin nick -> (
      do_addselect (Load.default_comp_env ()) false true nick;
      doopen true
    )
    | Keymap.Newboth -> donewboth true
    | Keymap.Newcolormap -> donewcolormap true
    | Keymap.Nextcolormap -> donewcolormap false
    | Keymap.Memorize -> Kernel.memorize !ac !matrice !time; Interactive.setstatus "Configuration memorized"
    | Keymap.Load -> (
      Kernel.load !ac !matrice !time !steps; 
      Interactive.setstatus "Configuration loaded"; 
      Interactive.refresh true !ac !matrice !time !fps !steps false;
    )
    | Keymap.Onestep -> (
      onestep := true;
      Interactive.running ()
    )
    | Keymap.AddCache -> (
      try
	Kernel.addcache !ac
		with Kernel.NoNeedCache -> Interactive.setstatus "No need to add cache"
    )
    | Keymap.Output -> Kernel.outputrule !ac
    | Keymap.Screenshot ->  dosnapshot !ac !matrice !time !steps
    | Keymap.Measurement m -> Kernel.do_single m !ac !matrice !time !steps
    | Keymap.Saverule -> dosave !savfile
    | Keymap.SaveAs f -> dosave (Some f)
    | Keymap.Newneno -> donewneno ()
    | Keymap.Nothing -> ()
    | Keymap.Help -> Interactive.launch_help !ac
    | Keymap.Licence -> Interactive.launch_licence ()
    | Keymap.Fullscreen -> (
      Interactive.toggle_fullscreen ();
      Interactive.colorcache !ac;
      Interactive.unworking ();
      Interactive.refresh true !ac !matrice !time !fps !steps false;		  
    )
    | Keymap.Zoomout -> (
      Interactive.zoomout ();
      Interactive.infozoom (); 
      if (not !run) then Interactive.refresh true !ac !matrice !time !fps !steps false
      else clearonce :=true
    )
    | Keymap.Zoomat (x,y) -> (
      Interactive.infozoom (); 
      if (not !run) then Interactive.refresh true !ac !matrice !time !fps !steps false
      else clearonce :=true
    )
    | Keymap.Centerzoom -> (
      Interactive.centerzoom (); 
      Interactive.infozoom (); 
      Interactive.refresh true !ac !matrice !time !fps !steps false;
      
    )
    | Keymap.Slow -> (
      if !Interactive.showrate > 1 then decr Interactive.showrate;
      Interactive.refreshranges ()
    )
    | Keymap.Accel -> (
      incr Interactive.showrate;
      Interactive.refreshranges ()
    )		
    | Keymap.Askexit -> (exit_not_asked := false; if !Interactive.interactive then Interactive.closeAll ())
    | Keymap.Forcedexit -> debugline Dbgwarning "Brutal exit!"; exit 0;
    | Keymap.Redraw -> Interactive.refresh true !ac !matrice !time !fps !steps false
    | Keymap.Resize (x,y) -> (
      Interactive.window_resize x y;
      Interactive.recalibrate ();
      Interactive.refresh true !ac !matrice !time !fps !steps false
    )
    | Keymap.Unmark -> (
      Kernel.setoverdrawings !ac (List.filter (function Kernel.Markcell _ -> false | _ -> true) 
				    (Kernel.getoverdrawings !ac))
    )
    | Keymap.HistoryClear -> (
      clean_history ();
      Interactive.puttitle (history_memsize ()) !ac;
    )
    | Keymap.Switchhistory -> history_set_keep (not (history_keep ()))
    | Keymap.Newhazard -> sysdeprandominit ()		    
    | Keymap.Switchgrid -> ((match !ac.Kernel.infotodraw.Kernel.over with
	[] -> (!ac.Kernel.infotodraw.Kernel.over <- [Kernel.GenGrid (1,1,250,250,250)])
	| _ -> (!ac.Kernel.infotodraw.Kernel.over <- []);
    );
			    Interactive.refresh true !ac !matrice !time !fps !steps !run
    )
    | Keymap.Zone (src, ((ax,ay),(bx,by))) -> 
      domodifzone src ax ay bx by
    | Keymap.ZoneDiag (src, ((ax,ay),(bx,by))) -> 
      domodifzonediag src ax ay bx by
    | Keymap.ZoneSide (src, ((ax,ay),(bx,by))) -> 
      domodifzoneside src ax ay bx by
    | Keymap.Shortcut text -> (
      try 
	do_shortcut !matrice !time !steps false text
      with Load.FatalSubProcessError | Dynlink.Error _ | Load.NothingDone -> 
	comboinfo Dbgwarning "Problem in compilation or linking: nothing loaded!"
    )
    | Keymap.SwitchPlugin (plug,active) -> (
      (if active then Kernel.plugin_on else Kernel.plugin_off)  plug.Kernel.pg_name;
      if plug.Kernel.pg_repetition = Kernel.Apreschaqueobjet then donewneno ()
    )
    | Keymap.Revert_from_file -> (
      let dothejob () =
	try
	  Load.revert_from_file ();
	  doopen true
	with _ -> comboinfo Dbgwarning "Problem in source file compilation: nothing loaded!"
      in if Kernel.last_load_is_plugin () then dothejob () else 
	  match Kernel.get_src_file !ac with
	      None -> comboinfo Dbginteract "Current object is built-in: nothing to do!"
	    | Some _ -> dothejob ()
    )
  in
  let doshooting () =
    percentprogression ();
    if (!shooting && (!shootcnt = !intervalshoot)) then (
      dosnapshot !ac !matrice !time !steps;
      shootcnt := 0
    ) else incr shootcnt
  in
    (
      Kernel.postset_builtin ();
      Arg.parse myargs group_shortcuts acmlwelcome;
      !execstack (fun ()-> ()) (fun () -> ());

      (******************************************)
      (* main loop                              *)
      (******************************************)

      while (!exit_not_asked && (!Interactive.interactive || !run)) do 
	testtimetrap ();
	if (!run || !onestep) then 
	  (

            (* do a snapshot if shooting mode and set interval ask so *)
            doshooting ();
            
            (* Run one step (and do all measurements if any) *)
 	    Kernel.run !ac !matrice time !steps;

	    
	    (* Apply eventually some plugins *)
 	    Kernel.apply_plugins !ac Kernel.Apreschaqueetape;

	  );
	onestep := false;
	if (!Interactive.interactive && (!time = !Kernel.stoptime)) || (not !Interactive.interactive && !time >= !Kernel.stoptime && !Kernel.stoptime >= 0) 
	then (
          run := false;
          (* do a snapshot if shooting mode and set interval ask so *)
          doshooting ()
        );
	(* Interpret scenario events *)
	Keymap.play_scenario react_to;
	if (!Interactive.interactive) 
	then (
	  Interactive.regulate_cadence (!ac).Kernel.dimension !time;
	  if (!run || !onestep) then dofpsinfo fps;
	  Interactive.refresh !clearonce !ac !matrice !time !fps !steps !run;		    
	  clearonce := false;
	  react_to (if !run then Interactive.pollInteraction !ac false else Interactive.waitInteraction !ac)
	)
      done
    )

        
let _ = (
  (* Is this useful or harmful? *)
  (* Gc.set { (Gc.get()) with Gc.minor_heap_size = Kernel.ints_in_cache }; *)
  at_exit (fun () -> flush stdout; flush stderr);  
  try 
    main ()
  with Kernel.Unimplemented s -> (
    debugline Dbgfatal ("Sorry! you tried to use a feature ("^s^") which is not yet implemented.");
    exit 1;
  )
    | Load.FatalSubProcessError | Dynlink.Error _ -> (
	debugline Dbgfatal "Unexpected error with source file so terminates!";
	exit 3
      )
    | Kernel.BadACML s -> (
	debugline Dbgfatal ("ACML object with the current parameters is not valid: "^s)
      )
    | Kernel.AbortRule s -> (
	debugline Dbgfatal ("ACML object with the current parameters is not valid: "^s)
      )
    | e -> (
	debugline Dbgfatal ("Sorry! the program contains a bug."); 
	debugline Dbgfatal ("Here is some info: '"^(Printexc.to_string e)^"'");
	unlessdebug Dbgdev (fun () -> debugline Dbgfatal 
			      "You should try to reproduce the problem with the '--debug' option.");
	exit 2;
      )
)
