(* This example shows how to use parameters. This is one of the
   benefits of using transformations over objects instead of
   objects directly.

   A parameter is just a function returning a value chosen via the
   user interface; each time the user wants to test a new parameter
   value, he (or she) sets the parameter and then asks for a new
   evaluation of the transformation function. 
*)



open Acml

let transform a =
  
  (* First we create an int parameter between 2 and 50 (3 by default)
     which can be controlled via the user interface under the
     parameter name 'Number of states' *)
  let n = param a "Number of states" 2 50 3  in
    set_cardinality a n;
    
    (* Same idea for dimension and neighbourhood. By the way, we
       discover another kind of parameter called 'choice\_param' and
       returning a string among a list of possible choices.*)
    set_dimension a (param a "Dimension" 1 2 1);
    let r = param a "Radius" 1 50 1 in
      (match 
	 choice_param a "neighborhood" [ "Moore"; "von Neumann"]  
       with 
	   "von Neumann" -> set_vonneumann_neighbourhood a r
	 | _ -> set_moore_neighbourhood a r);   
      
      (* Now we set the local rule: sum modulo 'n' of the neighbours. The form
	 of the local rule is independant of dimension and neighbourhood
	 but depends on the number of states.*)
    set_rule a (fun x -> Array.fold_left (+) 0 x mod n)
      

let _ = valid transform

(* From the user interface one can now easily explore the algebraic
'sum modulo n' rule for any n (up to 50), any radius (up to 50), and
dimension 1 or 2. *)

(* There are other parameter types in the API. *)

