(* This example defines the famous Brian's Brain automaton inspired
from the excitable medium metaphor. Two aspects of the API are illustrated here: polymorphic API calls and multiset rules. *)

open Acml;;

(* To make the definition more readable, we can use any OCaml type for
   states.
*)
type state = Quiescent | Excited | Refractory

(* Then we must declare the set of objects of that type which will form the set of states of the cellular automaton. *)
let alphabet = AlphaArray [|Quiescent; Excited; Refractory|]

(* From that point on, we can use 'polymorphic' versions of functions from the API. *)

let transform a =  
  set_dimension a 2;
  set_moore_neighbourhood a 1;

  (* We now define the local transition rule of the automaton. Here, states are OCaml values of type 'state' (and not integers). Moreover, the local function defined has two parameters: the first (of type 'state') gives the current state of the cell, and the second of type 'state $\rightarrow$ int' gives for each state its number of occurrences in the neighborhood. *) 
  let transition_rule cell howmany = match cell with
      Excited -> Refractory
    | Quiescent when howmany Excited = 2 -> Excited
    | _ -> Quiescent
  in
    
    (* This local function is registered via the appropriate API call. *)
    poly_set_multiset_rule a ~alphabet transition_rule;
    set_name a "Brian's Brain"
    
let _ = valid transform;;
