(* This example shows how to use (pseudo-)randomness in rule
   definitions and/or configurations. 

   A cellular automaton can be deterministic or stochastic depending
   on the definition of the local rule.  If you give a deterministic
   local rule, you get a deterministic CA. If you give a local rule
   using (pseudo-)randomness, there are two possibilities: either you
   want to define a stochastic CA, or you want to define a
   deterministic CA where some transitions are chosen at random.

   ACML can deal with all scenarii provided you specify what you
   want. Moreover, be it for rules or configurations, ACML can manage
   random seed and transparently save/load them to allow
   reproducibility of experiments.

   For things to work as you expect, observe the following discipline:
   \begin{enumerate}
   \item use functions from the API for (pseudo-)random generation
   \begin{itemize}
   \item \texttt{ca\_random} when it concerns the rule,
   \item \texttt{conf\_random} when it concerns the configuration;
   \end{itemize}
   \item specify whether you want a deterministic or a stochastic CA.
   \end{enumerate}
*)


open Acml;;

let transform a =  
  set_dimension a (param a "dimension" 1 2 2);
  set_cardinality a (param a "states" 2 100 5);
  set_moore_neighbourhood a (param a "radius" 1 50 1);

  (* Now we define a local function which uses (pseudo-)randomness: it chooses at random one of the states appearing in the neighbourhood. Since it concerns the local rule, we use \texttt{ca\_random} from the API to get random numbers. *)
  let local_function x = x.(ca_random (Array.length x)) in

    (* We must tell ACML what we want (stochastic or deterministic) before registering the local rule. To easily test all possibilities via the user interface, let's define some choice parameter. *)
    (match choice_param a "determinism type" ["I want it stochastic"; 
					      "I want it deterministic";
					     "I foret to specify what I want" ] with
	 "I want it stochastic" -> declare_stochastic a
       | "I want it deterministic" -> ensure_determinism a
       | _ -> ());

    (* Then the function is registered as usual. *)
    set_rule a local_function
      
let _ = valid transform

(* Let's explain what happens in each situation:
\begin{itemize}
   \item \textit{I want it stochastic}: you get always the same stochastic CA;
   \item \textit{I want it deterministic}: you get a deterministic CA chosen randomly among 'captive cellular automata' (i.e. CA for which any transition leads to a state which was already present in the neighbourhood);
   \item \textit{I foret to specify what I want}: by default, ACML considers that the local rule you give is deterministic, and, for efficiency, it can decide to add a cache mecanism; therefore you get a deterministic CA chosen randomly among captive CA when the cache mecanism is added (for small size CA), and the stochastic CA mentioned above when no cache mecanism is added (for large CA).
\end{itemize}
*)
